/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano;

import java.awt.BorderLayout;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.LinkedList;
import javax.media.opengl.GLCanvas;
import javax.media.opengl.GLCapabilities;
import javax.swing.JFrame;

import openpiano.control.ControlPanel;
import openpiano.control.Controller;

/**
 * The main application
 *
 * @author  Sascha Westermann
 * @version 1.0, 07/30/08
 */
public class OpenPiano extends JFrame implements WindowListener, ComponentListener {
	private static final long serialVersionUID = -1875366596758027498L;
	private Controller controller;
	private int mouseX, mouseY;
	private final int NOTSET = 0;
	private final int HORIZONTAL = 1;
	private final int VERTICAL = 2;
	private int direction = NOTSET;
	private LinkedList<Integer> directionList;
	int mouseButton;
    
	/**
	 * Constructs a new main application
	 */
	public OpenPiano() {
		super("OpenPiano");
		
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		setLayout(new BorderLayout());
		
		controller = new Controller(this);
		
		GLCapabilities glc = new GLCapabilities();
		glc.setSampleBuffers(true);
		glc.setNumSamples(4);
		GLCanvas glCanvas = new GLCanvas(glc);
        glCanvas.addGLEventListener(controller.getGLCore());
        
        glCanvas.addMouseMotionListener(new MouseMotionListener() {
        	public void mouseDragged(MouseEvent e) {
        		int draggedX = e.getX();
        		int draggedY = e.getY();
        		if (direction == NOTSET) {
        			if (directionList.size() != 5) {
		        		if (Math.abs(draggedX - mouseX) > Math.abs(draggedY - mouseY))
		        			directionList.add(HORIZONTAL);
		        		else
		        			directionList.add(VERTICAL);
        			}
        			else {
        				int hcounter = 0;
        				for (Integer i : directionList) {
        					if (i == HORIZONTAL)
        						hcounter++;
        				}
        				if (hcounter > (int)(directionList.size()/2.0f))
        					direction = HORIZONTAL;
        				else
        					direction = VERTICAL;
        				}
        		
        				
        		}
        		else {
        			if (direction == HORIZONTAL) {
        				float oldX = controller.getCamera().getXPosition();
        				int difference = e.getX() - mouseX;
       					controller.getCamera().setXPosition(oldX + difference*0.1f);
        			}
        			else {
        				float oldAngle = controller.getSettings().getAngle();
        				int difference = e.getY() - mouseY;
        				controller.getCamera().setAngle(oldAngle + difference*0.3f);
        			}
        		}
        			
           		mouseX = e.getX();
        		mouseY = e.getY(); 
        	}
        	public void mouseMoved(MouseEvent e) {}
        });
        
        glCanvas.addMouseListener(new MouseListener() {
        	public void mousePressed(MouseEvent e) {
           		direction = NOTSET;
           		directionList = new LinkedList<Integer>();
           		mouseX = e.getX();
        		mouseY = e.getY();
        		if (e.getButton() == MouseEvent.BUTTON1)
        			mouseButton = MouseEvent.BUTTON1;
        		else if (e.getButton() == MouseEvent.BUTTON3)
        			mouseButton = MouseEvent.BUTTON3;
        	}
        	public void mouseReleased(MouseEvent e) { 
        		if (e.getButton() == MouseEvent.BUTTON1)
        			mouseButton = -1;
        		else if (e.getButton() == MouseEvent.BUTTON3)
        			mouseButton = -1;
        		mouseX = e.getX();
        		mouseY = e.getY();
        	}
        	public void mouseClicked(MouseEvent e) {}
        	public void mouseEntered(MouseEvent e) {}
        	public void mouseExited(MouseEvent e) {}
        });
        
        glCanvas.addMouseWheelListener(new MouseWheelListener() {
        	public void mouseWheelMoved(MouseWheelEvent e) {
        		controller.getCamera().setZoomFactor(controller.getSettings().getZoomFactor() - (e.getWheelRotation()/20.0f));
        	}
        });
        
        add(glCanvas, BorderLayout.CENTER);
        
        add(new ControlPanel(controller), BorderLayout.SOUTH);

        addWindowListener(this);
        addComponentListener(this);
		setSize(controller.getSettings().getWindowWidth(), controller.getSettings().getWindowHeight());
		setLocation(controller.getSettings().getWindowPosX(), controller.getSettings().getWindowPosY());
		setResizable(true);
        setVisible(true);
	}
   
	public void windowActivated(WindowEvent we) {}
	public void windowClosed(WindowEvent we) {}
	public void windowDeactivated(WindowEvent we) {}
	public void windowDeiconified(WindowEvent we) {}
	public void windowIconified(WindowEvent we) {}
	public void windowOpened(WindowEvent we) {}
	
	/**
	 * Closes the applications, saves the settings and mutes active notes
	 * 
	 * @param we the event handler
	 */
	public void windowClosing(WindowEvent we) { 
		controller.getSettings().saveSettings();
		if (controller.getSettings().getMidiOutput() != null)
			controller.getSettings().getMidiOutput().allNotesOff(); 
	}
	
	/**
	 * Sets the new dimension of the application window
	 * 
	 * @param e the event handler
	 */
	public void componentResized(ComponentEvent e) {
		controller.getSettings().setWindowHeight(e.getComponent().getHeight());
		controller.getSettings().setWindowWidth(e.getComponent().getWidth());
	}
	
	/**
	 * Sets the new position of the application window
	 * 
	 * @param e the event handler
	 */
	public void componentMoved(ComponentEvent e) {
		if (e.getComponent().getWidth() >= e.getComponent().getGraphicsConfiguration().getBounds().getWidth())
			controller.getSettings().setWindowPosX(0);
		else
			controller.getSettings().setWindowPosX(e.getComponent().getX());

		if (e.getComponent().getHeight() >= e.getComponent().getGraphicsConfiguration().getBounds().getHeight())
			controller.getSettings().setWindowPosY(0);
		else
			controller.getSettings().setWindowPosY(e.getComponent().getY());
	}

	public void componentShown(ComponentEvent e) {}
	public void componentHidden(ComponentEvent e) {}
	
	/**
	 * Creates the application, no arguments are used
	 * 
	 * @param args not relevant
	 */
	public static void main(String[] args) {
		new OpenPiano();
	}
}