/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.control;

import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JPanel;

/**
 * A panel to handle control elements
 * 
 * @author Sascha Westermann
 * @version 1.0, 07/30/08
 */

public class ControlPanel extends JPanel implements ActionListener, MouseListener {
	private static final long serialVersionUID = -2122319809085500512L;
	private JButton onoffButton, playButton, stopButton, pauseButton, loadButton, repeatButton, settingsButton;
	private SlidePanel slidePanel;
	private Controller controller;
	private ClassLoader cl;
	
	/**
	 * Construct a new control panel
	 * 
     * @param controller the main controller of the application
	 */
	public ControlPanel(Controller controller) {
		this.controller = controller;
		slidePanel = new SlidePanel(controller);

		controller.registerControlPanel(this);
		controller.registerSlidePanel(slidePanel);
		
		Border empty = new EmptyBorder(0, 0, 0, 0);
        cl = this.getClass().getClassLoader();

		onoffButton = new JButton();
		playButton = new JButton();
        stopButton = new JButton();
        pauseButton = new JButton();
        repeatButton = new JButton();
        loadButton = new JButton();
        settingsButton = new JButton();
        
        onoffButton.setContentAreaFilled(false);
        playButton.setContentAreaFilled(false);
        stopButton.setContentAreaFilled(false);
        pauseButton.setContentAreaFilled(false);
        repeatButton.setContentAreaFilled(false);
        loadButton.setContentAreaFilled(false);
        settingsButton.setContentAreaFilled(false);

        // Create icons
        onoffButton.setIcon(new ImageIcon(cl.getResource("icons/onoff.png")));
        playButton.setIcon(new ImageIcon(cl.getResource("icons/play.png")));
        stopButton.setIcon(new ImageIcon(cl.getResource("icons/stop.png")));
        pauseButton.setIcon(new ImageIcon(cl.getResource("icons/pause.png")));
        repeatButton.setIcon(new ImageIcon(cl.getResource("icons/repeat.png")));
        loadButton.setIcon(new ImageIcon(cl.getResource("icons/load.png")));
        settingsButton.setIcon(new ImageIcon(cl.getResource("icons/settings.png")));

        onoffButton.setBorder(empty);
        playButton.setBorder(empty);
        stopButton.setBorder(empty);
        pauseButton.setBorder(empty);
        repeatButton.setBorder(empty);
        loadButton.setBorder(empty);
        settingsButton.setBorder(empty);
        
        onoffButton.setFocusable(false);
        playButton.setFocusable(false);
        stopButton.setFocusable(false);
        pauseButton.setFocusable(false);
        repeatButton.setFocusable(false);
        loadButton.setFocusable(false);
        settingsButton.setFocusable(false);
        
        onoffButton.addActionListener(this);
        playButton.addActionListener(this);
        pauseButton.addActionListener(this);
        stopButton.addActionListener(this);
        repeatButton.addActionListener(this);
        loadButton.addActionListener(this);
        settingsButton.addActionListener(this);

        onoffButton.addMouseListener(this);
        playButton.addMouseListener(this);
        pauseButton.addMouseListener(this);
        stopButton.addMouseListener(this);
        repeatButton.addMouseListener(this);
        loadButton.addMouseListener(this);
        settingsButton.addMouseListener(this);
        
        setLayout(new GridBagLayout());
        GridBagConstraints gbc = new GridBagConstraints();

        gbc.fill = GridBagConstraints.BOTH;
        gbc.anchor = GridBagConstraints.WEST;
        gbc.weighty = 1.0;
        gbc.weightx = 0.2;
        gbc.gridx = 0;
        gbc.gridy = 0;
        
        add(playButton, gbc);
        gbc.gridx = 1;
        add(pauseButton, gbc);
        gbc.gridx = 2;
        add(stopButton, gbc);
        
        gbc.anchor = GridBagConstraints.CENTER;
        gbc.gridx = 3;
        gbc.weightx = 1.0;
        add(slidePanel, gbc);

        gbc.anchor = GridBagConstraints.EAST;
        gbc.gridx = 4;
        gbc.weightx = 0.2;
        add(repeatButton, gbc);
        gbc.gridx = 5;
        add(loadButton, gbc);
        gbc.gridx = 6;
        add(settingsButton, gbc);
        gbc.gridx = 7;
        add(onoffButton, gbc);
	}
	
	/**
	 * Show an inactive load icon
	 */
	public void midiLoaderGUIClosed() {
		loadButton.setIcon(new ImageIcon(cl.getResource("icons/load.png")));
	}

	/**
	 * Show an inactive settings icon
	 */
	public void settingsGUIClosed() {
		settingsButton.setIcon(new ImageIcon(cl.getResource("icons/settings.png")));
	}
	
	/**
	 * Show an inactive repeat icon
	 */
	public void repeatAborted() {
		repeatButton.setIcon(new ImageIcon(cl.getResource("icons/repeat.png")));
	}
	
	/**
	 * Show an inactive play icon
	 */
	public void playStopped() {
		playButton.setIcon(new ImageIcon(cl.getResource("icons/play.png")));
	}

	/**
	 * Show an active play icon
	 */
	public void playStarted() {
		playButton.setIcon(new ImageIcon(cl.getResource("icons/play_active.png")));
	}

	public void actionPerformed(ActionEvent e) {
		  Object comp = e.getSource();
		  if (comp instanceof JButton) {
			  JButton b = (JButton)comp;
			  if (b == onoffButton) {
					controller.getSettings().saveSettings();
					if (controller.getSettings().getMidiOutput() != null)
						controller.getSettings().getMidiOutput().allNotesOff(); 
					System.exit(0);
			  }
			  else if (b == playButton) {
				  if (controller.getMidiLoader() != null) {
					  controller.playNotes();
				  }
				  else {
					  playStarted();
					  e.setSource(loadButton);
					  actionPerformed(e);
					  controller.playNotes();
				  }
			  }
			  else if (b == pauseButton)
				  controller.pauseNotes();
			  else if (b == stopButton)
				  controller.stopNotes();
			  else if (b == repeatButton) {
				  if (controller.isRepeatOn()) {
					  repeatButton.setIcon(new ImageIcon(cl.getResource("icons/repeat.png")));
					  controller.setRepeatOn(false);
				  }
				  else {
					  repeatButton.setIcon(new ImageIcon(cl.getResource("icons/repeat_active.png")));
					  controller.setRepeatOn(true);
				  }
			  }
			  else if (b == loadButton) {
				  if (controller.isMidiLoaderOn()) {
					  loadButton.setIcon(new ImageIcon(cl.getResource("icons/load.png")));
					  controller.handleMidiLoader();
				  }
				  else {
					  loadButton.setIcon(new ImageIcon(cl.getResource("icons/load_active.png")));
					  controller.handleMidiLoader();
				  }
			  }
			  else if (b == settingsButton) {
				  if (controller.isSettingsOn()) {
					  settingsButton.setIcon(new ImageIcon(cl.getResource("icons/settings.png")));
					  controller.handleSettings();
				  }
				  else {
					  settingsButton.setIcon(new ImageIcon(cl.getResource("icons/settings_active.png")));
					  controller.handleSettings();
				  }
			  }
		  }
	}
	
	public void mousePressed(MouseEvent e) {
		Object comp = e.getSource();
		if (comp instanceof JButton && e.getButton() == MouseEvent.BUTTON1) {
			JButton b = (JButton)comp;
			if (b == onoffButton)
				onoffButton.setIcon(new ImageIcon(cl.getResource("icons/onoff_active.png")));
			else if (b == stopButton)
				stopButton.setIcon(new ImageIcon(cl.getResource("icons/stop_active.png")));
			else if (b == pauseButton)
				pauseButton.setIcon(new ImageIcon(cl.getResource("icons/pause_active.png")));
		}
	}
	
	public void mouseReleased(MouseEvent e) {
		Object comp = e.getSource();
		if (comp instanceof JButton && e.getButton() == MouseEvent.BUTTON1) {
			JButton b = (JButton)comp;
			if (b == onoffButton)
				onoffButton.setIcon(new ImageIcon(cl.getResource("icons/onoff.png")));
			else if (b == stopButton)
				stopButton.setIcon(new ImageIcon(cl.getResource("icons/stop.png")));
			else if (b == pauseButton)
				pauseButton.setIcon(new ImageIcon(cl.getResource("icons/pause.png")));
		}
	}

	public void mouseClicked(MouseEvent e) {}
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
}
