/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.control;

import java.awt.Dimension;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.LinkedList;
import javax.media.opengl.GL;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.ShortMessage;
import javax.swing.JDialog;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import openpiano.OpenPiano;
import openpiano.gl.Camera;
import openpiano.gl.GLCore;
import openpiano.midi.MidiLoader;
import openpiano.midi.MidiLoaderGUI;
import openpiano.midi.MidiTrack;
import openpiano.notes.Notes;
import openpiano.piano.Piano;
import openpiano.settings.Constants;
import openpiano.settings.MidiSettings;
import openpiano.settings.Settings;
import openpiano.settings.SettingsGUI;

/**
 * A class managing the main application
 *
 * @author  Sascha Westermann
 * @version 1.0, 07/30/08
 */

public class Controller {
	private Boolean adjustSlider = false;
	private Camera camera;
	private ControlPanel controlPanel;
	private long elapsedTicks;
	private GL gl;
	private GLCore glCore;
	private int midiKeyboardChannel = 0;
	private MidiLoader midiLoader; 
	private JDialog midiLoaderDialog, settingsDialog;
	private MidiLoaderGUI midiLoaderGUI;
	private JScrollPane midiLoaderScroll;
	private Notes notes;
	private OpenPiano openPiano;
	private Piano piano;
	private Boolean playMode = false;
	private Boolean repeat = false;
	private Settings settings;
	private SettingsGUI settingsGUI;
	private Boolean slide = false;
	private SlidePanel slidePanel;
	
	/**
	 * Constructs a new controller class which constructs all other parts of the application
	 * 
	 * @param openPiano the reference to the main application
	 */
	public Controller(OpenPiano openPiano) {
		this.openPiano = openPiano;
		settings = new Settings(this);
		settingsGUI = new SettingsGUI(this, settings);
		glCore = new GLCore(this);
		gl = glCore.getGL();
		camera = new Camera(settings);
		notes = new Notes(this, settings);
	}
	
	/**
	 * Creates and shows the settings dialog
	 * 
	 * @param refresh <tt>true</tt> if settings dialog should only be refreshed
	 */
	public void createSettingsDialog(Boolean refresh) {
		if (refresh) {
			if (settingsDialog != null && settingsDialog.isVisible())
				settingsDialog.setVisible(false);
			settingsGUI = new SettingsGUI(this, settings);
		}
		settingsDialog = new JDialog(openPiano, "Settings", false);
		settingsDialog.add(settingsGUI);
		settingsDialog.pack();
		settingsDialog.setLocation(settings.getWindowPosX(), settings.getWindowPosY());
		settingsDialog.addWindowListener(new WindowListener() {
        	public void windowActivated(WindowEvent e) {}

        	public void windowClosed(WindowEvent e) {}
        	public void windowClosing(WindowEvent e) {
        		controlPanel.settingsGUIClosed();
        	}
        	public void windowDeactivated(WindowEvent e) {}
        	public void windowDeiconified(WindowEvent e) {}
        	public void windowIconified(WindowEvent e) {}
        	public void windowOpened(WindowEvent e) {}

        });
		
		if (refresh)
			settingsDialog.setVisible(true);
	}
	
	/**
	 * Returns the camera
	 * 
	 * @return the camera
	 */
	public Camera getCamera() { return camera; }
	
	/**
	 * Returns the settings of the current MIDI file
	 * 
	 * @return the settings of the current MIDI file
	 */
	public MidiSettings getCurrentMidiSettings() {
    	return settings.getMidiSettings(midiLoader.isFileKnown());
    }
	
	/**
	 * Returns the current GL context
	 * 
	 * @return the current GL context
	 */
	public GL getGL() { return gl; }
	
	/**
	 * Returns the GL core
	 * 
	 * @return the GL core
	 */
	public GLCore getGLCore() { return glCore; }
	

	/**
	 * Returns the file loader for MIDI files
	 * @return the file loader for MIDI files
	 */
	public MidiLoader getMidiLoader() { return midiLoader; };

	/**
	 * Returns the list of MIDI tracks
	 * 
	 * @return the list of MIDI tracks
	 */
	public LinkedList<MidiTrack> getMidiTracks() { return midiLoader.getMidiTracks(); }
	
	/**
	 * Returns the notes
	 * 
	 * @return the notes
	 */
	public Notes getNotes() { return notes; }
	
	/**
	 * Returns the piano
	 * 
	 * @return the piano
	 */
	public Piano getPiano() { return piano; }
	
	/**
	 * Returns the main application
	 * 
	 * @return the main application
	 */
	public OpenPiano getOpenPiano() { return openPiano; }

	/**
	 * Returns the settings
	 * 
	 * @return the settings
	 */
	public Settings getSettings() { return settings; }
	
	/**
	 * Returns the settings dialog
	 * 
	 * @return the settings dialog
	 */
	public JDialog getSettingsDialog() { return settingsDialog; }
	
	/**
	 * Calculates a GL position to a slider position
	 * 
	 * @param glPos the GL position to calculate
	 * @return the slider position of the GL position
	 */
	public int glToSlider(float glPos) { return tickToSlider(glToTick(glPos)); }
	
	/**
	 * Calculates a GL position to a MIDI tick
	 * 
	 * @param glPos the GL position to calculate
	 * @return the tick of the GL position
	 */
	public long glToTick(float glPos) {
		return Math.round( Constants.VIEWBEAT * midiLoader.getTicksPerBeat() * glPos / (float)Constants.VIEWAREA );
	}

	/**
	 * Handles the appearance of the MidiLoader dialog
	 */
	public void handleMidiLoader() {
		if (midiLoaderDialog != null) {
			midiLoaderDialog.setLocation(settings.getWindowPosX(), settings.getWindowPosY());
			if (midiLoaderDialog.isVisible())
				midiLoaderDialog.setVisible(false);
			else
				midiLoaderDialog.setVisible(true);
		}
		else {
			midiLoaderGUI = new MidiLoaderGUI(this, openPiano, slidePanel, settings);
			if (midiLoader != null)
				resetMidiLoaderDialog();
			else {
				controlPanel.midiLoaderGUIClosed();
				controlPanel.playStopped();
			}
		}
	}
	
	
	/**
	 * Handles the appearance of the settings dialog
	 */
	public void handleSettings() {
		if (settingsDialog != null) {
			settingsDialog.setLocation(settings.getWindowPosX(), settings.getWindowPosY());
			if (settingsDialog.isVisible())
				settingsDialog.setVisible(false);
			else {
				settingsDialog.setVisible(true);
			}
		}
		else { 
			createSettingsDialog(false);
			settingsDialog.setVisible(true);
		}
	}
	
	/**
	 * Initializes notes previously loaded
	 */
	private void initNotes() {
		slidePanel.resetPosSlider();
		Thread midiTimerThread = new Thread(midiLoader.getMidiTimer());
		midiTimerThread.start();
		midiLoader.getMidiTimer().initTimer();
		elapsedTicks = midiLoader.getMidiTimer().getElapsedTicks();
		setVolume();
	}
	
	/**
	 * Creates the piano
	 * 
	 * @param gl the relevant GL context
	 */
	public void initPiano(GL gl) { piano = new Piano(this, gl); }

	/**
	 * Returns <tt>true</tt> if the audio function is selected for a specified track
	 * 
	 * @param track the relevant track
	 * @return <tt>true</tt> if the audio function is selected
	 */
    public Boolean isAudioOn(int track) {
		return settings.getMidiSettings().get(midiLoader.isFileKnown()).getMidiTrackSettings().get(track).isAudioOn();
    }
	
    /**
     * Returns <tt>true</tt> if the clock is stopped
     * 
     * @return <tt>true</tt> if the clock is stopped
     */
	public Boolean isBpmStopped() {
    	if (midiLoader.getMidiTimer().getBeatsPerMinuteFactor() == 0.0f)
    		return true;
    	else
    		return false;
    }
	
	/**
	 * Returns <tt>true</tt> if the MidiLoader dialog is visible
	 * 
	 * @return <tt>true</tt> if the MidiLoader dialog is visible
	 */
	public Boolean isMidiLoaderOn() {
		if (midiLoaderDialog != null)
			return midiLoaderDialog.isVisible();
		else
			return false;
	}
	
	/**
	 * Returns <tt>true</tt> if the play button is activated
	 * 
	 * @return <tt>true</tt> if the play button is activated
	 */
	public Boolean isPlayModeSet() { return playMode; }
	
	/**
	 * Returns <tt>true</tt> if the repeat button is activated
	 * @return <tt>true</tt> if the repeat button is activated
	 */
	public Boolean isRepeatOn() { return repeat; }
	
	/**
	 * Returns <tt>true</tt> if the settings dialog is visible
	 * @return <tt>true</tt> if the settings dialog is visible
	 */
	public Boolean isSettingsOn() {
		if (settingsDialog != null)
			return settingsDialog.isVisible();
		else
			return false;
	}
	
	/**
	 * Returns <tt>true</tt> if the slider is currently in use
	 * 
	 * @return <tt>true</tt> if the slider is currently in use
	 */
	public Boolean isSliding() { return slide; }
	
	/**
	 * Returns <tt>true</tt> if the visual function is selected for a specified track
	 * 
	 * @param track the relevant track
	 * @return <tt>true</tt> if the visual function is selected
	 */
	public Boolean isVisualOn(int track) {
		return settings.getMidiSettings().get(midiLoader.isFileKnown()).getMidiTrackSettings().get(track).isVisualOn();
    }
	
	/**
	 * Loads new notes according to a MidiLoader object
	 * 
	 * @param newMidiLoader the MidiLoader object
	 */
	public void loadNotes(MidiLoader newMidiLoader) {
		Boolean play = false;
		if (playMode)
			play = true;
		
		if (midiLoader != null) {
			stopNotes();
		}
		midiLoader = newMidiLoader;
		notes.setNotes(midiLoader);
		initNotes();
		if (play)
			playNotes();
	}
	
	/**
	 * Stops or advances notes
	 */
	public void pauseNotes() {
		playMode = false;
		controlPanel.playStopped();
		if (midiLoader != null) {
			if (!midiLoader.getMidiTimer().isRunning()) {
				// advance a half beat
				slidePanel.getPositionSlider().setValue(slidePanel.getPositionSlider().getValue()+Math.round(tickToSlider(midiLoader.getTicksPerBeat())/2.0f));
				slideTo(slidePanel.getPositionSlider().getValue());
			}
			else {
				midiLoader.getMidiTimer().stopTimer();
				if (settings.getMidiOutput() != null)
					settings.getMidiOutput().allNotesOff();
				piano.releaseAllKeys();
			}
		}
	}
	
	/**
	 * Plays notes, starts from the beginning if end reached
	 */
	public void playNotes() {
		//TODO, wird schon vorher abgefangen
		if (midiLoader != null) {
			if (midiLoader.getMidiTimer().getElapsedTicks() == 
				midiLoader.getEndTick() + Constants.VIEWBEAT * midiLoader.getTicksPerBeat())
					stopNotes();
			if (!midiLoader.getMidiTimer().isRunning()) {
				playMode = true;
				controlPanel.playStarted();
				midiLoader.getMidiTimer().startTimer();
			}
/*			else {
				playMode = false;
				controlPanel.playStopped();
				pauseNotes();
			} */
		}
	}
	
	/**
	 * Resets the control messages of a MIDI file to the current tick
	 */
	public void resetControlMessages() {
		for (int i=0; i<getMidiTracks().size(); i++) {
			ShortMessage resetController = new ShortMessage();
			int channel = getMidiTracks().get(i).getChannel();
			try {
				resetController.setMessage( ShortMessage.CONTROL_CHANGE, channel, 121, 0 );
			} catch (Exception ex) {}
			if (settings.getMidiOutput() != null)
				settings.getMidiOutput().getReceiver().send(resetController, -1);
			getMidiTracks().get(i).setControlPos(0);
		}
	}
	
	/**
	 *  Resets the MidiLoader Dialog to rearrange items
	 */
	public void resetMidiLoaderDialog() {
		if (midiLoaderDialog != null)
			midiLoaderDialog.setVisible(false);

		midiLoaderDialog = new JDialog(openPiano, "MIDI Loader", false);
		midiLoaderScroll = new JScrollPane(midiLoaderGUI);
		midiLoaderDialog.add(midiLoaderScroll);
		midiLoaderDialog.addWindowListener(new WindowListener() {
			public void windowActivated(WindowEvent e) {}
		
			public void windowClosed(WindowEvent e) {}
			public void windowClosing(WindowEvent e) {
				controlPanel.midiLoaderGUIClosed();
			}
			public void windowDeactivated(WindowEvent e) {}
			public void windowDeiconified(WindowEvent e) {}
			public void windowIconified(WindowEvent e) {}
			public void windowOpened(WindowEvent e) {}
		});

		midiLoaderScroll.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
		midiLoaderDialog.pack();
		int height = midiLoaderScroll.getSize().height;
		int width = midiLoaderScroll.getSize().width;
		if (height < 300) {
	        midiLoaderScroll.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
	        midiLoaderDialog.pack();
			midiLoaderScroll.setPreferredSize(new Dimension(midiLoaderScroll.getSize().width, height));
		}
		else
			midiLoaderScroll.setPreferredSize(new Dimension(width, Constants.midiLoaderMaxHeight));
			
        midiLoaderDialog.pack();
		midiLoaderDialog.setLocation(settings.getWindowPosX(), settings.getWindowPosY());

		midiLoaderDialog.setVisible(true);
	}
	
	/**
	 * Resets the all relevant MIDI messages of a MIDI file to the current tick
	 */
	public void resetMidiMessages() {
		resetControlMessages();
		setInstruments();
		setVolume();
	}
	
	
	/**
	 * Resumes playing if the play button is active
	 */
	public void resumePlaying() {
		if (playMode)
			midiLoader.getMidiTimer().startTimer();
	}
	
	/**
	 * Handles and interprets a MIDI command
	 * 
	 * @param midiMessage the MIDI command to handle
	 */
	public void send(ShortMessage midiMessage) {
		if (midiMessage.getCommand() == ShortMessage.NOTE_ON || midiMessage.getCommand() == ShortMessage.NOTE_OFF) {
			int key = midiMessage.getData1()-21;
			int velocity = midiMessage.getData2();
			
			if (notes != null)
				notes.sendKey(key, velocity);
			
			if (piano != null)
				piano.sendKey(key, velocity, false);
		}
		
		// set volume
		if (midiMessage.getData1() == 7) {
			if (midiLoader != null)
				midiLoaderGUI.setVolume(midiKeyboardChannel, midiMessage.getData2());
		}
		// set instrument
		else if (midiMessage.getData1() == ShortMessage.PROGRAM_CHANGE) {
			if (midiLoader != null)
				midiLoaderGUI.setInstrument(midiKeyboardChannel, midiMessage.getData2());
		}

		try {
			midiMessage.setMessage(midiMessage.getCommand(), midiKeyboardChannel, midiMessage.getData1(), midiMessage.getData2());
		} catch (Exception ex) {}
		
		if (settings.getMidiOutput() != null)
			settings.getMidiOutput().send(midiMessage);
	}

	/**
	 * Sets the status of the audio function for specified track
	 * 
	 * @param track the relevant track
	 * @param audio <tt>true</tt> if the audio function should be enabled
	 */
	public void setAudio(int track, Boolean audio) {
		settings.getMidiSettings().get(midiLoader.isFileKnown()).getMidiTrackSettings().get(track).setAudioOn(audio);
		//TODO: ???
//		if (b && settings.isAwaitNotesOn())
//			midiTimer.updateBpm(settings.getMidiSettings().get(midiLoader.isFileKnown()).getBeatsPerMinuteValue()/100.0f);
	}
	
	/**
	 * Sets the status of the await notes function
	 * 
	 * @param awaitNotes <tt>true</tt> if the await notes function should be enabled
	 */
	public void setAwaitNote(Boolean awaitNotes) {
		settings.getMidiSettings().get(midiLoader.isFileKnown()).setAwaitNotesOn(awaitNotes);
		if (!awaitNotes)
			midiLoader.getMidiTimer().updateBpm(settings.getMidiSettings().get(midiLoader.isFileKnown()).getBeatsPerMinuteFactor()/100.0f);
	}

	/**
	 * Sets the control panel
	 * @param controlPanel the control panel
	 */
	public void registerControlPanel(ControlPanel controlPanel) {
		this.controlPanel = controlPanel;
	}
	
	/**
	 * Registers the slider panel
	 * @param slidePanel the slider panel
	 */
	public void registerSlidePanel(SlidePanel slidePanel) {
		this.slidePanel = slidePanel;
	}
	
	/**
	 * Sets the status of fast forward
	 * 
	 * @param fastForward <tt>true</tt> if fast forward should be enabled
	 */
	public void setFastForward(Boolean fastForward) {
		if (midiLoader != null) {
			if (fastForward) {
				if (settings.isAwaitNotesOn())
					setAwaitNote(false);

				//TODO: muss der timer sein, wenn er nicht spielt?
				midiLoader.getMidiTimer().seek(true, Constants.SeekFactor);
				if (!midiLoader.getMidiTimer().isRunning())
					midiLoader.getMidiTimer().startTimer();
			}
			else {
				if (midiLoaderGUI.isAwaitNotesChecked()) {
					setAwaitNote(true);
				}
				midiLoader.getMidiTimer().seek(false, Constants.SeekFactor);
				if (!playMode)
					midiLoader.getMidiTimer().stopTimer();
			}
		}
	}

	/**
	 * Sets a previously selected instrument for a specified track
	 * 
	 * @param track the relevant track
	 */
	public void setInstrument(int track) {
		int instrument = settings.getMidiSettings().get(midiLoader.isFileKnown()).getMidiTrackSettings().get(track).getInstrument();
		int channel = midiLoader.getMidiTracks().get(track).getChannel();
		setInstrument(track, channel, instrument);
	}
	
	/**
	 * Sets an instrument for a specified track and its channel
	 * 
	 * @param track the relevant track
	 * @param channel the corresponding channel
	 * @param instrument the instrument to set
	 */
	public void setInstrument(int track, int channel, int instrument) {
		//TODO: neu setzen nötig/sinnvoll?
		settings.getMidiSettings().get(midiLoader.isFileKnown()).getMidiTrackSettings().get(track).setInstrument(instrument);
		if (instrument != 128) {
			ShortMessage shortMessage = new ShortMessage();
			try {
				shortMessage.setMessage(ShortMessage.PROGRAM_CHANGE, channel, instrument, 0);
				if (settings.getMidiOutput() != null)
					settings.getMidiOutput().getReceiver().send(shortMessage, -1);
			} catch (InvalidMidiDataException e) {}
		}
	}
	
	/**
	 * Sets all previously selected instruments
	 */
	//TODO
	public void setInstruments() {
		int tracks = settings.getMidiSettings().get(midiLoader.isFileKnown()).getMidiTrackSettings().size();
		for (int i=0; i<tracks; i++) {
			int instrument = settings.getMidiSettings().get(midiLoader.isFileKnown()).getMidiTrackSettings().get(i).getInstrument();
			int channel = midiLoader.getMidiTracks().get(i).getChannel();
			setInstrument(i, channel, instrument);
		}
	}

	/**
	 * Sets the status of the mark keys function
	 * @param markKey <tt>true</tt> if the mark keys function should be enabled
	 */
	//TODO
	public void setMarkKey(Boolean markKey) {
		settings.getMidiSettings().get(midiLoader.isFileKnown()).setMarkKeysOn(markKey);
	}
	
	/**
	 * Sets the channel the MIDI keyboard will play on
	 * @param channel the channel to play on
	 */
	public void setMidiKeyboard(int channel) {
		midiKeyboardChannel = channel;
	}

	/**
	 * Sets the status of the repeat button
	 * @param repeat <tt>true</tt> if repeat button is active
	 */
	public void setRepeatOn(Boolean repeat) { this.repeat = repeat; }
	
	/**
	 * Sets the status of rewind
	 * 
	 * @param rewind <tt>true</tt> if rewind should be enabled
	 */
	public void setRewind(Boolean rewind) {
		setRewind(rewind, playMode);
	}
	
	/**
	 * Sets the status of rewind
	 * 
	 * @param rewind <tt>true</tt> if rewind should be enabled
	 * @param oldPlayMode <tt>true</tt> if the play button had been activated previously
	 */
	public void setRewind(Boolean rewind, Boolean oldPlayMode) {
		if (midiLoader != null) {
			if (rewind) {
				if (settings.isAwaitNotesOn())
					setAwaitNote(false);
				playMode = false;
				midiLoader.getMidiTimer().seek(true, -Constants.SeekFactor);
				//TODO: timer nötig?
				if (midiLoader.getMidiTimer().isRunning()) {
					if (settings.getMidiOutput() != null)
						settings.getMidiOutput().allNotesOff();
					piano.releaseAllKeys();
				}
				else
					midiLoader.getMidiTimer().startTimer();
			}
			else {
				if (midiLoaderGUI.isAwaitNotesChecked()) {
					setAwaitNote(true);
				}
				playMode = oldPlayMode;
				if (!playMode)
					midiLoader.getMidiTimer().stopTimer();
				midiLoader.getMidiTimer().seek(false, -Constants.SeekFactor);
			}
		}
	}

	/**
	 * Sets the tempo of a MIDI file
	 * 
	 * @param value the tempo to set
	 */
	public void setTempo(int value) {
		float factor = value / 100.0f;
		settings.getMidiSettings().get(midiLoader.isFileKnown()).setBeatsPerMinuteFactor(value);
		if (!isBpmStopped())
			midiLoader.getMidiTimer().updateBpm(factor);
	}

	/**
	 * Sets the status of the visual function for specified track
	 * 
	 * @param track the relevant track
	 * @param visual <tt>true</tt> if the visual function should be enabled
	 */
	public void setVisual(int track, Boolean visual) {
		settings.getMidiSettings().get(midiLoader.isFileKnown()).getMidiTrackSettings().get(track).setVisualOn(visual);
	}
    
	/**
	 * Sets the status of the visual function for specified track
	 * 
	 * @param track the relevant track
	 * @param visual <tt>true</tt> if the visual function should be enabled
	 * @param color the color of the notes
	 */
	public void setVisual(int track, Boolean visual, int color) {
		settings.getMidiSettings().get(midiLoader.isFileKnown()).getMidiTrackSettings().get(track).setVisualOn(visual);
	}

	/**
	 * Sets the previously set volume of all tracks
	 */
    public void setVolume() {
		int tracks = settings.getMidiSettings().get(midiLoader.isFileKnown()).getMidiTrackSettings().size();
		for (int i=0; i<tracks; i++)
			setVolume(i);
	}
    
    /**
     * Sets a previously set volume to a track
     * 
     * @param track the corresponding track
     */
    public void setVolume(int track) {
		int volume = settings.getMidiSettings().get(midiLoader.isFileKnown()).getMidiTrackSettings().get(track).getVolume();
		setVolume(track, volume);
	}
    
    /**
     * Sets the volume of a track
     * 
     * @param track the corresponding track
     * @param volume the volume to set
     */
    public void setVolume(int track, int volume) {
		ShortMessage volMessage = new ShortMessage();
		int channel = midiLoader.getMidiTracks().get(track).getChannel();
		try {
			volMessage.setMessage( ShortMessage.CONTROL_CHANGE, channel, 7, volume);
		} catch (Exception ex) {}
		if (settings.getMidiOutput() != null)
			settings.getMidiOutput().getReceiver().send(volMessage, -1);
	}

    /**
	 * Calculates a slider position to a GL position
	 * 
	 * @param sliderPos the slider position to calculate
	 * @return the GL position of the slider position
	 */
    public float sliderToGL(int sliderPos) {
		return tickToGL(sliderToTick(sliderPos));
	}

    /**
	 * Calculates a slider position to a MIDI tick
	 * 
	 * @param sliderPos the slider position to calculate
	 * @return the MIDI tick of the slider position
	 */
    public long sliderToTick(int sliderPos) {
		return Math.round( ( sliderPos / (float)slidePanel.getPositionSlider().getMaximum() ) 
				* (midiLoader.getEndTick() + Constants.VIEWBEAT * midiLoader.getTicksPerBeat()));
	}

    /**
     * Slide to a position in the MIDI file
     * 
     * @param sliderPos the position on the slider to slide to
     */
    // TODO: name verwirrend
	public void slideTo(int sliderPos) {
		slide = true;
		if (midiLoader != null) {
			if (repeat) {
				if (sliderPos < slidePanel.getSliderStartMarker() || sliderPos > slidePanel.getSliderEndMarker()) {
					repeat = false;
					controlPanel.repeatAborted();
				}
			}
			
			//TODO: eventuell unnötig
			if (midiLoader.getMidiTimer().isRunning()) {
				piano.releaseAllKeys();
				if (settings.getMidiOutput() != null)
					settings.getMidiOutput().allNotesOff();
				midiLoader.getMidiTimer().stopTimer();
			}
	
			// schlecht
			notes.setTick(sliderToTick(sliderPos)-(Constants.VIEWBEAT * midiLoader.getTicksPerBeat()), true);
			midiLoader.getMidiTimer().setElapsedTicks(sliderToTick(sliderPos));
			slidePanel.setTime(midiLoader.getBpmList(), sliderToTick(sliderPos), midiLoader.getTicksPerBeat());
		}
		slide = false;
	}
    
    /**
     * Slide to a position in the MIDI file
     * 
     * @param slider the slider that holds the value to slide to
     */
    public void slideTo(JSlider slider) {
		if (midiLoader != null) {
			if (slider.getValueIsAdjusting()) {
				slide = true;
				adjustSlider = true;
				if (midiLoader.getMidiTimer().isRunning()) {
					piano.releaseAllKeys();
					if (settings.getMidiOutput() != null)
						settings.getMidiOutput().allNotesOff();
					midiLoader.getMidiTimer().stopTimer();
				}
		
				// schlecht
				notes.setTick(sliderToTick(slider.getValue())-(Constants.VIEWBEAT * midiLoader.getTicksPerBeat()), true);
				midiLoader.getMidiTimer().setElapsedTicks(sliderToTick(slider.getValue()));
				slidePanel.setTime(midiLoader.getBpmList(), sliderToTick(slider.getValue()), midiLoader.getTicksPerBeat());
			}
			else if (adjustSlider) {
				adjustSlider = false;
				if (repeat) {
					if (slider.getValue() < slidePanel.getSliderStartMarker() || slider.getValue() > slidePanel.getSliderEndMarker()) {
						repeat = false;
						controlPanel.repeatAborted();
					}
				}
				if (playMode)
					midiLoader.getMidiTimer().startTimer();
				slide = false;
			}
		}
	}

    /**
     * Starts the clock with the current beats per minute settings
     */
    public void startBpm() {
    	midiLoader.getMidiTimer().updateBpm(getCurrentMidiSettings().getBeatsPerMinuteFactor()/100.0f);
    }
    
    /**
     * Stops the clock
     */
    public void stopBpm() {
    	midiLoader.getMidiTimer().updateBpm(0.0f);
    }
    
    /**
     * Stops the notes
     */
    public void stopNotes() {
		playMode = false;
		controlPanel.playStopped();
		if (midiLoader != null) {
			if (midiLoader.getMidiTimer().isRunning()) {
				midiLoader.getMidiTimer().stopTimer();
				//TODO: gehört in den midiTimer als selbsterkennung rein
				midiLoader.getMidiTimer().updateBpm();
				if (settings.getMidiOutput() != null)
					settings.getMidiOutput().allNotesOff();
				piano.releaseAllKeys();
			}
			slideTo(slidePanel.getSliderStartMarker());
			slidePanel.getPositionSlider().setValue(slidePanel.getSliderStartMarker());
			resetMidiMessages();
		}
		slidePanel.getTimerLabel().setText("");
		midiLoader.getMidiTimer().updateBpm(settings.getMidiSettings().get(midiLoader.isFileKnown()).getBeatsPerMinuteFactor()/100.0f);
	}
    
	/**
	 * Calculates a MIDI tick to a GL position
	 * 
	 * @param tick the tick to calculate
	 * @return the GL position of the MIDI tick
	 */
    public float tickToGL(long tick) {
		return Constants.VIEWAREA * tick / (float)(Constants.VIEWBEAT * midiLoader.getTicksPerBeat());
	}
    
	/**
	 * Calculates a MIDI tick to a slider position
	 * 
	 * @param tick the tick to calculate
	 * @return the slider position of the MIDI tick
	 */
    public int tickToSlider(long tick) {
		return Math.round( ( tick / (float)(midiLoader.getEndTick() + Constants.VIEWBEAT * midiLoader.getTicksPerBeat()) )
				* slidePanel.getPositionSlider().getMaximum() );
	}

    /**
     * Updates MIDI control messages to a specified tick
     * 
     * @param tick the tick to update control messages to
     */
    public void updateControlMessages(long tick) {
		//TODO
		if (tick < elapsedTicks)
			resetMidiMessages();
		
		for (int i=0; i<getMidiTracks().size(); i++) {
			for (int j=getMidiTracks().get(i).getControlPos(); j<getMidiTracks().get(i).getMidiControls().size(); j++) {
				if (getMidiTracks().get(i).getMidiControls().get(j).getTick() <= tick) {
					MidiMessage message = getMidiTracks().get(i).getMidiControls().get(j).getMidiMessage();
					if (settings.getMidiOutput() != null)
						settings.getMidiOutput().getReceiver().send(message, -1);
					getMidiTracks().get(i).setControlPos(j+1);
					
					// reset controller
					if (((ShortMessage)message).getData1() == 121) {
						setInstrument(i);
						setVolume(i);
					}
				}
				else
					j = getMidiTracks().get(i).getMidiControls().size();
			}
		}
		elapsedTicks = tick;
	}
}