/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.control;

import java.util.LinkedList;
import openpiano.midi.Bpm;
import openpiano.notes.Notes;
import openpiano.settings.Constants;

/**
 * MidiTimer is the central clock of the application. It controls the speed of notes and sets their position
 *
 * @author  Sascha Westermann
 * @version 1.0, 07/30/08
 */

public class MidiTimer implements Runnable {
	private int bpmIndex = 0;
	private LinkedList<Bpm> bpmList;
	private Controller controller;
	private SlidePanel slidePanel;
	private Notes notes;
	private boolean running;
	private long stopTime, lastTime;
	private int tpb;
	private float viewTicks, tps, bps, bpm, factor, elapsedTicks;

    /**
     * Constructs a new MidiTimer based on conditions from a MIDI file
     *
     * @param controller the main controller of the application
     * @param slidePanel the panel containing the position slider
     * @param bpmList a list of timed speed-changing events according to the MIDI file
     * @param tpb the amount of ticks per beat according to the MIDI file
     * @param factor a speed factor for beats per minute
     */
	public MidiTimer(Controller controller, SlidePanel slidePanel, LinkedList<Bpm> bpmList, int tpb, float factor) {
		this.factor = factor;
		this.controller = controller;
		this.slidePanel = slidePanel;
		elapsedTicks = 0.0f;
		notes = controller.getNotes();
		this.bpmList = bpmList;
		bpm = (float)bpmList.get(0).getBpm();
		this.tpb = tpb;
		bps = (factor*bpm)/60.0f;
		tps = (float)tpb * bps;
		viewTicks = (float)(Constants.VIEWBEAT * tpb);
		running = false;
	}
	
    /**
     * Returns the current beats per minute factor
     *
     * @return the current beats per minute factor
     */
	public float getBeatsPerMinuteFactor() { return factor; }

    /**
     * Returns the elapsed ticks of the current MIDI file
     *
     * @return the elapsed ticks
     */
	public long getElapsedTicks() { return (long)elapsedTicks; }
		
    /**
     * Returns the ticks per beat defined in the MIDI file
     *
     * @return the ticks per beat
     */
	public long getTicksPerBeat() { return Math.round(tpb); }

    /**
     * Returns the ticks per second defined in the MIDI file
     *
     * @return the ticks per second
     */
	public long getTicksPerSecond() { return Math.round(tps); }

	
    /**
     * Initializes the MidiTimer 
     */
	public void initTimer() {
		lastTime = System.currentTimeMillis();
		stopTime = System.currentTimeMillis();
		elapsedTicks = 0.0f;
	}

    /**
     * Returns <tt>true</tt> if the MidiTimer thread is running
     *
     * @return <tt>true</tt> if the MidiTimer thread is running
     */
	public boolean isRunning() { return running; }

    /**
     * Updates the application according to the time passed by. This is the central clock dealing 
     * with timed events like moving blocks and the slider position
     */
	public void run() {
		while (true) {
			try {
				Thread.sleep(10);
			} catch (Exception e) {}
			if (running) {
				long currentTime = System.currentTimeMillis();

				// calculate ticks based on time elapsed since last update
				float currentTicks = (float)((currentTime - lastTime) / 1000.0f) * tps;
				
				for (int i=bpmIndex+1; i<bpmList.size(); i++) {
					if (bpmList.get(i).getTick() <= (elapsedTicks - viewTicks)) {
						bpmIndex++;
						bpm = (float)bpmList.get(i).getBpm();
						bps = (factor*bpm)/60.0f;
						tps = (float)tpb * bps;
						currentTicks = (float)((currentTime - lastTime) / 1000.0f) * tps;
					}
				}
				
				elapsedTicks += currentTicks;
				
				// handle repeat
				if (controller.isRepeatOn()) {
					if ( (elapsedTicks > controller.sliderToTick(slidePanel.getSliderEndMarker())) 
					  || (elapsedTicks < controller.sliderToTick(slidePanel.getSliderStartMarker())) ) {
						controller.stopNotes();
						controller.playNotes();
					}
				}
				
				// cleanup
				if (elapsedTicks < 0)
					elapsedTicks = 0;
				else if (elapsedTicks > (controller.getMidiLoader().getEndTick() + viewTicks)) {
					elapsedTicks = controller.getMidiLoader().getEndTick() + viewTicks;
				}
				
				slidePanel.setTime(bpmList, Math.round(elapsedTicks), tpb);

				notes.setTick(Math.round(elapsedTicks-viewTicks), false);
				
				controller.updateControlMessages(Math.round(elapsedTicks - viewTicks));
				slidePanel.getPositionSlider().setValue(Math.round(slidePanel.getPositionSlider().getMaximum() * elapsedTicks / (controller.getMidiLoader().getEndTick() + viewTicks)));

				lastTime = currentTime;
			}
		}
	}
	
    /**
     * Sets the seek mode and speed. Used for rewind and fast forward
     *
     * @param seek <tt>true</tt> if seeking, <tt>false</tt> if resume
     * @param factor speed-factor of seeking
     */
	public void seek(boolean seek, float factor) {
		if (seek)
			this.factor*=factor;
		else
			this.factor/=factor;
		
		updateBpm();
	}

    /**
     * Sets the actual position in the MIDI file
     *
     * @param ticks current tick position
     */
	public void setElapsedTicks(float ticks) { 
		if (ticks < elapsedTicks) {
			updateBpm();
		}
		elapsedTicks = ticks;
	}

	/**
     * Starts the MidiTimer
     */
	public void startTimer() {
		if (!running) {
			long stopDuration = System.currentTimeMillis() - stopTime;
			lastTime+=stopDuration;
			running = true;
		}
	};

	/**
     * Stops the MidiTimer
     */
	public void stopTimer() {
		if (running) {
			running = false;
			stopTime = System.currentTimeMillis();
		}
	};
	
	/**
     * Updates beats per minutes according to the position
     */
	public void updateBpm() {
		bpm = bpmList.get(bpmIndex).getBpm();
		bps = (factor*bpm)/60.0f;
		tps = (float)tpb * bps;
		bpmIndex = 0;
	}
	
	/**
     * Updates beats per minutes according to the position based on a given factor
     * 
     * @param factor speed-factor
     */
	public void updateBpm(float factor) {
		this.factor = factor;
		bpm = bpmList.get(bpmIndex).getBpm();
		bps = (this.factor*bpm)/60.0f;
		tps = (float)tpb * bps;
		bpmIndex = 0;
	}
}