/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.control;

import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.Color;
import java.awt.event.MouseEvent;
import javax.swing.JSlider;
import javax.swing.UIManager;
import javax.swing.plaf.metal.MetalLookAndFeel;
import javax.swing.plaf.metal.MetalSliderUI;

/**
 * A modified version of the MetalSliderUI
 * 
 * @author Sascha Westermann
 * @version 1.0, 07/30/08
 */

public class PositionSliderUI extends MetalSliderUI {
	private int startMarker = 0;
	private int endMarker = 100;

	/**
	 * Creates a new UI for a slider
	 */
	public PositionSliderUI() {
		super();
	}
	
	/**
	 * Paints the track on the slider
	 */
    public void paintTrack(Graphics g)  {
   		oceanPaintTrack(g);
    }
    
    /**
     * Sets a start marker according to a slider's value
     * 
     * @param slider the slider which value is used
     */
    public void setStartMarker(JSlider slider) {
    	float value = (float)slider.getValue();
    	float maximum = (float)slider.getMaximum();
    	startMarker = Math.round((value / maximum) * 100.0f);
    }
    
    /**
     * Sets a start marker according to a value
     * 
     * @param value the value to set the marker to
     * @param slider the relevant slider
     */
    public void setStartMarker(int value, JSlider slider) {
    	float maximum = (float)slider.getMaximum();
    	startMarker = Math.round(((float)value / maximum) * 100.0f);
    }

    /**
     * Calculates the middle of a slider thumb
     * 
     * @return the middle of a slider thumb
     */
    public int getMiddleOfThumb() {
        Rectangle paintRect = getPaintTrackRect();
        int middleOfThumb = thumbRect.x + (thumbRect.width / 2) - paintRect.x;
        return middleOfThumb;
    }
    
    /**
     * Sets an end marker according to a slider's value
     * 
     * @param slider the slider which value is used
     */
    public void setEndMarker(JSlider slider) {
    	float value = (float)slider.getValue();
    	float maximum = (float)slider.getMaximum();
    	endMarker = Math.round((value / maximum) * 100.0f);
    }
    	
    /**
     * Sets an end marker according to a value
     * 
     * @param value the value to set the marker to
     * @param slider the relevant slider
     */
    public void setEndMarker(int value, JSlider slider) {
    	float maximum = (float)slider.getMaximum();
    	endMarker = Math.round(((float)value / maximum) * 100.0f);
    }
    
    
    /**
     * Draws the markers
     * 
     * @param g the relevant graphics context
     * @param startMarkerPos the position of the start marker
     * @param endMarkerPos the position of the end marker
     */
    private void drawMarkers(Graphics g, int startMarkerPos, int endMarkerPos) {
    	// start marker
    	g.setColor(MetalLookAndFeel.getBlack());
        g.drawLine(startMarkerPos-3, -6, startMarkerPos-3, 11);
        g.drawLine(startMarkerPos-2, -6, startMarkerPos-2, 11);
        g.drawLine(startMarkerPos-1, -6, startMarkerPos-1, 11);
        
        g.drawLine(startMarkerPos-3, -6, startMarkerPos+2, -6);
        g.drawLine(startMarkerPos-3, -5, startMarkerPos+2, -5);
        g.drawLine(startMarkerPos-3, -4, startMarkerPos+2, -4);
        g.drawLine(startMarkerPos-3, 11, startMarkerPos+2, 11);
        g.drawLine(startMarkerPos-3, 10, startMarkerPos+2, 10);
        g.drawLine(startMarkerPos-3, 9, startMarkerPos+2, 9);
        
        // end marker
        g.drawLine(endMarkerPos+3, -6, endMarkerPos+3, 11);
        g.drawLine(endMarkerPos+2, -6, endMarkerPos+2, 11);
        g.drawLine(endMarkerPos+1, -6, endMarkerPos+1, 11);
        
        g.drawLine(endMarkerPos+3, -6, endMarkerPos-2, -6);
        g.drawLine(endMarkerPos+3, -5, endMarkerPos-2, -5);
        g.drawLine(endMarkerPos+3, -4, endMarkerPos-2, -4);
        
        g.drawLine(endMarkerPos+3, 11, endMarkerPos-2, 11);
        g.drawLine(endMarkerPos+3, 10, endMarkerPos-2, 10);
        g.drawLine(endMarkerPos+3, 9, endMarkerPos-2, 9);
        
        // fill room
        g.setColor(MetalLookAndFeel.getWhite());
        g.drawLine(startMarkerPos+1, 1, endMarkerPos-1, 1);
        g.drawLine(startMarkerPos+1, 2, endMarkerPos-1, 2);
        g.drawLine(startMarkerPos+1, 3, endMarkerPos-1, 3);
        g.drawLine(startMarkerPos+1, 4, endMarkerPos-1, 4);
    }
    
    private void oceanPaintTrack(Graphics g)  {
	    Color sliderAltTrackColor = (Color)UIManager.get(
	                                "Slider.altTrackColor");
	    
	    // Translate to the origin of the painting rectangle
	    Rectangle paintRect = getPaintTrackRect();
	    g.translate(paintRect.x, paintRect.y);
	
	    // Width and height of the painting rectangle.
	    int w = paintRect.width;
	    int h = paintRect.height;

		int middleOfThumb = thumbRect.x + (thumbRect.width / 2) -
                            paintRect.x;
        int fillMinX;
        int fillMaxX;
        
        int absoluteStartMarker = Math.round((w / 100.0f) * (float)startMarker);
        int absoluteEndMarker = Math.round((w / 100.0f) * (float)endMarker);
        
        g.setColor(MetalLookAndFeel.getControlDarkShadow());
        g.drawRect(0, 0, w, h-1);
        g.setColor(MetalLookAndFeel.getPrimaryControlShadow());
        g.drawLine(1, 1, w - 1, 1);

        fillMinX = absoluteStartMarker;
        fillMaxX = middleOfThumb;
        
    	drawMarkers(g, absoluteStartMarker, absoluteEndMarker);
        
        if (middleOfThumb < absoluteStartMarker) {
        	fillMinX = absoluteStartMarker;
        	fillMaxX = absoluteStartMarker;
        }
        else if (middleOfThumb > absoluteEndMarker)
        	fillMaxX = absoluteEndMarker;

        g.setColor(MetalLookAndFeel.getWhite());
        g.drawLine(fillMinX, 1, fillMaxX, 1);
        g.setColor(sliderAltTrackColor);
        g.drawLine(fillMinX, 2, fillMaxX, 2);
        g.setColor(MetalLookAndFeel.getControlShadow());
        g.drawLine(fillMinX, 3, fillMaxX, 3);
        g.setColor(MetalLookAndFeel.getPrimaryControlShadow());
        g.drawLine(fillMinX, 4, fillMaxX, 4);
     
        g.translate(-paintRect.x, -paintRect.y);
    }
    
    private Rectangle getPaintTrackRect() {
        int trackLeft = 0, trackRight = 0, trackTop = 0, trackBottom = 0;
        if (slider.getOrientation() == JSlider.HORIZONTAL) {
            trackBottom = (trackRect.height - 1) - getThumbOverhang();
            trackTop = trackBottom - (getTrackWidth() - 1);
            trackRight = trackRect.width - 1;
        }
        else {
	    if (true) {
	        trackLeft = (trackRect.width - getThumbOverhang()) - 
		                                         getTrackWidth();
		trackRight = (trackRect.width - getThumbOverhang()) - 1;
	    }
	    else {
	        trackLeft = getThumbOverhang();
		trackRight = getThumbOverhang() + getTrackWidth() - 1;
	    }
            trackBottom = trackRect.height - 1;
        }
        return new Rectangle(trackRect.x + trackLeft, trackRect.y + trackTop,
                             trackRight - trackLeft, trackBottom - trackTop);
    }
    
    protected  void handlePressTrack(MouseEvent e) {
      trackListener.mousePressed(e);
    }

    protected CustomTrackListener createTrackListener(JSlider slider) {
        return new CustomTrackListener();
    }

    protected class CustomTrackListener extends TrackListener {
    	TrackListener trackListener;
    	CustomTrackListener() {
    		super();
    		trackListener = new TrackListener();
    	}
        public void mouseReleased(MouseEvent e) {
        	if (e.getButton() != MouseEvent.BUTTON3)
        		trackListener.mouseReleased(e);
        }
        public void mousePressed(MouseEvent e) {
        	if (e.getButton() != MouseEvent.BUTTON3)
        		trackListener.mousePressed(e);
        }
    	
    }
}