/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.control;

import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.text.NumberFormat;
import java.util.LinkedList;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.MouseInputListener;

import openpiano.midi.Bpm;
import openpiano.settings.Constants;

/**
 * A panel to handle a slider
 * 
 * @author Sascha Westermann
 * @version 1.0, 07/30/08
 */

public class SlidePanel extends JPanel implements ActionListener, MouseInputListener, ChangeListener {
	private static final long serialVersionUID = -2883433681228589162L;
	private JButton startplayButton, rewindButton, fastforwardButton, endplayButton;
	private JSlider posSlider;
	private PositionSliderUI posSliderUI;
	private JLabel timerLabel;
	private JPanel timerPanel;
	private Controller controller;
	private ClassLoader cl;
	private Boolean startMarkerClicked = false;
	private Boolean endMarkerClicked = false;
	private Boolean rewindPlayModeState;
	private int sliderStartMarker, sliderEndMarker;

	/**
	 * Construct a new slider panel
	 * 
     * @param controller the main controller of the application
	 */
	public SlidePanel(Controller controller) {
		this.controller = controller;
		Border empty = new EmptyBorder(0, 0, 0, 0);
        cl = this.getClass().getClassLoader();

        startplayButton = new JButton();
        rewindButton = new JButton();
        fastforwardButton = new JButton();
        endplayButton = new JButton();

        startplayButton.setContentAreaFilled(false);
        rewindButton.setContentAreaFilled(false);
        fastforwardButton.setContentAreaFilled(false);
        endplayButton.setContentAreaFilled(false);
        
        startplayButton.setFocusable(false);
        rewindButton.setFocusable(false);
        fastforwardButton.setFocusable(false);
        endplayButton.setFocusable(false);

        startplayButton.setIcon(new ImageIcon(cl.getResource("icons/startplay.png")));
        rewindButton.setIcon(new ImageIcon(cl.getResource("icons/rewind.png")));
        fastforwardButton.setIcon(new ImageIcon(cl.getResource("icons/ff.png")));
        endplayButton.setIcon(new ImageIcon(cl.getResource("icons/endplay.png")));

        startplayButton.setBorder(empty);
        rewindButton.setBorder(empty);
        fastforwardButton.setBorder(empty);
        endplayButton.setBorder(empty);

        startplayButton.addActionListener(this);
        rewindButton.addActionListener(this);
        fastforwardButton.addActionListener(this);
        endplayButton.addActionListener(this);

        startplayButton.addMouseListener(this);
        rewindButton.addMouseListener(this);
        fastforwardButton.addMouseListener(this);
        endplayButton.addMouseListener(this);
        
        timerPanel = new JPanel();
        timerLabel = new JLabel();
        timerLabel.setFont(new Font("Sans", Font.BOLD, 20));
        // simple workaround to get a reasonable size
        timerLabel.setText("00:00");
        timerLabel.setPreferredSize(new Dimension(timerLabel.getPreferredSize().width+6, timerLabel.getPreferredSize().height+3));
        timerLabel.setText("");
        timerPanel.setLayout(new GridBagLayout());
        timerPanel.add(timerLabel, new GridBagConstraints());

        posSlider = new JSlider(0, 10000, 0);
        posSliderUI = new PositionSliderUI();
        posSlider.setUI(posSliderUI);
        posSlider.setEnabled(false);
        posSlider.addChangeListener(this);
        posSlider.addMouseListener(this);
        posSlider.addMouseMotionListener(this);
        resetPosSlider();
        
        setLayout(new GridBagLayout());
        GridBagConstraints gbc = new GridBagConstraints();
        
        gbc.ipady = 10;
        gbc.ipadx = 20;
        gbc.weightx = 1.0;
        gbc.weighty = 1.0;
        gbc.fill = GridBagConstraints.BOTH;
        
        gbc.gridwidth = 5;
        gbc.gridx = 0;
        gbc.gridy = 0;
        gbc.weightx = 0.0;
        gbc.anchor = GridBagConstraints.CENTER;
        add(posSlider, gbc);
        
        gbc.gridwidth = 1;
        gbc.gridy = 1;
        gbc.gridx = 0;
        gbc.anchor = GridBagConstraints.WEST;
        add(rewindButton, gbc);

        gbc.gridx++;
        add(startplayButton, gbc);

        gbc.anchor = GridBagConstraints.CENTER;
        gbc.gridx++;
        gbc.weightx = 1.0;
        add(timerPanel, gbc);

        gbc.anchor = GridBagConstraints.EAST;
        gbc.gridx++;
        gbc.weightx = 0.0;
        add(endplayButton, gbc);

        gbc.gridx++;
        add(fastforwardButton, gbc);
        
	}
	
	public void actionPerformed(ActionEvent e) {
		  Object comp = e.getSource();
		  if (comp instanceof JButton) {
			  JButton b = (JButton)comp;
			  if (b == startplayButton && posSlider.isEnabled())
				  setSliderStartMarker(posSlider.getValue());
			  else if (b == endplayButton && posSlider.isEnabled())
				  setSliderEndMarker(posSlider.getValue());
		  }
	}
	
	public void mousePressed(MouseEvent e) {
		Object comp = e.getSource();
		if (comp instanceof JButton) {
			JButton b = (JButton)comp;
		
			if (b == startplayButton) {
				startplayButton.setIcon(new ImageIcon(cl.getResource("icons/startplay_active.png")));
			}
			else if (b == rewindButton) {
				rewindButton.setIcon(new ImageIcon(cl.getResource("icons/rewind_active.png")));
				rewindPlayModeState = controller.isPlayModeSet();
				controller.setRewind(true);
			}
			else if (b == fastforwardButton) {
				fastforwardButton.setIcon(new ImageIcon(cl.getResource("icons/ff_active.png")));
				controller.setFastForward(true);
			}
			else if (b == endplayButton) {
				endplayButton.setIcon(new ImageIcon(cl.getResource("icons/endplay_active.png")));
			}
		}
		else if (comp instanceof JSlider) {
			if (e.getButton() == MouseEvent.BUTTON1 && posSlider.isEnabled()) {
				PositionSliderUI sliderUI = (PositionSliderUI)((JSlider)comp).getUI();
				int pos = sliderUI.valueForXPosition(e.getX());
				((JSlider)comp).setValue(pos);
				sliderUI.handlePressTrack(e);
			}
			else if (e.getButton() == MouseEvent.BUTTON3 && posSlider.isEnabled()) {
				PositionSliderUI sliderUI = (PositionSliderUI)((JSlider)comp).getUI();

				int pos = sliderUI.valueForXPosition(e.getX());
				int startMarkerDistance = Math.abs(sliderStartMarker - pos);
				int endMarkerDistance = Math.abs(sliderEndMarker - pos);
				float tolerance = (((JSlider)comp).getMaximum()/100.0f)*2.0f;
				int thumb = ((JSlider)comp).getValue();

				if (startMarkerDistance < tolerance)
					startMarkerClicked = true;
				else if (endMarkerDistance < tolerance)
					endMarkerClicked = true;
				else {
					if (thumb - pos < 0) {
				        if (sliderStartMarker >= thumb) {
				        	if (pos > sliderEndMarker)
				        		endMarkerClicked = true;
				        	else
				        		startMarkerClicked = true;
				        }
				        else
							endMarkerClicked = true;
					}
				    else {
				        if (sliderEndMarker < thumb) {
				        	if (pos <= sliderStartMarker)
				        		startMarkerClicked = true;
				        	else
				        		endMarkerClicked = true;
				        }
				        else
							startMarkerClicked = true;
				    }
					
					if (startMarkerClicked) {
						setSliderStartMarker(pos);
						controller.slideTo(pos);
					}
					else if (endMarkerClicked) {
						setSliderEndMarker(pos);
						controller.slideTo(pos);
					}
				}
			}
		}

	}

	public void mouseReleased(MouseEvent e) {
		Object comp = e.getSource();
		if (comp instanceof JButton) {
			JButton b = (JButton)comp;
			
			if (b == startplayButton)
				startplayButton.setIcon(new ImageIcon(cl.getResource("icons/startplay.png")));
			else if (b == rewindButton) {
        		controller.setRewind(false, rewindPlayModeState);
				rewindButton.setIcon(new ImageIcon(cl.getResource("icons/rewind.png")));
			}
			else if (b == fastforwardButton) {
        		controller.setFastForward(false);
				fastforwardButton.setIcon(new ImageIcon(cl.getResource("icons/ff.png")));
			}
			else if (b == endplayButton)
				endplayButton.setIcon(new ImageIcon(cl.getResource("icons/endplay.png")));
		}
		else if (comp instanceof JSlider) {
			if (e.getButton() == MouseEvent.BUTTON3) {
				controller.slideTo(posSlider.getValue());
				controller.resumePlaying();
				startMarkerClicked = false;
				endMarkerClicked = false;
			}
		}
	}

	public void mouseDragged(MouseEvent e) {
		Object comp = e.getSource();
		if (comp instanceof JSlider) {
			if (startMarkerClicked) {
				PositionSliderUI sliderUI = (PositionSliderUI)((JSlider)comp).getUI();
				int pos = sliderUI.valueForXPosition(e.getX());
				setSliderStartMarker(pos);
				controller.slideTo(pos);
			}
			else if (endMarkerClicked) {
				PositionSliderUI sliderUI = (PositionSliderUI)((JSlider)comp).getUI();
				int pos = sliderUI.valueForXPosition(e.getX());
				setSliderEndMarker(pos);
				controller.slideTo(pos);
			}
		}
	}
	public void mouseMoved(MouseEvent e) {}
	public void mouseClicked(MouseEvent e) {}
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
	
	public void stateChanged(ChangeEvent e) {
	  Object comp = e.getSource();
	  if (comp instanceof JSlider) {
		  controller.slideTo((JSlider)comp);
	  }
	}
	
	/**
	 * Resets the slider
	 */
	public void resetPosSlider() {
		setSliderStartMarker(0);
		setSliderEndMarker(posSlider.getMaximum());
		controller.slideTo(sliderStartMarker);
		posSlider.setValue(0);
		timerLabel.setText("");
	}
	
	/**
	 * Sets the position of the end marker of the slider
	 * 
	 * @param value the value of the position
	 */
	private void setSliderEndMarker(int value) {
		posSliderUI.setEndMarker(value, posSlider);
		sliderEndMarker = value;
		if (sliderEndMarker < sliderStartMarker) {
			posSliderUI.setStartMarker(sliderEndMarker, posSlider);
			sliderStartMarker = sliderEndMarker;
		}
		posSlider.repaint();
	}
	
	/**
	 * Sets the position of the start marker of the slider
	 * 
	 * @param value the value of the position
	 */
	private void setSliderStartMarker(int value) {
		posSliderUI.setStartMarker(value, posSlider);
		sliderStartMarker = value;
		if (sliderStartMarker > sliderEndMarker) {
			posSliderUI.setEndMarker(sliderStartMarker, posSlider);
			sliderEndMarker = sliderStartMarker;
		}
		posSlider.repaint();
	}
	
	/**
	 * Returns the position of the end marker on the slider
	 * 
	 * @return the position of the end marker
	 */
	public int getSliderEndMarker() { return sliderEndMarker; }

	/**
	 * Returns the position of the start marker on the slider
	 * 
	 * @return the position of the start marker
	 */
	public int getSliderStartMarker() {	return sliderStartMarker; }
	
	/**
	 * Returns the position slider
	 * 
	 * @return the position slider
	 */
	public JSlider getPositionSlider() { return posSlider; }

	/**
	 * Sets a tick as current tick
	 * 
	 * @param tick the tick to make current
	 */
	public void setTime(LinkedList<Bpm> bpmList, long tick, int tpb) {
    	if (tick < tpb*Constants.VIEWBEAT)
    		setTimeCountdown(tick, tpb);
    	else {
    		tick-= tpb*Constants.VIEWBEAT;
	    	Boolean allDone=false;
	    	float seconds = 0.0f;
	    	
	    	for (int i=1; i<bpmList.size(); i++) {
	    		float ticksPerSecond = (tpb * bpmList.get(i-1).getBpm() / 60.0f);
	    		if (tick < bpmList.get(i).getTick()) {
	    			seconds+= (tick - bpmList.get(i-1).getTick()) / ticksPerSecond;
	    			i = bpmList.size();
	    			allDone = true;
	    		}
	    		else
	    			seconds+= (bpmList.get(i).getTick() - bpmList.get(i-1).getTick()) / ticksPerSecond;
	    	}
	    	
	    	if (!allDone) {
	    		float ticksPerSecond = (tpb * bpmList.getLast().getBpm() / 60.0f);
	    		seconds+= (tick - bpmList.getLast().getTick()) / ticksPerSecond; 
	    	}
	
	    	int sec = Math.round(seconds) % 60;
	    	int min = (Math.round(seconds) - sec) / 60;
	    	
	    	NumberFormat n = NumberFormat.getIntegerInstance();
	    	n.setMinimumIntegerDigits(2);
	    	
	    	timerLabel.setText(n.format(min) + ":" + n.format(sec));
    	}
    }
	
	private void setTimeCountdown(long tick, int tpb) {
    	int allTicks = tpb*Constants.VIEWBEAT;
    	float distance = allTicks / 3.0f;
    	int wait = 3 - (int)Math.floor(tick / distance);
    	
    	timerLabel.setText(wait + "...");
    }
	
	/**
	 * Returns the timer label
	 * 
	 * @return the timer label
	 */
	public JLabel getTimerLabel() { return timerLabel; }
}
