/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.gl;

import openpiano.settings.Constants;
import openpiano.settings.Settings;

/**
 * Camera class to set up the user's point of view
 * 
 * @author Sascha Westermann
 * @version 1.0, 07/30/08
 */

public class Camera {
	private Settings settings;
	private float centerView, actualView;
	private float visible;

	/**
	 * Construct a camera ontroller controllerwith default settings
	 */
	public Camera(Settings settings) {
		this.settings = settings;
		setDestination(settings.getFirstKey(), settings.getLastKey() - settings.getFirstKey());
		actualView = centerView;
		setAngle(settings.getAngle());
		setZoomFactor(settings.getZoomFactor());
	}

	private void setDestination(int start, int center) {
		int centerKey = start + center/2;
		if (centerKey < 3)
			centerView = Constants.POS_START[centerKey];
		else
			centerView = (int)((centerKey-3)/12)*Constants.OCTAVELENGTH + Constants.POS_OCTAVE[(centerKey-3)%12];
	}

	/**
	 * Sets the actual x-position
	 * 
	 * @param x the actual x-position
	 */
	public void setXPosition(float x) {
		if (x < 0.0f)
			actualView = 0.0f;
		else if (x > 7 * Constants.OCTAVELENGTH + Constants.POS_END + Constants.KEYLENGTH)
			actualView = 7 * Constants.OCTAVELENGTH + Constants.POS_END + Constants.KEYLENGTH;
		else
			actualView = x;
	}

	/**
	 * Returns the actual x-position
	 * 
	 * @return the actual x-position
	 */
	public float getXPosition() {
		return actualView;
	}
	
	/**
	 * Sets the actual angle

	 * @param angle the actual angle
	 */
	public void setAngle(float angle) {
		float newAngle;
		
		if (angle < 0.0f)
			newAngle = 0.0f;
		else if (angle > 90.0f)
			newAngle = 90.0f;
		else
			newAngle = angle;

		//TODO
		double lineA = Math.sqrt((37.0f*37.0f)+(10.0f*10.0f));
		double beta = 90.0f - newAngle + (90.0f - Math.toDegrees(Math.atan(10.0f/37.0f)));
		double gamma = 30.0f + Math.toDegrees(Math.atan(10.0f/37.0f));
		
		if (beta + gamma > 180.0f) {
			visible = 3000.0f;
		}
		else {
			double alpha = 180 - beta - gamma;
			visible = (float)(lineA * Math.sin(Math.toRadians(gamma)) / Math.sin(Math.toRadians(alpha)));
		}
		
		settings.setAngle(newAngle);
	}
	
	/**
	 * Returns the last visible GLFloat
	 * 
	 * @return the last visible GLFloat
	 */
	//TODO
	public float getVisibleArea() {
		if ((visible/settings.getZoomFactor()) + 25.0f > 3000.0f)
			return 3000.0f;
		else
			return (visible/settings.getZoomFactor()) + 25.0f;
	}
	
	/**
	 * Sets the actual zoom factor
	 * 
	 * @param zoom
	 */
	public void setZoomFactor(float zoomFactor) {
		float newZoomFactor;
		if (zoomFactor < 0.3)
			newZoomFactor = 0.3f;
		else if (zoomFactor > 3.5)
			newZoomFactor = 3.5f;
		else
			newZoomFactor = zoomFactor;
		
		settings.setZoomFactor(newZoomFactor);
	}
}
