/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.gl;

import java.io.BufferedReader;
import java.util.LinkedList;
import java.util.StringTokenizer;
import javax.media.opengl.GL;

/**
 * Wavefront model loader class. Based on code examples of JautOGL by Evangelos Pournaras
 *
 * @author Sascha Westermann
 * @version 1.0, 07/30/08
 */
public class GLModel implements Cloneable {
    private LinkedList<Float[]> vertexsets;
    private LinkedList<Float[]> vertexsetsnorms;
    private LinkedList<Integer[]> faces;
    private LinkedList<Integer[]> facesnorms;
    private int objectlist;
    private float toppoint = 0.0f;
    private float bottompoint = 0.0f;
    private float leftpoint = 0.0f;
    private float rightpoint = 0.0f;
    private float farpoint = 0.0f;
    private float nearpoint = 0.0f;

    /**
     * Constructs a simple GLModel (vertices and faces)
     * 
     * @param file model file to load
	 * @param gl the relevant GL context
     */
    public GLModel(BufferedReader file, GL gl){
        vertexsets = new LinkedList<Float[]>();
        vertexsetsnorms = new LinkedList<Float[]>();
        faces = new LinkedList<Integer[]>();
        facesnorms = new LinkedList<Integer[]>();
        loadobject(file);
        opengldrawtolist(gl);
        cleanup();
    }

	private void cleanup(){
        vertexsets.clear();
        vertexsetsnorms.clear();
        faces.clear();
        facesnorms.clear();
    }

    private void loadobject(BufferedReader br) {
        try {
            boolean firstpass = true;
            String newline;
            while ( (newline = br.readLine()) != null ) {
            	if (newline.length() > 0) {
                    newline = newline.trim();
                    
                    if (newline.startsWith("v ")) {
                        Float coords[] = new Float[3];
                        newline = newline.substring(2, newline.length());
                        StringTokenizer st = new StringTokenizer(newline, " ");
                        for (int i = 0; st.hasMoreTokens(); i++)
                            coords[i] = Float.parseFloat(st.nextToken());

                        if (firstpass) {
                            rightpoint = coords[0];
                            leftpoint = coords[0];
                            toppoint = coords[1];
                            bottompoint = coords[1];
                            nearpoint = coords[2];
                            farpoint = coords[2];
                            firstpass = false;
                        }
                        else {
	                        if(coords[0] > rightpoint)
	                            rightpoint = coords[0];
	                        if(coords[0] < leftpoint)
	                            leftpoint = coords[0];
	                        if(coords[1] > toppoint)
	                            toppoint = coords[1];
	                        if(coords[1] < bottompoint)
	                            bottompoint = coords[1];
	                        if(coords[2] > nearpoint)
	                            nearpoint = coords[2];
	                        if(coords[2] < farpoint)
	                            farpoint = coords[2];
                        }

                        vertexsets.add(coords);
                    }
                    else if (newline.startsWith("vn ")) {
                        Float coords[] = new Float[3];
                        newline = newline.substring(3, newline.length());
                        StringTokenizer st = new StringTokenizer(newline, " ");
                        for(int i=0; st.hasMoreTokens(); i++)
                            coords[i] = Float.parseFloat(st.nextToken());

                        vertexsetsnorms.add(coords);
                    }
                    else
                    
                    if (newline.startsWith("f ")) {
                        newline = newline.substring(2, newline.length());
                        StringTokenizer st = new StringTokenizer(newline, " ");
                        int count = st.countTokens();
                        Integer f[] = new Integer[count];
                        Integer fn[] = new Integer[count];
                        
                        for (int i=0; i<count; i++) {
                            char chars[] = st.nextToken().toCharArray();
                            StringBuffer sb = new StringBuffer();
                            char lc = 'x';
                            for(int k = 0; k < chars.length; k++){
                                if(chars[k] == '/' && lc == '/')
                                    sb.append('0');
                                lc = chars[k];
                                sb.append(lc);
                            }

                            StringTokenizer st2 = new StringTokenizer(sb.toString(), "/");
                            int num = st2.countTokens();
                            f[i] = Integer.parseInt(st2.nextToken());
                            if(num > 1)
                            	st2.nextToken();

                            if(num > 2)
                                fn[i] = Integer.parseInt(st2.nextToken());
                            else
                                fn[i] = 0;
                        }

                        faces.add(f);
                        facesnorms.add(fn);
                    }
                }
            }
        } catch (Exception e) {}
    }

    private void opengldrawtolist(GL gl){
    	objectlist = gl.glGenLists(1);
		gl.glNewList(objectlist, GL.GL_COMPILE);

		for (int i=0;i<faces.size();i++) {
			Integer[] tempfaces = faces.get(i);
			Integer[] tempfacesnorms = facesnorms.get(i);
			
			int polytype;
			if (tempfaces.length == 3) {
				polytype = GL.GL_TRIANGLES;
			} else if (tempfaces.length == 4) {
				polytype = GL.GL_QUADS;
			} else {
				polytype = GL.GL_POLYGON;
			}
			gl.glBegin(polytype);
			
			for (int w=0;w<tempfaces.length;w++) {
				if (tempfacesnorms[w] != 0) {
					float normtempx = vertexsetsnorms.get(tempfacesnorms[w] - 1)[0];
					float normtempy = vertexsetsnorms.get(tempfacesnorms[w] - 1)[1];
					float normtempz = vertexsetsnorms.get(tempfacesnorms[w] - 1)[2];
					gl.glNormal3f(normtempx, normtempy, normtempz);
				}
				
				float tempx = vertexsets.get(tempfaces[w] - 1)[0];
				float tempy = vertexsets.get(tempfaces[w] - 1)[1];
				float tempz = vertexsets.get(tempfaces[w] - 1)[2];
				gl.glVertex3f(tempx,tempy,tempz);
			}

			gl.glEnd();
			
			
		}
		gl.glEndList();
	}
    
    /**
     * Draws the model
     * 
	 * @param gl the relevant GL context
     */
    public void draw(GL gl){
        gl.glCallList(objectlist);
    }
}