/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.midi;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.security.MessageDigest;
import java.util.LinkedList;
import javax.sound.midi.MetaMessage;
import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.Sequence;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Track;
import openpiano.control.Controller;
import openpiano.control.MidiTimer;
import openpiano.control.SlidePanel;
import openpiano.settings.MidiSettings;
import openpiano.settings.MidiTrackSettings;
import openpiano.settings.Settings;

/**
 * The MidiLoader class loads the MIDI events of a MIDI file. The MIDI formats 0 and 1 are supported 
 * 
 * @author Sascha Westermann
 * @version 1.0, 07/30/08
 */

public class MidiLoader {
	private LinkedList<Bpm> bpmList;
	private MidiTimer midiTimer;
	private Boolean failed = false;
	private int knownMidiFile;
	private int midiFormat;
	private LinkedList<MidiTrack> midiTracks;
	private Sequence sequence;
	private long ticks;
	private int tpb;

	/**
	 * Constructs a new list of MIDI events based on a MIDI files.
	 * If the MIDI file is already known, the pre-settings will be adapted
	 * 
	 * @param controller the main controller of the application
	 * @param slidePanel the panel containing the position slider
	 * @param settings a reference to the settings object
	 * @param midiFile the MIDI file to load
	 */
	public MidiLoader(Controller controller, SlidePanel slidePanel, Settings settings, File midiFile) {
		String md5Checksum = "";

		try	{
			sequence = MidiSystem.getSequence(midiFile);
			midiFormat = MidiSystem.getMidiFileFormat(midiFile).getType();
			md5Checksum = getMD5Checksum(midiFile);
		} catch (Exception e) { failed = true; }
		  
		if (!failed) {
			knownMidiFile = settings.lookupForFile(md5Checksum);
			
			Track[] tracks = sequence.getTracks();
			ticks = sequence.getTickLength();
			tpb = sequence.getResolution();
			bpmList = new LinkedList<Bpm>();
			
			if (midiFormat == 0)
				handleMidiFormat0(tracks);
			else if (midiFormat == 1)
				handleMidiFormat1(tracks);
			else
				failed = true;
			
			if (!failed && knownMidiFile == -1) {
				knownMidiFile = settings.addMidiSettings(new MidiSettings(md5Checksum, 0, 100, false, false));
				LinkedList<MidiTrackSettings> midiTrackSettings = settings.getMidiSettings(knownMidiFile).getMidiTrackSettings();
				for (MidiTrack m : midiTracks)
					midiTrackSettings.add(new MidiTrackSettings(m.getName(), m.getInstrument(), m.getVolume(), true, false));
			}

			float factor = settings.getMidiSettings().get(knownMidiFile).getBeatsPerMinuteFactor()/100.0f;
			midiTimer = new MidiTimer(controller, slidePanel, bpmList, tpb, factor);
		}
	}
	
	/**
	 * Deletes tracks without MIDI notes
	 */
	private void cleanupTracks() {
		// cleanup non-relevant channels, else sort elements
		for (int i=0; i<midiTracks.size(); i++) {
			if (midiTracks.get(i).getMidiNotes().size() == 0) {
				midiTracks.remove(i);
				i--;
			}
			else
				midiTracks.get(i).sortElements();
		}
		
		// correct sound volumes and instruments
		for (int i=0; i<midiTracks.size(); i++) {
			if (midiTracks.get(i).getVolume() == -1) {
				midiTracks.get(i).setVolume(63);
			}
			if (midiTracks.get(i).getInstrument() == 128) {
				if (midiTracks.get(i).getChannel() != 10)
					midiTracks.get(i).setInstrument(0);
			}
		}
		
	}
	
	/**
	 * Returns the MD5 checksum of the MIDI file
	 * 
	 * @param midiFile the MIDI file to use
	 * @return the MD5 checksum of the MIDI file
	 * @throws Exception
	 */
	private byte[] createMD5Checksum(File midiFile) throws Exception {
	    byte[] buffer = new byte[1024];

	    InputStream is =  new FileInputStream(midiFile);
    	MessageDigest md5 = MessageDigest.getInstance("MD5");

	    int i;
	    do {
	    	i = is.read(buffer);
	    	if (i > 0)
	    		md5.update(buffer, 0, i);
	    } while (i != -1);

	    is.close();
	    return md5.digest();
	}
	
	private MidiNote createNote(MidiMessage midiMessage, long startTick, Track track, int pos) {
		long endTick = startTick;
		
		for (int i=pos+1; i<track.size(); i++) {
			MidiEvent midiEvent = track.get(i);
			MidiMessage	message = midiEvent.getMessage();
			// next equal note must be of type noteOff, otherwise we
			// have to deal with a malformed midi file
			if (isEqualNote((ShortMessage)midiMessage, message)) {
				endTick = midiEvent.getTick();
				i = track.size();
			}
		}
		
		// make sure the note is complete
		if (startTick != endTick)
			return new MidiNote(startTick, endTick, midiMessage);
		else
			return null;
	}
	
	/**
	 * Returns the MIDI file's list of beats per minute events
	 *
	 * @return the list of beats per minute events
	 */
	public LinkedList<Bpm> getBpmList() { return bpmList; }
	
	/**
	 * Returns the last tick of the MIDI file
	 * 
	 * @return the last tick of the MIDI file
	 */
	public long getEndTick() { return ticks; }
	
	private String getMD5Checksum(File midiFile) throws Exception {
	    byte[] md5 = createMD5Checksum(midiFile);
	    String result = "";
	    for (int i=0; i < md5.length; i++)
	    	result += Integer.toString( ( md5[i] & 0xff ) + 0x100, 16).substring( 1 );
	    
	    return result;
	}
	
	/**
	 * Returns the list of MIDI events per MIDI track
	 * 
	 * @return the list of MIDI events
	 */
	public LinkedList<MidiTrack> getMidiTracks() { return midiTracks; }
	
	/**
	 * Returns the ticks per beat of the MIDI file
	 * 
	 * @return the ticks per beat
	 */
	public int getTicksPerBeat() { return tpb; }
	
	private void handleMidiFormat0(Track[] tracks) {
		// create 16 channels (midi rfc)
		midiTracks = new LinkedList<MidiTrack>();
		for (int i=0; i<16; i++) {
			midiTracks.add(new MidiTrack("Channel " + i, i));
		}
		
		// use channels to handle midi format 0
		for (Track t : tracks) {
			for (int i=0; i<t.size(); i++) {
				MidiEvent midiEvent = t.get(i);
				MidiMessage midiMessage = midiEvent.getMessage();
				if (isNoteOn(midiMessage)) {
					int channel = ((ShortMessage)midiMessage).getChannel();
					MidiNote midiNote = createNote(midiMessage, midiEvent.getTick(), t, i);
					if (midiNote != null)
						midiTracks.get(channel).addMidiNote(midiNote);
				}
				// always use latest instrument (handle multiple changes on one channel)
				else if (isInstrumentChange(midiMessage)) {
					setInstrument((ShortMessage)midiMessage, ((ShortMessage)midiMessage).getChannel());
				}
				else if (isTempoChange(midiMessage)) {
					setTempoChange((MetaMessage)midiMessage, midiEvent.getTick());
				}
				else if (isControlChange(midiMessage)) {
					setControlChange(midiEvent.getTick(), midiMessage, ((ShortMessage)midiMessage).getChannel());
				}
			}
		}
		
		cleanupTracks();
	}
	
	private void handleMidiFormat1(Track[] tracks) {
		midiTracks = new LinkedList<MidiTrack>();
		
		// use tracks (and split to channels) to handle midi format 1
		int trackPos = 0;
		for (Track t : tracks) {
			String trackName = null;
			int trackStart = trackPos;
			int[] translateChannels = new int[16];
			for (int i=0; i<16; i++)
				translateChannels[i] = -1;
			
			for (int i=0; i<t.size(); i++) {
				MidiEvent midiEvent = t.get(i);
				MidiMessage midiMessage = midiEvent.getMessage();
				if (isNoteOn(midiMessage)) {
					int channel = ((ShortMessage)midiMessage).getChannel();
					if (translateChannels[channel] == -1) {
						midiTracks.add(new MidiTrack(channel));
						translateChannels[channel] = trackPos;
						trackPos++;
					}
					MidiNote midiNote = createNote(midiMessage, midiEvent.getTick(), t, i);
					if (midiNote != null)
						midiTracks.get(translateChannels[channel]).addMidiNote(midiNote);
				}
				// always use latest instrument (workaround multiple changes on one channel)
				else if (isInstrumentChange(midiMessage)) {
					int channel = ((ShortMessage)midiMessage).getChannel();
					if (translateChannels[channel] == -1) {
						midiTracks.add(new MidiTrack(channel));
						translateChannels[channel] = trackPos;
						trackPos++;
					}
					setInstrument((ShortMessage)midiMessage, translateChannels[channel]);
				}
				else if (isTempoChange(midiMessage)) {
					setTempoChange((MetaMessage)midiMessage, midiEvent.getTick());
				}
				else if (isTrackNameChange(midiMessage)) {
					trackName = new String(((MetaMessage)midiMessage).getData());
				}
				else if (isControlChange(midiMessage)) {
					int channel = ((ShortMessage)midiMessage).getChannel();
					if (translateChannels[channel] == -1) {
						midiTracks.add(new MidiTrack(channel));
						translateChannels[channel] = trackPos;
						trackPos++;
					}
					setControlChange(midiEvent.getTick(), midiMessage, translateChannels[channel]);
				}
			}
			
			// name channels according to their source track
			for (int i=trackStart; i<trackPos; i++)
				midiTracks.get(i).setTrackName(trackName);
		}
		
		cleanupTracks();
	}
	
	/**
	 * Returns <tt>true</tt> if loading of a MIDI file failed
	 * 
	 * @return <tt>true</tt> if loading of a MIDI file failed
	 */
	public Boolean initFailed() { return failed; }
	
	private Boolean isControlChange(MidiMessage midiMessage) {
		if (midiMessage instanceof ShortMessage) {
			ShortMessage shortMessage = (ShortMessage)midiMessage;
			if (shortMessage.getCommand() == ShortMessage.CONTROL_CHANGE)
				return true;
			else
				return false;
		}
		else
			return false;
	}

	private Boolean isEqualNote(ShortMessage refShortMessage, MidiMessage midiMessage) {
		if (midiMessage instanceof ShortMessage) {
			ShortMessage shortMessage = (ShortMessage)midiMessage;
			if ( (refShortMessage.getData1() == shortMessage.getData1()) &&
				 (refShortMessage.getChannel() == shortMessage.getChannel()) &&
				 ( (refShortMessage.getCommand() == ShortMessage.NOTE_OFF) ||
				   (refShortMessage.getCommand() == ShortMessage.NOTE_ON) ) )
						return true;
			else
				return false;
		}
		else
			return false;
	}
	
	/**
	 * Returns <tt>true</tt> if the MIDI file is already known
	 * 
	 * @return <tt>true</tt> if the MIDI file is already known
	 */
	public int isFileKnown() { return knownMidiFile; }

	private Boolean isInstrumentChange(MidiMessage midiMessage) {
		if (midiMessage instanceof ShortMessage) {
			ShortMessage shortMessage = (ShortMessage)midiMessage;
			if (shortMessage.getCommand() == ShortMessage.PROGRAM_CHANGE)
				return true;
			else
				return false;
		}
		else
			return false;
	}

	private Boolean isNoteOn(MidiMessage midiMessage) {
		if (midiMessage instanceof ShortMessage) {
			ShortMessage shortMessage = (ShortMessage)midiMessage;
			if ( (shortMessage.getCommand() == ShortMessage.NOTE_ON) && (shortMessage.getData2() > 0) )
				return true;
			else
				return false;
		}
		else
			return false;
	};
	
	private Boolean isTempoChange(MidiMessage midiMessage) {
		if (midiMessage instanceof MetaMessage) {
			MetaMessage metaMessage = (MetaMessage)midiMessage;
			if (metaMessage.getType() == 0x51) { // set tempo message
				return true;
			}
			else
				return false;
		}
		else
			return false;
	}
	
	private Boolean isTrackNameChange(MidiMessage midiMessage) {
		 if (midiMessage instanceof MetaMessage) {
			 MetaMessage metaMessage = (MetaMessage)midiMessage;
			 if (metaMessage.getType() == 3)
				 return true;
			 else
				 return false;
		 }
		 else
			 return false;
	}
	
	private void setControlChange(long tick, MidiMessage midiMessage, int track) {
		// set first found volume for the track
		if (((ShortMessage)midiMessage).getData1() == 7) {
			if (midiTracks.get(track).getVolume() == -1)
				midiTracks.get(track).setVolume(((ShortMessage)midiMessage).getData2());
		}
		else
			midiTracks.get(track).addMidiControl(new MidiControl(midiMessage, tick));
	}
	
	private void setInstrument(ShortMessage shortMessage, int track) {
		int instrument = shortMessage.getData1();
		//set first found instrument for the track
		if (midiTracks.get(track).getInstrument() == 128)
			midiTracks.get(track).setInstrument(instrument);
	}
	
	//TODO: der aufwand nötig?
	private void setTempoChange(MetaMessage metaMessage, long startTick) {
		bpmList.add(new Bpm(metaMessage, startTick));
	}
	
	/**
	 * Returns the clock
	 * 
	 * @return the clock
	 */
	public MidiTimer getMidiTimer() { return midiTimer; }
}
