/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.midi;

import java.util.LinkedList;
import java.awt.Color;
import java.awt.Component;
import java.awt.ComponentOrientation;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.Collections;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.border.BevelBorder;
import openpiano.OpenPiano;
import openpiano.control.Controller;
import openpiano.control.SlidePanel;
import openpiano.settings.Constants;
import openpiano.settings.MidiTrackSettings;
import openpiano.settings.Settings;

/**
 * The MidiLoaderGUI class creates a graphical frontend for the MidiLoader 
 * 
 * @author Sascha Westermann
 * @version 1.0, 07/30/08
 */

public class MidiLoaderGUI extends JPanel {
	private static final long serialVersionUID = -101768350270394237L;
	private MidiLoader midiLoader;
	private Controller controller;
	private LinkedList<MidiTrackGUI> midiTracks;
	private JSpinner spinnerTempo;
	private MidiLoaderGUI midiLoaderGUI;
	private JCheckBox checkBoxAwaitNote;
	private JCheckBox checkBoxMarkKey;
	private Boolean initialStart = true;
	private SlidePanel slidePanel;
	private Settings settings;
	
	/**
	 * Constructs a new fronted for the MidiLoader
	 * 
     * @param controller the main controller of the application
	 * @param openPiano the reference to the application window
	 * @param slidePanel the panel containing the position slider
	 */
	public MidiLoaderGUI(Controller controller, OpenPiano openPiano, SlidePanel slidePanel, Settings settings) {
		this.controller = controller;
		this.slidePanel = slidePanel;
		this.settings = settings;
		midiLoaderGUI = this;
		setLayout(new GridBagLayout());
		setVisible(false);
		openFileDialog(openPiano);
	}
        
	private void openFileDialog(Component parent) {
		JFileChooser fileChooser = new JFileChooser();
		fileChooser.setCurrentDirectory(controller.getSettings().getFileDir());
		
        if (fileChooser.showOpenDialog(parent) == JFileChooser.APPROVE_OPTION) {
        	controller.getSettings().setFileDir(fileChooser.getCurrentDirectory());
    		midiLoader = new MidiLoader(controller, slidePanel, controller.getSettings(), fileChooser.getSelectedFile().getAbsoluteFile());
    		if (!midiLoader.initFailed()) {
    			removeAll();
	    		controller.loadNotes(midiLoader);
	    		
	    		GridBagConstraints gbc = new GridBagConstraints();
	            gbc.fill = GridBagConstraints.BOTH;
	            gbc.anchor = GridBagConstraints.WEST;
	            gbc.weightx = 0.0;
	            add(new FilePanel(fileChooser.getSelectedFile().getName()), gbc);
	            
	            gbc.anchor = GridBagConstraints.CENTER;
	            gbc.insets = new Insets(0, 70, 0, 70);
	            add(new JPanel(), gbc);
	            
	            gbc.insets = new Insets(0, 0, 0, 0);
	            gbc.anchor = GridBagConstraints.EAST;
	            gbc.gridwidth = GridBagConstraints.REMAINDER;
	            add(new OptionPanel(), gbc);
	            
	            addMidiTrackPanels(gbc);
	            
	            revalidate();
	            setVisible(true);
	
	            controller.resetMidiMessages();
	            if (!initialStart)
	            	controller.resetMidiLoaderDialog();
            	else
            		initialStart = false;
	            
	            slidePanel.getPositionSlider().setEnabled(true);
    		}
		}
		
	}
	
	/**
	 * Returns the a reasonable color for notes
	 * 
	 * @param the track to use the color for
	 * @return the next available color
	 */
	public int getColor(int track) {
		int[] colors = new int[midiTracks.size()];
		LinkedList<ColorOccurence> colorOccurence = new LinkedList<ColorOccurence>();

		for (int i=0; i<(Constants.COLORSTART.length/3); i++)
			colorOccurence.add(new ColorOccurence(i));
		
		for (int i=0; i<midiTracks.size(); i++) {
			colors[i] = midiTracks.get(i).getColor();
			if (colors[i] != -1)
				colorOccurence.get(colors[i]).addOccurence();
		}
		
		Collections.sort(colorOccurence);
		
		int index = 0;
		if (track > 0 && colors[track-1] == colorOccurence.get(index).getColor())
			index++;
		if (track < midiTracks.size()-1 && colors[track+1] == colorOccurence.get(index).getColor())
			index++;
		
		return colorOccurence.get(index).getColor();
	}
	

	private class ColorOccurence implements Comparable<ColorOccurence>{
		private int color;
		private int occurence = 0;
		
		public ColorOccurence(int color) {
			this.color = color;
		}
		
		public void addOccurence() { occurence++; }
		public int getOccurence() { return occurence; }
		public int getColor() { return color; }
		
		public int compareTo(ColorOccurence colorOccurence) {
			return (int)(this.occurence - colorOccurence.getOccurence());
		}
	}

	
	private void addMidiTrackPanels(GridBagConstraints gbc) {
		midiTracks = new LinkedList<MidiTrackGUI>();
        gbc.fill = GridBagConstraints.NONE;
        gbc.weighty = 0.0;
        gbc.weightx = 0.0;
        gbc.gridwidth = GridBagConstraints.REMAINDER;

        for (int i=0; i<midiLoader.getMidiTracks().size(); i++) {
        	midiTracks.add(new MidiTrackGUI(controller, settings, midiLoader.getMidiTracks().get(i), i, this));
            gbc.anchor = GridBagConstraints.NORTHWEST;
			add(midiTracks.getLast(), gbc);
        }
		
        gbc.weightx = 1.0;
        gbc.weighty = 1.0;
        gbc.fill = GridBagConstraints.BOTH;
        add(new JPanel(), gbc);

        setKeyboardToTrack(controller.getSettings().getMidiSettings(controller.getMidiLoader().isFileKnown()).getKeyboardTrack());
	}

	final class FilePanel extends JPanel {
		private static final long serialVersionUID = 9113542777459409375L;

		FilePanel(String fileString) {
			JButton fileOpenButton = new JButton("Open File");
			fileOpenButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					openFileDialog(midiLoaderGUI);
				}
			});
			
			JButton resetSettingsButton = new JButton("Reset Settings");
			resetSettingsButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					resetSettings();
				}
			});
			
			 JTextField fileText = new JTextField(30);
			 fileText.setText(fileString);
			
			 setLayout(new GridBagLayout());
			 GridBagConstraints gbc = new GridBagConstraints();
			 
			 gbc.anchor = GridBagConstraints.WEST;
			 gbc.weightx = 0.0;
			 gbc.insets = new Insets(5,5,5,5);
			 add(fileOpenButton, gbc);

			 add(resetSettingsButton, gbc);
			 
			 gbc.insets = new Insets(0,0,0,0);
			 gbc.gridwidth = GridBagConstraints.REMAINDER;
			 add(new JPanel(), gbc);
			 
			 gbc.weightx = 0.0;
			 gbc.insets = new Insets(0,5,5,0);
			 add(fileText, gbc);
		 }
	 }

	 final class OptionPanel extends JPanel {
		private static final long serialVersionUID = -1440764926067542392L;

		OptionPanel() {
			 checkBoxMarkKey = new JCheckBox("Mark Key");
			 checkBoxMarkKey.setSelected(controller.getSettings().getMidiSettings(controller.getMidiLoader().isFileKnown()).isMarkKeysOn());
			 checkBoxMarkKey.addItemListener(new ItemListener() {
				 public void itemStateChanged(ItemEvent e) {
					 if (checkBoxMarkKey.isSelected())
						 controller.setMarkKey(true);
					 else
						 controller.setMarkKey(false);
				 }
			 });
			
			checkBoxAwaitNote = new JCheckBox("Await Note");
			checkBoxAwaitNote.setSelected(controller.getSettings().getMidiSettings(controller.getMidiLoader().isFileKnown()).isAwaitNotesOn());
			checkBoxAwaitNote.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					if (checkBoxAwaitNote.isSelected())
						controller.setAwaitNote(true);
					else
						controller.setAwaitNote(false);
				}
			});
			
			int bpmvalue = controller.getSettings().getMidiSettings(controller.getMidiLoader().isFileKnown()).getBeatsPerMinuteFactor();
			SpinnerNumberModel model = new SpinnerNumberModel(new Integer(bpmvalue),
					new Integer(1), new Integer(200), new Integer(1)); 
			spinnerTempo = new JSpinner(model);
			spinnerTempo.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					controller.setTempo((Integer)spinnerTempo.getValue());
				}
			});
			 
			setComponentOrientation(ComponentOrientation.RIGHT_TO_LEFT);
			setLayout(new GridBagLayout());
			GridBagConstraints gbc = new GridBagConstraints();
			
			gbc.anchor = GridBagConstraints.EAST;
			gbc.weightx = 0.0;
			add(spinnerTempo, gbc);
			
			gbc.gridwidth = GridBagConstraints.REMAINDER;
			add(new JLabel("Beats per Minutes (in %): "), gbc);
			
			JPanel p = new JPanel(new FlowLayout());
			p.add(checkBoxAwaitNote, gbc);
			p.add(checkBoxMarkKey, gbc);
			add(p, gbc);
		 }
	 }
	
	/**
	 * Sets the track the MIDI keyboard plays
	 * 
	 * @param track the track the MIDI keyboard plays
	 */ 
	public void setKeyboardToTrack(int track) {
		controller.getSettings().getMidiSettings(controller.getMidiLoader().isFileKnown()).setKeyboardTrack(track);
		controller.setMidiKeyboard(getMidiChannel(track));
		for (int i=0; i<midiTracks.size(); i++) {
			if (i == track)
				midiTracks.get(i).setKeyboard();
			else
				midiTracks.get(i).unsetKeyboard();
		}
		revalidate();
	}
	
	/**
	 * Returns the corresponding MIDI channel of a track
	 * 
	 * @param track the track 
	 * @return the corresponding MIDI channel of a track
	 */
	public int getMidiChannel(int track) {
		return controller.getMidiLoader().getMidiTracks().get(track).getChannel();
	}

	/**
	 * Sets the instrument of a channel
	 * 
	 * @param channel the channel
	 * @param instrument the instrument for the channel
	 */
	public void setInstrument(int channel, int instrument) {
		for (int i=0; i<midiTracks.size(); i++) {
			if (getMidiChannel(i) == channel) {
				midiTracks.get(i).changeInstrument(instrument);
			}
		}
	}
	
	/**
	 * Sets the volume of a channel
	 * 
	 * @param channel the channel
	 * @param volume the volume for the channel
	 */
	public void setVolume(int channel, int volume) {
		for (int i=0; i<midiTracks.size(); i++) {
			if (getMidiChannel(i) == channel) {
				midiTracks.get(i).changeVolume(volume);
			}
		}
	}
	
	/**
	 * Resets the settings of a MIDI file
	 */
	public void resetSettings() {
		setKeyboardToTrack(0);
		spinnerTempo.setValue(new Integer(100));
		checkBoxMarkKey.setSelected(false);
		checkBoxAwaitNote.setSelected(false);
		for (MidiTrackGUI m : midiTracks)
			m.resetMidiTrackSettings();
	}

	//TODO: wo ist mark keys?
	/**
	 * Returns <tt>true</tt> if the await notes function is enabled
	 * 
	 * @return <tt>true</tt> if the await notes function is enabled
	 */
	public Boolean isAwaitNotesChecked() {
		return checkBoxAwaitNote.isSelected();
	}
}

class MidiTrackGUI extends JPanel implements ItemListener {
	private static final long serialVersionUID = -6360249971502138670L;
	private JTextField textTrack;
	private JComboBox comboInstrument;
	private JSlider slideVolume;
	private JPanel checkBoxPanel;
	private JCheckBox checkBoxAudio;
	private JCheckBox checkBoxVisual;
	private JButton buttonSet;
	private int number;
	private int color = -1;
	private final Controller controller;
	private MidiLoaderGUI mainMenu;
	private Color defaultBackground;
	private int instrument, volume;
	private String name;
	private Boolean audio, visual;
	private MidiTrack midiTrack;
	private int knownFile;
	private MidiTrackSettings midiTrackSettings;
	private final Settings settings;

	public MidiTrackGUI(Controller finalController, Settings finalSettings, MidiTrack m, int n, MidiLoaderGUI mlg) {
		this.controller = finalController;
		this.settings = finalSettings;
		defaultBackground = getBackground();
		number = n;
		knownFile = controller.getMidiLoader().isFileKnown();
		midiTrack = m;
		midiTrackSettings = controller.getSettings().getMidiSettings(knownFile).getMidiTrackSettings().get(number);
		mainMenu = mlg;

		name = midiTrackSettings.getTrackName();
		instrument = midiTrackSettings.getInstrument();
		volume = midiTrackSettings.getVolume();
		audio = midiTrackSettings.isAudioOn();
		visual = midiTrackSettings.isVisualOn();
		
		setLayout(new FlowLayout());
		textTrack = new JTextField(name, 15);
		textTrack.getDocument().addDocumentListener(new DocumentListener() {
		    public void insertUpdate(DocumentEvent ev) {
		    	controller.getSettings().getMidiSettings(knownFile).getMidiTrackSettings().get(number).setTrackName(textTrack.getText());
		    }
		    
		    public void removeUpdate(DocumentEvent ev) {
		    	controller.getSettings().getMidiSettings(knownFile).getMidiTrackSettings().get(number).setTrackName(textTrack.getText());
		    }
		    
		    public void changedUpdate(DocumentEvent ev) {}
		    
		});

		// handle percussion in a special way
		// TODO: channel handling ok?
		if (instrument == 128 || mainMenu.getMidiChannel(number) == 9) {
			String[] percussion = new String[openpiano.settings.Constants.INSTRUMENTS.length+1]; 
			for (int i=0; i<openpiano.settings.Constants.INSTRUMENTS.length; i++)
				percussion[i] = openpiano.settings.Constants.INSTRUMENTS[i];
			percussion[Constants.INSTRUMENTS.length] = "Percussion";
			comboInstrument = new JComboBox(percussion);
			comboInstrument.setSelectedIndex(Constants.INSTRUMENTS.length);
			comboInstrument.setEnabled(false);
		}
		else {
			comboInstrument = new JComboBox(openpiano.settings.Constants.INSTRUMENTS);
			comboInstrument.setSelectedIndex(instrument);
		}
		
		slideVolume = new JSlider(0, 127, volume);
        slideVolume.setPreferredSize(new Dimension(100,32));
        slideVolume.addChangeListener(new ChangeListener() {
        	public void stateChanged(ChangeEvent e) {
        		controller.getSettings().getMidiSettings().get(controller.getMidiLoader().isFileKnown()).getMidiTrackSettings().get(number).setVolume(slideVolume.getValue());
       			controller.setVolume(number, slideVolume.getValue());
        	}
        });

        checkBoxPanel = new JPanel(new FlowLayout());

        checkBoxAudio = new JCheckBox("Audio");
       	checkBoxAudio.setSelected(audio);
       	checkBoxAudio.addItemListener(this);

        checkBoxVisual = new JCheckBox("Visual");
        checkBoxVisual.setSelected(visual);
       	checkBoxVisual.addItemListener(this);

       	checkBoxPanel.add(checkBoxAudio);
        checkBoxPanel.add(checkBoxVisual);

		buttonSet = new JButton("Set Keyboard");
	    buttonSet.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				mainMenu.setKeyboardToTrack(number);
			}
		});
	    
	    checkBoxPanel.add(buttonSet);

		add(textTrack);
		add(comboInstrument);
		add(slideVolume);
		add(checkBoxPanel);
		comboInstrument.addItemListener(this);
		
		if (visual) {
			color = midiTrackSettings.getColor();
			setColor(color);
		}

	}

	public void itemStateChanged( ItemEvent e ) {
	  Object comp = e.getSource();
	  if (comp instanceof JCheckBox) {
		  String status = ((JCheckBox)comp).getText();
		  if (((JCheckBox)comp).isSelected()) {
			  if (status.equals("Audio"))
				  controller.setAudio(number, true);
			  else if (status.equals("Visual")) {
				  controller.setVisual(number, true);
				  color = mainMenu.getColor(number);
				  setColor(color);
				  midiTrackSettings.setColor(color);
			  }
		  }
		  else {
			  if (status.equals("Audio")) {
				  if (settings.getMidiOutput() != null)
					  settings.getMidiOutput().allNotesOff(mainMenu.getMidiChannel(number));
				  controller.setAudio(number, false);
			  }
			  else if (status.equals("Visual")) {
				  controller.setVisual(number, false);
				  color = -1;
				  unsetColor();
				  midiTrackSettings.setColor(color);
			  }
		  }
	  }
	  else if (comp instanceof JComboBox) {
		  if (e.getStateChange() == ItemEvent.SELECTED)
			  controller.setInstrument(number, mainMenu.getMidiChannel(number), ((JComboBox)comp).getSelectedIndex());
	  }
	}
	
	public void setKeyboard() {
		setBorder(BorderFactory.createBevelBorder(BevelBorder.RAISED));
//		setBorder(BorderFactory.createLineBorder(Color.black, 3));
	}
	
	public void unsetKeyboard() {
		setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
	}
	
	public void changeInstrument(int instrument) {
		  controller.setInstrument(number, mainMenu.getMidiChannel(number), instrument);
		  comboInstrument.setSelectedIndex(instrument);
	}
	
	public void changeVolume(int v) {
		slideVolume.setValue(v);
	}
	
	public void resetMidiTrackSettings() {
		name = midiTrack.getName();
		instrument = midiTrack.getInstrument();
		volume = midiTrack.getVolume();
		audio = true;
		visual = false;

		textTrack.setText(name);
		comboInstrument.setSelectedIndex(instrument);
		slideVolume.setValue(volume);
		checkBoxAudio.setSelected(true);
		checkBoxVisual.setSelected(false);
	}
	
	public int getColor() { return color; }
	
	public void setColor(int colorID) {
		float r = Constants.COLORSTART[colorID*3 + 0];
		float g = Constants.COLORSTART[colorID*3 + 1];
		float b = Constants.COLORSTART[colorID*3 + 2];
		
		Color color = new Color(r, g, b);
		
		setBackground(color);
		checkBoxPanel.setBackground(color);
		slideVolume.setBackground(color);
		checkBoxAudio.setBackground(color);
		checkBoxVisual.setBackground(color);
	}
	
	public void unsetColor() {
		setBackground(defaultBackground);
		checkBoxPanel.setBackground(defaultBackground);
		slideVolume.setBackground(defaultBackground);
		checkBoxAudio.setBackground(defaultBackground);
		checkBoxVisual.setBackground(defaultBackground);
	}
}