/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.midi;

import javax.sound.midi.MidiMessage;
import javax.sound.midi.ShortMessage;

/**
 * The MidiNote class represents a single MIDI note
 * 
 * @author Sascha Westermann
 * @version 1.0, 07/30/08
 */

public class MidiNote implements Comparable<MidiNote>{
	long startTick, endTick;
	int velocity;
	private MidiMessage midiMessage;

	/**
	 * Constructs a new MIDI note
	 * 
	 * @param startTick the start tick of the note
	 * @param endTick the end tick of the note
	 * @param midiMessage the MIDI message containing the note
	 */
	public MidiNote(long startTick, long endTick, MidiMessage midiMessage) {
		this.startTick = startTick;
		this.endTick = endTick;
		this.midiMessage = midiMessage;
		velocity = ((ShortMessage)midiMessage).getData2();
	}

	/**
	 * Returns the MIDI message for the note
	 * 
	 * @return the MIDI message for the note
	 */
	public MidiMessage getMidiMessage() { return midiMessage; }
	
	/**
	 * Returns the start tick of the note
	 * 
	 * @return the start tick of the note
	 */
	public long getStartTick() { return startTick; }
	
	/**
	 * Returns the end tick of the note
	 * 
	 * @return the end tick of the note
	 */
	public long getEndTick() { return endTick; }
	
	/**
	 * Returns the velocity of the note
	 * 
	 * @return the velocity of the note
	 */
	public int getVelocity() { return velocity; }
	
	/**
	 * Returns the corresponding key for the note
	 * 
	 * @return the corresponding key for the note
	 */
	public int getKey() { return ((ShortMessage)midiMessage).getData1()-21; }
	
	/**
	 * Compares the start ticks of MIDI notes
	 * 
	 * @param midiNote the MIDI note to compare with
	 * @return the difference of the start ticks
	 */
	public int compareTo(MidiNote midiNote) {
		return (int)(this.startTick - midiNote.getStartTick());
	}
}