/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.midi;

import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiDevice;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Receiver;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Transmitter;

/**
 * The MidiOutput class transmits events to a selected MIDI out port
 * 
 * @author Sascha Westermann
 * @version 1.0, 07/30/08
 */

public class MidiOutput implements Transmitter {
	private MidiDevice midiSequencer;
	private Receiver midiOut;

	/**
	 * Constructs a new Transmitter to a MIDI out port
	 * 
	 * @param midiDevice a MIDI out port
	 */
	public MidiOutput(MidiDevice.Info midiDevice) {
		try {
			midiSequencer = MidiSystem.getMidiDevice(midiDevice);
			midiSequencer.open();
		} catch (MidiUnavailableException e) { System.out.println("MIDI sequencer not available"); }

		if (midiSequencer != null) {
			try	{
				midiOut = midiSequencer.getReceiver();

				
				// dirty bugfix
				ShortMessage resetController = new ShortMessage();
				try {
					resetController.setMessage( ShortMessage.CONTROL_CHANGE, 0, 121, 0 );
				} catch (Exception ex) {}
				midiOut.send(resetController, -1);
				
				ShortMessage shortMessage = new ShortMessage();
				try {
					shortMessage.setMessage(ShortMessage.PROGRAM_CHANGE, 0, 0, 0);
				} catch (InvalidMidiDataException e) {}
				midiOut.send(shortMessage, -1);
			} catch (MidiUnavailableException e) {
				System.out.println("cannot connect the MIDI sequencer to MidiOutput");
			}
		}
	}

	/**
	 * Closes the transmitter und mutes all sounds
	 */
	public void close() {
		allNotesOff();
		if (midiOut != null)
			midiOut.close();
		if (midiSequencer != null && midiSequencer.isOpen())
			midiSequencer.close();
	}

	/**
	 * Returns the receiver for the transmitter
	 * 
	 * @return the receiver for the transmitter
	 */
	public Receiver getReceiver() { return midiOut; }
	
	/**
	 * Sets the receiver for the transmitter
	 * 
	 * @param receiver the receiver for the transmitter
	 */
	public void setReceiver(Receiver receiver) { midiOut = receiver; }

	/**
	 * Mutes all notes
	 */
	public void allNotesOff() {
		for (int i=0; i<16; i++) {
			ShortMessage noteoff = new ShortMessage();
			try {
				noteoff.setMessage( ShortMessage.CONTROL_CHANGE, i, 123, 0 );
			} catch (Exception ex) {}
			if (midiOut != null && midiSequencer != null && midiSequencer.isOpen())
				midiOut.send(noteoff, -1);
		}
	}
	
	/**
	 * Mutes all notes on a specified channel
	 * 
	 * @param channel the channel to mute
	 */
	public void allNotesOff(int channel) {
		ShortMessage noteoff = new ShortMessage();
		try {
			noteoff.setMessage( ShortMessage.CONTROL_CHANGE, channel, 123, 0 );
		} catch (Exception ex) {}
		if (midiOut != null && midiSequencer != null && midiSequencer.isOpen())
			midiOut.send(noteoff, -1);
	}

	/**
	 * Mutes a specified note
	 * 
	 * @param midiMessage the specified note
	 */
	public void noteOff(ShortMessage midiMessage) {
		ShortMessage noteOff = new ShortMessage();
		try {
			noteOff.setMessage( ShortMessage.NOTE_OFF, midiMessage.getChannel(), midiMessage.getData1(), 0 );
		} catch (Exception ex) {}
		if (midiOut != null && midiSequencer != null && midiSequencer.isOpen())
			midiOut.send(noteOff, -1);
	}
	
	/**
	 * Sends a MIDI event to the synthesizer
	 * 
	 * @param midiMessage the MIDI event
	 */
	public void send(MidiMessage midiMessage) {
		if (midiOut != null && midiSequencer != null && midiSequencer.isOpen())
			midiOut.send(midiMessage, -1);
	}
}