/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.midi;

import java.util.LinkedList;
import java.util.Collections;

/**
 * The MidiTrack class represents single tracks of a MIDI file
 * 
 * @author Sascha Westermann
 * @version 1.0, 07/30/08
 */

public class MidiTrack {
	private String trackName;
	private int instrument;
	private int channel;
	private int volume;
	private int controlPos;
	private LinkedList<MidiNote> midiNotes;
	private LinkedList<MidiControl> midiControls;

	/**
	 * Constructs a new MIDI track
	 * 
	 * @param trackName the name of the track
	 * @param channel the channel this track relies on
	 */
	public MidiTrack(String trackName, int channel) {
		midiNotes = new LinkedList<MidiNote>();
		midiControls = new LinkedList<MidiControl>();
		this.trackName = trackName;
		instrument = 128;
		this.channel = channel;
		volume = -1;
		controlPos = 0;
	}

	/**
	 * Constructs a new MIDI track with undefined name
	 * 
	 * @param channel the channel this track relies on
	 */
	public MidiTrack(int channel) {
		midiNotes = new LinkedList<MidiNote>();
		midiControls = new LinkedList<MidiControl>();
		trackName = "Untitled";
		instrument = 128;
		this.channel = channel;
		volume = -1;
	}
	
	/**
	 * Adds a new MIDI note to the track
	 * @param midiNote the midiNote to add
	 */
	public void addMidiNote(MidiNote midiNote) { getMidiNotes().add(midiNote); }
	
	/**
	 * Adds a new MIDI control message to the track
	 * @param midiControl the MIDI control message to add
	 */
	public void addMidiControl(MidiControl midiControl) { getMidiControls().add(midiControl); }

	/**
	 * Returns the list of MIDI notes
	 * @return the list of MIDI notes
	 */
	public LinkedList<MidiNote> getMidiNotes() { return midiNotes; }

	/**
	 * Returns the list of MIDI control messages
	 * @return the list of MIDI control messages
	 */
	public LinkedList<MidiControl> getMidiControls() { return midiControls; }

	/**
	 * Sets the name of the track
	 * @param trackName the name of the track
	 */
	public void setTrackName(String trackName) { this.trackName = trackName; }
	
	/**
	 * Sets the instrument of the track
	 * @param instrument the instrument of the track
	 */
	public void setInstrument(int instrument) { this.instrument = instrument; }
	
	/**
	 * Sets the channel of the track
	 * @param channel the channel of the track
	 */
	public void setChannel(int channel) { this.channel = channel; }
	
	/**
	 * Sets the volume of the track
	 * 
	 * @param volume the volume of the track
	 */
	public void setVolume(int volume) { this.volume = volume; }

	/**
	 * Returns the name of the track
	 * 
	 * @return the name of the track
	 */
	public String getName() { return trackName; }
	
	/**
	 * Returns the instrument of the track
	 * 
	 * @return the instrument of the track
	 */
	public int getInstrument() { return instrument; }
	
	/**
	 * Returns the channel of the track
	 * 
	 * @return the channel of the track
	 */
	public int getChannel() { return channel; }
	
	/**
	 * Returns the volume of the track
	 * 
	 * @return the volume of the track
	 */
	public int getVolume() { return volume; }
	
	/**
	 * Returns the actual index of the relevant MIDI control message
	 * @return the actual index of the relevant MIDI control message
	 */
	public int getControlPos() { return controlPos; }

	/**
	 * Sets the actual index of the relevant MIDI control message
	 * @param controlPos the actual index of the relevant MIDI control message
	 */
	public void setControlPos(int controlPos) { this.controlPos = controlPos; }
	
	/**
	 * Sorts the elements of a MIDI track
	 */
	public void sortElements() {
		Collections.sort(midiNotes);
		Collections.sort(midiControls);
	}
}