/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.notes;

import java.util.Collections;
import java.util.LinkedList;
import javax.media.opengl.GL;
import javax.sound.midi.ShortMessage;
import openpiano.control.Controller;
import openpiano.midi.MidiNote;
import openpiano.settings.Constants;
import openpiano.settings.MidiTrackSettings;
import openpiano.settings.Settings;

/**
 * The Blocks class is a data-structure to handle single notes
 *
 * @author  Sascha Westermann
 * @version 1.0, 07/30/08
 */
public class Blocks {
	private class Block {
		private Boolean awaitedNote = false;
		private Boolean highlightedNote = false;
	    private Boolean markedKey = false;
		private Boolean playedNote = false;
		private Boolean autoPressedKey = false;
		private Boolean pressedKey = false;
		private Boolean lateCleanup = false;
		private float hit;
		private float[] pos;
		private float scale = 0.97f;
		private float[] size;
		private float startpoint;
		private float tempscale;
		private GL gl;
	    private MidiNote midiNote;
	    private MidiTrackSettings midiTrack;

		public Block(GL g, MidiTrackSettings midiTrack, int track, MidiNote m, float[] p, float[] s) {
			gl = g;
			midiNote = m;
			pos = p;
			startpoint = pos[2];
			size = s;
			size[2]*=scale;
			hit = 0.0f;
			// dirty bugfix to prevent bleeding
			tempscale = 0.01f * track;
			this.midiTrack = midiTrack;
		}
		
		private void createBlock() {
			gl.glPushMatrix();
			gl.glBegin(GL.GL_QUADS);
		        // front
		        gl.glNormal3f(   0.0f,    0.0f,     1.0f);
				gl.glVertex3f(   0.0f - tempscale,    0.0f - tempscale,      hit + tempscale);
				gl.glVertex3f(size[0] + tempscale,    0.0f - tempscale,      hit + tempscale);
				gl.glVertex3f(size[0] + tempscale, size[1] + tempscale,      hit + tempscale);
				gl.glVertex3f(   0.0f - tempscale, size[1] + tempscale,      hit + tempscale);

				// back
				gl.glNormal3f(   0.0f,    0.0f,    -1.0f);
				gl.glVertex3f(   0.0f - tempscale,    0.0f - tempscale, -size[2] - tempscale);
				gl.glVertex3f(   0.0f - tempscale, size[1] + tempscale, -size[2] - tempscale);
				gl.glVertex3f(size[0] + tempscale, size[1] + tempscale, -size[2] - tempscale);
				gl.glVertex3f(size[0] + tempscale,    0.0f - tempscale, -size[2] - tempscale);

				// top
				gl.glNormal3f(   0.0f,    1.0f,     0.0f);
				gl.glVertex3f(   0.0f - tempscale, size[1] + tempscale, -size[2] - tempscale);
				gl.glVertex3f(   0.0f - tempscale, size[1] + tempscale,      hit + tempscale);
				gl.glVertex3f(size[0] + tempscale, size[1] + tempscale,      hit + tempscale);
				gl.glVertex3f(size[0] + tempscale, size[1] + tempscale, -size[2] - tempscale);

				// bottom
				gl.glNormal3f(   0.0f,   -1.0f,     0.0f);
				gl.glVertex3f(   0.0f - tempscale,    0.0f - tempscale, -size[2] - tempscale);
				gl.glVertex3f(size[0] + tempscale,    0.0f - tempscale, -size[2] - tempscale);
				gl.glVertex3f(size[0] + tempscale,    0.0f - tempscale,      hit + tempscale);
				gl.glVertex3f(   0.0f - tempscale,    0.0f - tempscale,      hit + tempscale);

				// right
				gl.glNormal3f(   1.0f,    0.0f,     0.0f);
				gl.glVertex3f(size[0] + tempscale,    0.0f - tempscale, -size[2] - tempscale);
				gl.glVertex3f(size[0] + tempscale, size[1] + tempscale, -size[2] - tempscale);
				gl.glVertex3f(size[0] + tempscale, size[1] + tempscale,      hit + tempscale);
				gl.glVertex3f(size[0] + tempscale,    0.0f - tempscale,      hit + tempscale);

				// left
				gl.glNormal3f(  -1.0f,    0.0f,      0.0f);
				gl.glVertex3f(   0.0f - tempscale,    0.0f - tempscale,  -size[2] - tempscale);
				gl.glVertex3f(   0.0f - tempscale,    0.0f - tempscale,       hit + tempscale);
				gl.glVertex3f(   0.0f - tempscale,  size[1] + tempscale,      hit + tempscale);
				gl.glVertex3f(   0.0f - tempscale,  size[1] + tempscale, -size[2] - tempscale);
	        gl.glEnd();
			gl.glPopMatrix();
	    }

	    private void drawBlock(Boolean useable) {
			gl.glEnable(GL.GL_BLEND);
			gl.glBlendFunc (GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
			gl.glEnable(GL.GL_COLOR_MATERIAL);

			float blend = midiNote.getVelocity() / 127.0f;
			int color = midiTrack.getColor();
			float r = Constants.COLORSTART[color*3 + 0];
			float g = Constants.COLORSTART[color*3 + 1];
			float b = Constants.COLORSTART[color*3 + 2];
			float radd = Constants.COLOREND[color*3 + 0];
			float gadd = Constants.COLOREND[color*3 + 1];
			float badd = Constants.COLOREND[color*3 + 2];

			if (useable) {
				if (highlightedNote )
					gl.glColor4f(1.0f,0.2f,0.2f,0.7f);
				else
					gl.glColor4f(r-((r-radd)*blend), g-((g-gadd)*blend), b-((b-badd)*blend), 1.0f);
			}
			else {
				gl.glColor4f(r-((r-radd)*blend), g-((g-gadd)*blend), b-((b-badd)*blend), 0.3f);
			}

			createBlock();

			gl.glDisable(GL.GL_COLOR_MATERIAL);
			gl.glDisable(GL.GL_BLEND);

		}

		
		public long getEndTick() { return midiNote.getEndTick(); }

		public float getMarkValue(float position, long tick, int tpb) {
			if (hasReachedKeyboard(position))
				return 1.0f;
			else {
				float halfBeat = tpb / 2.0f;
				return 1.0f - ((midiNote.getStartTick() - tick) / halfBeat);
			}
		}
		public ShortMessage getMidiMessage() { return (ShortMessage)midiNote.getMidiMessage(); }
		public long getStartTick() { return midiNote.getStartTick(); }
		public Boolean hasAutoPressedKey() { return autoPressedKey; }
		public Boolean hasAwaitedNote() { return awaitedNote; }
		public Boolean hasHighlightedNote() { return highlightedNote; }
		public Boolean hasLateCleanup() { return lateCleanup; }
		public Boolean hasMarkedKey() { return markedKey; }
		public Boolean hasPlayedNote() { return playedNote; }
		public Boolean hasPressedKey() { return pressedKey; }
		public Boolean hasReachedKeyboard(float position) {
			if (startpoint + position >= 0.0f)
				return true;
			else
				return false;
		}

		public void reset() {
			hit = 0.0f;
			autoPressedKey = false;
			awaitedNote = false;
			lateCleanup = false;
			markedKey = false;
			playedNote = false;
			highlightedNote = false;
			pressedKey = false;
		}
		public void resetSize() {
			hit = 0.0f;
		}
		public void setAutoPressedKey(Boolean b) { autoPressedKey = b; }
		
		public void setAwaitedNote(Boolean b) { awaitedNote = b; }
		
		public void setHighlightedNote(Boolean b) { highlightedNote = b; }
		
		public void setLateCleanup(Boolean b) { lateCleanup = b; }
		
		public void setMarkedKey(Boolean b) { markedKey = b; }
		
		public void setPlayedNote(Boolean b) { playedNote = b; }
		
		public void setPressedKey(Boolean b) { pressedKey = b; }

		public void showBlock(Boolean useable, float position) {
			// handle scale factor
			// the note may be bigger, but we just show the scale factor
			if (hit != -size[2]) {
				gl.glPushMatrix();
				gl.glTranslatef(pos[0], pos[1], startpoint + position);
				drawBlock(useable);
				gl.glPopMatrix();
			}
		}
		
		public void updateSize(float position) {
			hit = -(startpoint + position);
			if (hit  < -size[2])
				hit = -size[2];
		}
	}

	private class SpecialBlock implements Comparable<SpecialBlock> {
		int key, track;
		Block block;
		SpecialBlock(Block b, int k, int t) {
			block = b;
			key = k;
			track = t;
		}
		public int getKey() { return key; }
		public int getTrack() { return track; }
		public Block getBlock() { return block; }
		
		public int compareTo(SpecialBlock b) {
			return (int)(this.block.getStartTick() - b.getBlock().getStartTick());
		}

	}
	
	private Controller controller;
	private Settings settings;
	private LinkedList<LinkedList<Integer>> blockIndex;
	private LinkedList<LinkedList<LinkedList<Block>>> blocks;
	private LinkedList<LinkedList<Block>> currentBlock;
	private LinkedList<SpecialBlock> lateCleanList;
	private int halfBeat, numKeys, numTracks, ticksPerBeat;
	private float currentPosition = 0.0f;
	private long currentTick = 0;
	private long lastVisibleTick = 0;
	private Boolean reverse = false;
	private Boolean isAwaitNotesOn, isMarkKeysOn;
	
	/**
	 * Constructs a data structure for single notes
	 * 
     * @param controller the main controller of the application
     * @param settings the settings object
	 * @param keys the amount of keys of the piano
	 * @param tracks the amount of tracks in the MIDI file
	 */
	public Blocks(Controller controller, Settings settings, int keys, int tracks) {
		this.controller = controller;
		this.settings = settings;
		numKeys = keys;
		numTracks = tracks;
		ticksPerBeat = controller.getMidiLoader().getTicksPerBeat();
		halfBeat = Math.round(controller.getMidiLoader().getTicksPerBeat()/2.0f);
		blocks = new LinkedList<LinkedList<LinkedList<Block>>>();
		blockIndex = new LinkedList<LinkedList<Integer>>();
		currentBlock = new LinkedList<LinkedList<Block>>();
		lateCleanList = new LinkedList<SpecialBlock>();
		
		for (int key=0; key<numKeys; key++) {
			blocks.add(new LinkedList<LinkedList<Block>>());
			blockIndex.add(new LinkedList<Integer>());
			currentBlock.add(new LinkedList<Block>());
			for (int track=0; track<numTracks; track++) {
				blocks.get(key).add(new LinkedList<Block>());
				blockIndex.get(key).add(new Integer(0));
				currentBlock.get(key).add(null);
			}
		}
	}
	
	/**
	 * Mark a block for late cleanup. This method adds a single block to a 
	 * list of blocks for late handling. This is useful for notes that has been
	 * played by the user, but did not pass the piano for cleanup yet 
	 * 
	 * @param block the block to deal with
	 * @param key the key of the block
	 * @param track the track of the block
	 */
	private void addBlockForLateCleanup(Block block, int key, int track) {
		if (block.hasMarkedKey()) {
			block.setMarkedKey(false);
			controller.getPiano().setKeyMarkerOff(key);
		}

		block.setLateCleanup(true);

		lateCleanList.add(new SpecialBlock(block, key, track));
		currentBlock.get(key).set(track, null);
	}
	
	/**
	 * Adds a single note/block to the list
	 * 
	 * @param midiNote the note according to the MIDI file
	 * @param key the key as defined by the note
	 * @param track the track according to the MIDI file
	 * @param refTicks a reference for the displayed size
	 * @param flatnote <tt>true</tt> if note is a flat note
	 * @param position the position in 3d space according to the start tick of the note
	 */
	//TODO: refTicks
	public void addNote(MidiNote midiNote, int key, int track, float refTicks, Boolean flatnote, float position) {
        float[] size = new float[3];
		if (flatnote) {
			size[0] = 1.0f;
			size[1] = 2.6f;
		}
		else {
			size[0] = 2.0f;
			size[1] = 2.0f;
		}
		size[2] = (Constants.VIEWAREA * (midiNote.getEndTick() - midiNote.getStartTick())) / refTicks;
        
	    float[] pos = new float[3];
        pos[0] = position;
        pos[1] = -1.0f;
        pos[2] = -(Constants.VIEWAREA + controller.tickToGL(midiNote.getStartTick()));

        MidiTrackSettings midiTrack = settings.getMidiSettings().get(controller.getMidiLoader().isFileKnown()).getMidiTrackSettings().get(track);
		blocks.get(key).get(track).add(new Block(controller.getGL(), midiTrack, track, midiNote, pos, size));
	}
	
	/**
	 * Cleans up a block and restores default values
	 * 
	 * @param block the block to deal with
	 * @param key the key of the block
	 * @param track the track of the block
	 */
	private void cleanupBlock(Block block, int key, int track) {
		Boolean markedKey = block.hasMarkedKey();
		Boolean playedNote = block.hasPlayedNote();
		Boolean autoPressedKey = block.hasAutoPressedKey();
		
		if (markedKey)
			controller.getPiano().setKeyMarkerOff(key);
		if (playedNote) {
			if (settings.getMidiOutput() != null)
				settings.getMidiOutput().noteOff(block.getMidiMessage());
		}
		if (autoPressedKey)
			controller.getPiano().sendKey(key, 0, true);
			
		block.reset();
	}

	/**
	 * Returns the actual block in the list
	 * 
	 * @param key the relevant key
	 * @param track the relevant track
	 * @return the actual block in the list
	 */
	private Block getActual(int key, int track) {
		return blocks.get(key).get(track).get(blockIndex.get(key).get(track));
	}
	
	/**
	 * Returns the end tick of the actual block in the list
	 * 
	 * @param key the relevant key
	 * @param track the relevant track
	 * @return the end tick of the actual block in the list
	 */
	private long getActualEndTick(int key, int track) {
		if (blockIndex.get(key).get(track) > -1)
			return blocks.get(key).get(track).get(blockIndex.get(key).get(track)).getEndTick();
		return Long.MIN_VALUE;
	}

	/**
	 * Returns the start tick of the actual block in the list
	 * 
	 * @param key the relevant key
	 * @param track the relevant track
	 * @return the start tick of the actual block in the list
	 */
	private long getActualStartTick(int key, int track) {
		if (blockIndex.get(key).get(track) < blocks.get(key).get(track).size())
			return blocks.get(key).get(track).get(blockIndex.get(key).get(track)).getStartTick();
		return Long.MAX_VALUE;
	}
	
	/**
	 * Returns the end tick of the forerunner of the actual block in the list
	 * 
	 * @param key the relevant key
	 * @param track the relevant track
	 * @return the end tick of the forerunner of the actual block in the list
	 */
	private long getLastEndTick(int key, int track) {
		if (blockIndex.get(key).get(track) > 0)
			return blocks.get(key).get(track).get(blockIndex.get(key).get(track)-1).getEndTick();
		return Long.MIN_VALUE;
	}

	/**
	 * Returns the start tick of the follower of the actual block in the list
	 * 
	 * @param key the relevant key
	 * @param track the relevant track
	 * @return the start tick of the follower of the actual block in the list
	 */
	private long getNextStartTick(int key, int track) {
		if (blockIndex.get(key).get(track)+1 < blocks.get(key).get(track).size())
			return blocks.get(key).get(track).get(blockIndex.get(key).get(track)+1).getStartTick();
		return Long.MAX_VALUE;
	}
	
	/**
	 * Returns the number of used keys
	 * 
	 * @return the number of used keys
	 */
	public int getNumKeys() { return numKeys; }
	
	/**
	 * Return the number of used tracks
	 * 
	 * @return the number of used tracks
	 */
	public int getNumTracks() { return numTracks; }

	/**
	 * Returns the index of the first visible block
	 * 
	 * @param key the relevant key
	 * @param track the relevant track
	 * @return the index of the first visible block
	 */
	private int getRemainingBlocks(int key, int track) {
		int pos = blockIndex.get(key).get(track);
		int max = blocks.get(key).get(track).size();
		int min = -1;
		int ret = 0;
		
		// increase index, if blocks have passed the keyboard
		for (int i=pos; i<max; i++) {
			if (currentTick >= blocks.get(key).get(track).get(i).getEndTick())
				ret++;
			else
				i = max;
		}
		
		// else - decrease index, if previous blocks have not passed the keyboard
		if (ret == 0) {
			for (int i=pos-1; i > min; i--) {
				if (currentTick < blocks.get(key).get(track).get(i).getEndTick())
					ret--;
				else
					i= min;
			}
		}
		return ret;
	}
	
	/**
	 * Central method to handle a block. Every action a block can receive or send is handled here.
	 * 
	 * @param block the block to deal with
	 * @param key the relevant key
	 * @param track the relevant track
	 * @param isAudioOn <tt>true</tt> if audio functions are selected for the block
	 * @param isVisualOn <tt>true</tt> if video functions are selected for the block
	 */
	private void handleBlock(Block block, int key, int track, Boolean isAudioOn, Boolean isVisualOn) {
		if (currentTick >= block.getEndTick()) {
			cleanupBlock(block, key, track);
			currentBlock.get(key).set(track, null);
		}
		else if (currentTick < block.getStartTick()) {
			Block lastBlock = null;
			if (blockIndex.get(key).get(track) > 0)
				lastBlock = blocks.get(key).get(track).get(blockIndex.get(key).get(track)-1);

			if ( (lastBlock != null && currentTick < lastBlock.getEndTick() && !lastBlock.hasLateCleanup())
				|| currentTick < block.getStartTick() - halfBeat) 
			{
				cleanupBlock(block, key, track);
				currentBlock.get(key).set(track, null);
			}
			else if (currentTick >= block.getStartTick() - halfBeat) {
				handleMarkKeys(block, key, track, isVisualOn);
				block.resetSize();
			}
		}
		else if (currentTick >= block.getStartTick()) {
			Boolean useable = isUseable(key);
			block.updateSize(currentPosition);
			handleAudio(block, isAudioOn, isVisualOn, useable);
			handleVisual(block, key, isAudioOn, isVisualOn, useable);
			handleMarkKeys(block, key, track, isVisualOn);
			if (useable)
				handleAwaitNotes(block, isAudioOn, isVisualOn);
		}
	}
	
	//TODO: allgemein nicht controller.isSliding, sondern lokale variable verwenden
	//KONSISTENZ!!

	/**
	 * Handle audio functions of a block
	 * 
	 * @param block the block to deal with
	 * @param isAudioOn <tt>true</tt> if audio functions are selected for the block
	 * @param isVisualOn <tt>true</tt> if video functions are selected for the block
	 * @param useable <tt>true</tt> if corresponding key is playable
	 */
	private void handleAudio(Block block, Boolean isAudioOn, Boolean isVisualOn, Boolean useable) {
		//TODO
		if (!block.hasPlayedNote() && (isAudioOn || (!useable && isVisualOn)) && !controller.isSliding() && controller.isPlayModeSet() && !reverse) {
			block.setPlayedNote(true);
			if (settings.getMidiOutput() != null)
				settings.getMidiOutput().send(block.getMidiMessage());
		}
		else if (block.hasPlayedNote() && ((!isAudioOn && useable) || controller.isSliding() || !controller.isPlayModeSet() || reverse)) {
			block.setPlayedNote(false);
			if (settings.getMidiOutput() != null)
				settings.getMidiOutput().noteOff(block.getMidiMessage());
		}
	}
	
	/**
	 * Handles the awaitNote function of a block
	 * 
	 * @param block the block to deal with
	 * @param isAudioOn <tt>true</tt> if audio functions are selected for the block
	 * @param isVisualOn <tt>true</tt> if video functions are selected for the block
	 */
	private void handleAwaitNotes(Block block, Boolean isAudioOn, Boolean isVisualOn) {
		if (!block.hasAwaitedNote() && isAwaitNotesOn && !isAudioOn && 
			isVisualOn && !controller.isSliding() && !reverse && !block.hasLateCleanup())
		{
			block.setAwaitedNote(true);
			if (!controller.isBpmStopped())
				controller.stopBpm();
		}
		else if (block.hasAwaitedNote() && (isAudioOn || !isVisualOn || !isAwaitNotesOn 
				|| controller.isSliding() || reverse) )
		{
			block.setAwaitedNote(false);
			if (controller.isBpmStopped())
				controller.startBpm();
			
		}
	}
	
	/**
	 * Handles the keys pressed by the user according to the available notes
	 * 
	 * @param isVisualOn List of tracks, <tt>true</tt> if the track's video function is selected
	 * @param isAudioOn List of tracks, <tt>true</tt> if the track's audio function is selected
	 */
	public void handleKeyPressed(LinkedList<Boolean> isVisualOn, LinkedList<Boolean> isAudioOn) {
		int firstKey = controller.getSettings().getFirstKey();
		int lastKey = controller.getSettings().getLastKey();
		
		LinkedList<SpecialBlock> currentNotes = new LinkedList<SpecialBlock>();
		for (int key=firstKey; key<=lastKey; key++) {
			for (int track=0; track<numTracks; track++) {
				if (currentBlock.get(key).get(track) != null && isVisualOn.get(track) && !isAudioOn.get(track))
					currentNotes.add(new SpecialBlock(currentBlock.get(key).get(track), key, track));
			}
		}
		
		if (!currentNotes.isEmpty()) {
			Collections.sort(currentNotes);
			
			//TODO: epsilon = zeit eines 1/64 takts (~menschenmöglich maximal)
			int epsilon = 20;
			long max = currentNotes.getFirst().getBlock().getStartTick() + epsilon;

			Boolean playedRelevantNotes = true;
			
			int pos = 0;
			for (int i=0; i<currentNotes.size(); i++) {
				int key = currentNotes.get(i).getKey();
				Block block = currentNotes.get(i).getBlock();
				if (block.getStartTick() < max) {
					pos++;
					if (!controller.getPiano().isKeyMarked(key) || !block.hasPressedKey()) {
						i = currentNotes.size();
						playedRelevantNotes = false;
					}
				}
				else
					i = currentNotes.size();
			}

			if (playedRelevantNotes) {
				for (int i=0; i<pos; i++) {
					Block block = currentNotes.get(i).getBlock();
					int key = currentNotes.get(i).getKey();
					int track = currentNotes.get(i).getTrack();

					block.setHighlightedNote(true);
					addBlockForLateCleanup(block, key, track);
				}
				if (isAwaitNotesOn && controller.isBpmStopped())
					controller.startBpm();
			}
		}
	}
	
	/**
	 * Handle the blocks marked for late cleanup
	 */
	public void handleLateCleanup() {
		for (int i=0; i<lateCleanList.size(); i++) {
			SpecialBlock b = lateCleanList.get(i);
			if (b.getBlock().getEndTick() <= currentTick
				|| b.getBlock().getStartTick() - halfBeat > currentTick)
			{
				cleanupBlock(b.getBlock(), b.getKey(), b.getTrack());
				lateCleanList.remove(i);
				i--;
			}
			else if (b.getBlock().getStartTick() <= currentTick) {
				b.getBlock().updateSize(currentPosition);
			}
			// startTick - halfBeat <= currentTick < startTick
			else
				b.getBlock().resetSize();
		}
	}
	
	/**
	 * Handles mark key functions of a block
	 * 
	 * @param block the block to deal with
	 * @param key the relevant key
	 * @param isVisualOn <tt>true</tt> if video functions are selected for the block
	 */
	private void handleMarkKeys(Block block, int key, int track, Boolean isVisualOn) {
		if (isMarkKeysOn && isVisualOn && !controller.isSliding() && controller.isPlayModeSet() && !reverse) {
			block.setMarkedKey(true);
			float markValue = block.getMarkValue(currentPosition, currentTick, ticksPerBeat);
			int color = settings.getMidiSettings().get(controller.getMidiLoader().isFileKnown()).getMidiTrackSettings().get(track).getColor();
			controller.getPiano().setKeyMarker(key, color, markValue);
		}
		else if (block.hasMarkedKey() && (!isMarkKeysOn || !isVisualOn || controller.isSliding() || !controller.isPlayModeSet() || reverse) ) {
			block.setMarkedKey(false);
			controller.getPiano().setKeyMarkerOff(key);
		}
	}
	
	/**
	 * Handles visual functions of a block
	 * 
	 * @param block the block to deal with
	 * @param key the relevant key
	 * @param isAudioOn <tt>true</tt> if audio functions are selected for the block
	 * @param isVisualOn <tt>true</tt> if video functions are selected for the block
	 * @param useable <tt>true</tt> if corresponding key is playable
	 */
	private void handleVisual(Block block, int key, Boolean isAudioOn, Boolean isVisualOn, Boolean useable) {
		if (!block.hasAutoPressedKey() && (isAudioOn || !useable) && isVisualOn && !controller.isSliding() && controller.isPlayModeSet() && !reverse) {
			block.setAutoPressedKey(true);
			controller.getPiano().sendKey(key, block.getMidiMessage().getData2(), true);
		}
		else if (block.hasAutoPressedKey() && ((!isAudioOn && useable)|| !isVisualOn || controller.isSliding() || !controller.isPlayModeSet() || reverse)) {
			block.setAutoPressedKey(false);
			controller.getPiano().sendKey(key, 0, true);
		}
	}
	/**
	 * Set the status of the await notes function
	 * 
	 * @param awaitNotes <tt>true</tt> if await notes function should be enabled
	 */
	public void setAwaitNotes(Boolean awaitNotes) { isAwaitNotesOn = awaitNotes; }

	/**
	 * Set a key on all tracks to status pressed
	 * 
	 * @param key the relevant key
	 */
	public void setKeyPressed(int key) {
		for (int track = 0; track < numTracks; track++) {
			Block block = currentBlock.get(key).get(track);
			if (block != null) {
				block.setPressedKey(true);
			}
		}
	}

	/**
	 * Sets a key on all tracks to status released
	 * 
	 * @param key the relevant key
	 */
	public void setKeyReleased(int key) {
		for (Block b : currentBlock.get(key)) {
			if (b != null) {
				b.setPressedKey(false);
			}
		}
		for (SpecialBlock b : lateCleanList) {
			if (b.getKey() == key && b.getBlock().hasHighlightedNote())
				b.getBlock().setHighlightedNote(false);
		}
	}
	
	/**
	 * Sets the current position in the 3D environment
	 *
	 * @param position the current position
	 */
	public void setCurrentPosition(float position) { currentPosition = position; }
	
	/**
	 * Sets the current tick
	 *
	 * @param tick the current tick
	 */
	public void setCurrentTick(long tick) { currentTick = tick; }
	
	/**
	 * Sets the actual block to its forerunner
	 * 
	 * @param key the relevant key
	 * @param track the relevant track
	 * @return the forerunner of the actual block
	 */
	private Block setLast(int key, int track) {
		if (blockIndex.get(key).get(track) > 0) {
			blockIndex.get(key).set(track, blockIndex.get(key).get(track)-1);
			return blocks.get(key).get(track).get(blockIndex.get(key).get(track));
		}
		return null;
	}

	/**
	 * Sets the distance of visible ticks
	 * 
	 * @param tick the visible ticks
	 */
	public void setLastVisibleTick(long tick) { lastVisibleTick = currentTick + tick; }
	
	/**
	 * Sets the status of the mark key function
	 * 
	 * @param markKeys <tt>true</tt> if mark keys function should be enabled
	 */
	public void setMarkKeys(Boolean markKeys) { isMarkKeysOn = markKeys; }
	
	private Boolean isUseable(int key) {
		if (controller.getSettings().getFirstKey() > key || controller.getSettings().getLastKey() < key)
			return false;
		else
			return true;
	}

	/**
	 * Sets the actual block to its follower
	 * 
	 * @param key the relevant key
	 * @param track the relevant track
	 * @return the follower of the actual block
	 */
	private Block setNext(int key, int track) {
		if (blockIndex.get(key).get(track)+1 < blocks.get(key).get(track).size()) {
			blockIndex.get(key).set( track, blockIndex.get(key).get(track)+1 );
			return blocks.get(key).get(track).get( blockIndex.get(key).get(track) );
		}
		return null;
	}
	
	/**
	 * Sets the reverse status
	 * 
	 * @param reverse <tt>true</tt> to set the reverse status
	 */
	public void setReverse(Boolean reverse) { this.reverse = reverse; }
	
	/**
	 * Draw the blocks
	 * 
	 * @param key the relevant key
	 * @param track the relevant track
	 * @param useable <tt>true</tt> if corresponding key is playable
	 * @param audio <tt>true</tt> if audio function is selected
	 * @param visual <tt>true</tt> if video function is selected
	 * @param markKeys <tt>true</tt> if mark keys function is selected 
	 */
	public void showBlocks(int key, int track, Boolean useable, Boolean audio, Boolean visual, Boolean markKeys) {
		if (blocks.get(key).get(track).size() != 0) {
			int add = getRemainingBlocks(key, track);
			for (int i=blockIndex.get(key).get(track)+add; i<blocks.get(key).get(track).size(); i++) {
				if (blocks.get(key).get(track).get(i).getStartTick() < lastVisibleTick)
					blocks.get(key).get(track).get(i).showBlock(useable, currentPosition);
				else
					i=blocks.get(key).get(track).size();
			}
		}
	}
	
	/**
	 * Updates the pointer to the current block
	 * 
	 * @param key the relevant key
	 * @param track the relevant track
	 * @param audio <tt>true</tt> if audio function is selected
	 * @param visual <tt>true</tt> if video function is selected
	 */
	public void updateBlocks(int key, int track, Boolean audio, Boolean visual) {
		if (blocks.get(key).get(track).size() != 0) {
			// handle current block
			if (currentBlock.get(key).get(track) != null)
				handleBlock(currentBlock.get(key).get(track), key, track, audio, visual);
			
			// if no block is available
			if (currentBlock.get(key).get(track) == null) {
				if (!reverse && currentTick >= getActualStartTick(key, track) - halfBeat) {
					if (currentTick < getActualEndTick(key, track)
						&& !getActual(key, track).hasLateCleanup())
					{
		 				currentBlock.get(key).set(track, getActual(key, track));
					}
						
					while (currentTick >= getNextStartTick(key, track) - halfBeat 
						   && ( currentTick >= getActualEndTick(key, track) 
							    || (getActual(key, track) != null && getActual(key, track).hasLateCleanup()) ) ) 
					{
						if (currentBlock.get(key).get(track) != null
							&& !currentBlock.get(key).get(track).hasLateCleanup() )
						{
							cleanupBlock(currentBlock.get(key).get(track), key, track);
						}
	
						currentBlock.get(key).set(track, setNext(key, track));
					}
	
					if (currentBlock.get(key).get(track) != null)
						handleBlock(currentBlock.get(key).get(track), key, track, audio, visual);
				}
				else if (reverse && currentTick < getActualEndTick(key, track)) {
					currentBlock.get(key).set(track, getActual(key, track));
					
					while (currentTick < getLastEndTick(key, track)) {
						if (currentBlock.get(key).get(track) != null)
							cleanupBlock(currentBlock.get(key).get(track), key, track);
	
						currentBlock.get(key).set(track, setLast(key, track));
					}
					
					if (currentBlock.get(key).get(track) != null)
						handleBlock(currentBlock.get(key).get(track), key, track, audio, visual);
				}
			}
		}
	}
}