/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.notes;

import java.util.LinkedList;
import javax.media.opengl.GL;
import openpiano.control.Controller;
import openpiano.midi.MidiLoader;
import openpiano.midi.MidiNote;
import openpiano.settings.Constants;
import openpiano.settings.Settings;

/**
 * The Notes class is the central class to manage blocks
 *
 * @author  Sascha Westermann
 * @version 1.0, 07/30/08
 */

public class Notes {
	private long lastTick = Long.MIN_VALUE;
	private LinkedList<Boolean> blockFlatnote;
	private LinkedList<Float> blockPos;
	private Blocks blocks;
	private Controller controller;
	private Settings settings;
	private GL gl;
	private Boolean[] notesToBePlayed;
	private Boolean settingTick = false;
	private Boolean showingNotes = false;
	private Boolean sendingKey = false;
	private Boolean settingNotes = false;
	
	/**
	 * Constructs a new Notes object for a 88 key piano without any pre-loaded notes
	 * 
     * @param controller the main controller of the application
	 */
	public Notes(Controller controller, Settings settings) {
		this.controller = controller;
		this.settings = settings;
		gl = controller.getGL();
		
		blockPos = new LinkedList<Float>();
		for (int i=0; i<3; i++)
			blockPos.add(Constants.POS_START[i]);
		for (int i=3; i<87; i++)
			blockPos.add((int)((i-3)/12) * Constants.OCTAVELENGTH + Constants.POS_OCTAVE[(i-3)%12]);
		blockPos.add(7*Constants.OCTAVELENGTH + Constants.POS_END);

		blockFlatnote = new LinkedList<Boolean>();
		blockFlatnote.add(false);
		blockFlatnote.add(true);
		blockFlatnote.add(false);
		for (int i=0; i<7; i++) {
			blockFlatnote.add(false);
			blockFlatnote.add(true);
			blockFlatnote.add(false);
			blockFlatnote.add(true);
			blockFlatnote.add(false);
			blockFlatnote.add(false);
			blockFlatnote.add(true);
			blockFlatnote.add(false);
			blockFlatnote.add(true);
			blockFlatnote.add(false);
			blockFlatnote.add(true);
			blockFlatnote.add(false);
		}
		blockFlatnote.add(false);
		
		notesToBePlayed = new Boolean[88];
		for (int i=0; i<notesToBePlayed.length; i++)
			notesToBePlayed[i] = false;
	}
	
	/**
	 * Sends a piano key to a the corresponding block, if available
	 * 
     * @param key the affected key
	 * @param velocity the velocity of the key, a velocity of 0 means release key
	 */
    public synchronized void sendKey(int key, int velocity) {
    	if (settingTick || showingNotes || settingNotes) {
    		try {
    			wait();
    		} catch (Exception e) {}
    	}
    	sendingKey = true;
		if (blocks != null) {
			if (velocity > 0)
				blocks.setKeyPressed(key);
			else
				blocks.setKeyReleased(key);
		}
		sendingKey = false;
	}

    /**
     * Sets new notes to handle
     * 
     * @param midiLoader a MidiLoader object handling a MIDI file
     */
	public synchronized void setNotes(MidiLoader midiLoader) {
    	if (settingTick || showingNotes || sendingKey) {
    		try {
    			wait();
    		} catch (Exception e) {}
    	}
    	settingNotes = true;

		float refTicks = midiLoader.getTicksPerBeat() * Constants.VIEWBEAT;
		int tracks = midiLoader.getMidiTracks().size();
		blocks = new Blocks(controller, settings, 88, tracks);

		for (int track=0; track<tracks; track++) {
			for (MidiNote midiNote : midiLoader.getMidiTracks().get(track).getMidiNotes()) {
				try {
					int key = midiNote.getKey();
					blocks.addNote(midiNote, key, track, refTicks, blockFlatnote.get(key), blockPos.get(key));
				} catch (Exception e) {}
			}
		}
		
		settingNotes = false;
		notify();
	}
	
	/**
	 * Sets the actual tick and performs updates to blocks
	 * 
	 * @param currentTick the actual tick in the MIDI file
	 * @param isSliding <tt>true</tt> if the user is using the slider
	 */
	public synchronized void setTick(long currentTick, Boolean isSliding) {
		if (showingNotes || settingNotes || sendingKey) {
			try {
				wait();
			} catch (Exception e) {}
		}
		settingTick = true;
		
		Boolean reverse = false;
		if (currentTick - lastTick < 0)
			reverse = true;
		
		blocks.setCurrentTick(currentTick);
		blocks.setCurrentPosition(controller.tickToGL(currentTick) + Constants.VIEWAREA);
		blocks.setReverse(reverse);

		blocks.setAwaitNotes(controller.getSettings().isAwaitNotesOn());
		blocks.setMarkKeys(controller.getSettings().isMarkKeysOn());

		LinkedList<Boolean> isVisualOn = new LinkedList<Boolean>();
		LinkedList<Boolean> isAudioOn = new LinkedList<Boolean>();
		for (int track=0; track<blocks.getNumTracks(); track++) {
			isVisualOn.add(new Boolean(controller.getSettings().isVisualOn(track)));
			isAudioOn.add(new Boolean(controller.getSettings().isAudioOn(track)));
		}
		
		blocks.handleKeyPressed(isVisualOn, isAudioOn);		
		blocks.handleLateCleanup();
		
		for (int key=0; key<blocks.getNumKeys(); key++) {
			for (int track=0; track<blocks.getNumTracks(); track++) {
				blocks.updateBlocks(key, track, isAudioOn.get(track), isVisualOn.get(track));
			}
		}
		
		lastTick = currentTick;
		settingTick = false;
		notify();
	}
	
	/**
	 * Shows the lines behind the piano keys
	 */
	public void showLines() {
		gl.glDisable(GL.GL_LIGHTING);
		gl.glDisable(GL.GL_LIGHT0);

		for (int key=controller.getSettings().getFirstKey(); key<=controller.getSettings().getLastKey(); key++) {
			if (!blockFlatnote.get(key)) {
				gl.glColor3f(0.85f, 0.85f, 0.85f);
				gl.glBegin(GL.GL_LINES);
					gl.glVertex3d(blockPos.get(key) + 1.0f, 0, -controller.getCamera().getVisibleArea());
					gl.glVertex3d(blockPos.get(key) + 1.0f, 0, 0);
				gl.glEnd();

				if ( (key-3)%12 == 0 ) {
					gl.glColor3f(0.55f, 0.55f, 0.55f);
			 		gl.glLineWidth(3.0f);
					gl.glBegin(GL.GL_LINES);
						gl.glVertex3d(blockPos.get(key), 0, -controller.getCamera().getVisibleArea());
						gl.glVertex3d(blockPos.get(key), 0, 0);
					gl.glEnd();
			 		gl.glLineWidth(1.0f);
				}
			}
		}
		gl.glEnable(GL.GL_LIGHTING);
		gl.glEnable(GL.GL_LIGHT0);
	}
	
	/**
	 * Shows the notes
	 */
	public synchronized void showNotes() {
		if (settingTick || settingNotes || sendingKey) {
			try {
				wait();
			} catch (Exception e) {}
		}
		showingNotes = true;
		int firstKey = controller.getSettings().getFirstKey();
		int lastKey = controller.getSettings().getLastKey();
		
		// due to a local copy, the settings are consistent for this run
		LinkedList<Boolean> isVisualOn = new LinkedList<Boolean>();
		LinkedList<Boolean> isAudioOn = new LinkedList<Boolean>();
		for (int track=0; track<blocks.getNumTracks(); track++) {
			isVisualOn.add(new Boolean(controller.getSettings().isVisualOn(track)));
			isAudioOn.add(new Boolean(controller.getSettings().isAudioOn(track)));
		}
		
		Boolean isMarkKeysOn = controller.getSettings().isMarkKeysOn();
		
		blocks.setLastVisibleTick(controller.glToTick(controller.getCamera().getVisibleArea()));

		
		// not useable notes
		for (int key=0; key<firstKey; key++) {
			for (int track=0; track<blocks.getNumTracks(); track++) {
				if (isVisualOn.get(track)) {
					blocks.showBlocks(key, track, false, false, false, false);
				}
			}
		}
		
		// useable notes
		for (int key=firstKey; key<=lastKey; key++) {
			for (int track=0; track<blocks.getNumTracks(); track++) {
				if (isVisualOn.get(track)) {
					blocks.showBlocks(key, track, true, isAudioOn.get(track), isVisualOn.get(track), isMarkKeysOn);
				}
			}
		}
		
		// not usable notes
		for (int key=lastKey+1; key<blocks.getNumKeys(); key++) {
			for (int track=0; track<blocks.getNumTracks(); track++) {
				if (isVisualOn.get(track)) {
					blocks.showBlocks(key, track, false, false, false, false);
				}
			}
		}
		
		showingNotes = false;
		notify();
	}
}

