/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.piano;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.concurrent.ConcurrentLinkedQueue;
import javax.media.opengl.GL;

import openpiano.control.Controller;
import openpiano.gl.GLModel;
import openpiano.settings.Constants;

/**
 * The Key class for keys in a piano
 *
 * @author  Sascha Westermann
 * @version 1.0, 07/30/08
 */
public class Key {
	private ClassLoader cl;
	private Controller controller;
	private Boolean flattone = false;
	private GL gl;
	private ConcurrentLinkedQueue<Boolean> keyQueue;
	private float limit, rotation = 0.0f;
	private Boolean mark = false;
	private float markKeyFactor = 0.0f;
	private GLModel model;
	private float[] pos;
	private int color;

	/**
	 * Constructs a key in a piano
	 *
     * @param controller the main controller of the application
	 * @param gl the relevant GL Context
	 * @param modeldir the directory containing model files
	 * @param modelname the name of the model
	 * @param distance absolute position on x-axis of the piano
	 */
	Key(Controller controller, GL gl, String modeldir, String modelname, float distance) {
		this.gl = gl;
		this.controller = controller;
		pos = new float[3];
		pos[0] = distance; pos[1] = 0.0f; pos[2] = 0.0f;
		keyQueue = new ConcurrentLinkedQueue<Boolean>();
        cl = this.getClass().getClassLoader();

		// flat tone or full tone
		if (modelname == "flatnote") {
			flattone = true;
			limit = 4.0f;
		}
		else
			limit = 6.0f;
		
		try {
			InputStream is = cl.getResourceAsStream("models/" + modelname + ".obj");
			BufferedReader br = new BufferedReader(new InputStreamReader(is));
			model = new GLModel(br, gl);
			br.close();
			is.close();
		} catch(Exception e) {}
	}

	/**
	 * Draws the model of a key
	 *
     * @param useable <tt>true</tt> if key is playable
	 */
	private void drawKeyModel(Boolean useable) {
		gl.glEnable(GL.GL_BLEND);
		gl.glEnable(GL.GL_COLOR_MATERIAL);
		gl.glBlendFunc (GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		
		if (flattone) {
			if (useable) {
				if (mark)
					gl.glColor4f(1.0f,0.2f,0.2f,0.7f);
				else if (controller.getSettings().isMarkKeysOn() && markKeyFactor > 0.0f) {
					float r = Constants.COLORSTART[color*3 + 0];
					float g = Constants.COLORSTART[color*3 + 1];
					float b = Constants.COLORSTART[color*3 + 2];
					float radd = Constants.COLOREND[color*3 + 0];
					float gadd = Constants.COLOREND[color*3 + 1];
					float badd = Constants.COLOREND[color*3 + 2];
					gl.glColor4f(r-((r-radd)*markKeyFactor), g-((g-gadd)*markKeyFactor), b-((b-badd)*markKeyFactor), 1.0f);
				}
				else
					gl.glColor4f(0.0f,0.0f,0.0f,0.8f);
			}
			else
				gl.glColor4f(0.0f,0.0f,0.0f,0.3f);
		}
		else {
			if (useable)
				if (mark)
					gl.glColor4f(1.0f,0.2f,0.2f,0.7f);
				else if (controller.getSettings().isMarkKeysOn() && markKeyFactor > 0.0f) {
					float r = Constants.COLORSTART[color*3 + 0];
					float g = Constants.COLORSTART[color*3 + 1];
					float b = Constants.COLORSTART[color*3 + 2];
					float radd = Constants.COLOREND[color*3 + 0];
					float gadd = Constants.COLOREND[color*3 + 1];
					float badd = Constants.COLOREND[color*3 + 2];
					gl.glColor4f(r-((r-radd)*markKeyFactor), g-((g-gadd)*markKeyFactor), b-((b-badd)*markKeyFactor), 1.0f);
				}
				else
					gl.glColor4f(1.0f,1.0f,1.0f,1.0f);
			else
				gl.glColor4f(1.0f,1.0f,1.0f,0.3f);
		}

		model.draw(gl);
		gl.glDisable(GL.GL_COLOR_MATERIAL);
		gl.glDisable(GL.GL_BLEND);
	}

	/**
	 * Returns <tt>true</tt> if a key is marked
	 * 
	 * @return <tt>true</tt> if a key is marked
	 */
	public Boolean isKeyMarked() { return mark; }
	
	/**
	 * Marks a key
	 * 
	 * @param mark <tt>true</tt> if a key should be marked
	 */
	public void markKey(Boolean mark) { this.mark = mark; }

	/**
	 * Presses a key
	 */
	public void pressKey() { keyQueue.add(true); }
	
	/**
	 * Releases a key
	 */
	public void releaseKey() { keyQueue.add(false); }

	/**
	 * Sets the factor of the color depth of the key (0.0-1.0)
	 * 
	 * @param factor the factor of the color depth
	 */
	public void setMarkKeyFactor(int color, float factor) {
		this.color = color;
		markKeyFactor = factor;
	}

	/**
	 * Sets the factor of the color depth of the key (0.0-1.0)
	 * 
	 * @param factor the factor of the color depth
	 */
	public void setMarkKeyOff() {
		markKeyFactor = 0.0f;
	}

	/**
	 * Draws the key
	 *
     * @param useable <tt>true</tt> if key is playable
	 */
	public void showKey(Boolean useable) {
		gl.glPushMatrix();

		// cleans up the queue, if there are too much jobs left
		// leaves the last element
		int max = 20;
		if (keyQueue.size() > max) {
			for (int i=0; i<keyQueue.size()-1; i++)
				keyQueue.poll();
		}
		
		Boolean keyPressed;
		if ( (keyPressed = keyQueue.poll()) != null) {
			if (keyPressed)
				rotation=limit;
			else
				rotation=0.0f;
		}
		
		gl.glRotatef(rotation, 1.0f, 0.0f, 0.0f);
		gl.glTranslatef(pos[0], pos[1], pos[2]);
		drawKeyModel(useable);
		gl.glPopMatrix();
	}
}
