/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.piano;

import javax.media.opengl.GL;
import openpiano.control.Controller;
import openpiano.settings.Constants;

/**
 * The Piano class loads a 3D-piano and handles the animations of keys
 *
 * @author  Sascha Westermann
 * @version 1.0, 07/30/08
 */

public class Piano {
	private Controller controller;
	private Key[] key;
	private final String MODEL_END = "Cend";
	private final String[] MODEL_OCTAVE = { "C", "flatnote", "D", "flatnote", "E", "F", 
		"flatnote", "G", "flatnote", "A", "flatnote", "H", };
	private final String[] MODEL_START = { "Astart", "flatnote", "H" };
	private final String modeldir = "models";

	/**
	 * Constructs a 3D-piano with 88 keys
	 *
     * @param controller the main controller of the application
	 * @param gl the relevant GL context
	 */
	public Piano(Controller controller, GL gl) {
		this.controller = controller;
		
		// create 88 keys 
		key = new Key[88];

		// create A0 to H0
		for (int i=0; i<3; i++)
			key[i] = new Key(controller, gl, modeldir, MODEL_START[i], Constants.POS_START[i]);

		// create 7 octaves
		for (int i=3; i<87; i++)
			key[i] = new Key(controller, gl, modeldir, MODEL_OCTAVE[(i-3)%12], (int)((i-3)/12)*Constants.OCTAVELENGTH + Constants.POS_OCTAVE[(i-3)%12]);

		// create C8
		key[87] = new Key(controller, gl, modeldir, MODEL_END, 7*Constants.OCTAVELENGTH + Constants.POS_END);
	}

	/**
	 * Return <tt>true</tt> if the key is marked
	 *
     * @param key the affected key
     * @return <tt>true</tt> if key is marked
	 */
	public Boolean isKeyMarked(int key) {
		return this.key[key].isKeyMarked();
	}
	
	/**
	 * Release all keys
	 */
	public void releaseAllKeys() {
		for (Key k : key)
			k.releaseKey();
	}

	/**
	 * Sends played notes to the piano and shows them
	 *
     * @param key the affected key
	 * @param velocity the velocity of the key, a velocity of 0 means release key
	 * @param autoPlayed <tt>true</tt> if the application, <tt>false</tt> if the user activated the key
	 */
	public void sendKey(int key, int velocity, Boolean autoPlayed) {
		if (velocity > 0) {
			if (!autoPlayed)
				this.key[key].markKey(true);
			this.key[key].pressKey();
		}
		else {
			if (!autoPlayed)
				this.key[key].markKey(false);
			this.key[key].releaseKey();
		}
	}

	/**
	 * Sets the depth of the color of a key
	 *
     * @param key the affected key
     * @param color the color of the affected key
	 * @param factor the depth factor of the color (0.0 - 1.0)
	 */
	public void setKeyMarker(int key, int color, float factor) {
		this.key[key].setMarkKeyFactor(color, factor);
	}
	
	/**
	 * Sets the key's color off
	 *
     * @param key the affected key
	 */
	public void setKeyMarkerOff(int key) {
		this.key[key].setMarkKeyOff();
	}

	/**
	 * Draws the piano
	 */
	public void showPiano() {
		for (int i = 0; i < controller.getSettings().getFirstKey(); i++)
			key[i].showKey(false);
		for (int i = controller.getSettings().getFirstKey(); i <= controller.getSettings().getLastKey(); i++)
			key[i].showKey(true);
		for (int i = controller.getSettings().getLastKey()+1; i < 88; i++)
			key[i].showKey(false);
	}
}