/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.settings;

import java.util.LinkedList;

/**
 * Class for global MIDI file settings
 *
 * @author  Sascha Westermann
 * @version 1.0, 07/30/08
 */
public class MidiSettings {
	private String md5Hash;
	private int keyboardTrack;
	private int bpmFactor;
	private Boolean awaitNotes, markKeys;
	private LinkedList<MidiTrackSettings> midiTrackSettings;
	
	/**
	 * Constructs a new MIDI file setting
	 * @param md5Hash the hash sum to identify the MIDI file
	 */
	public MidiSettings(String md5Hash) {
		this.md5Hash = md5Hash;
		midiTrackSettings = new LinkedList<MidiTrackSettings>();
	}
	
	/**
	 * Constructs a new MIDI file setting
	 *
	 * @param md5Hash the hash sum to identify the MIDI file
	 * @param keyboardTrack the track played by the keyboard
	 * @param bpmFactor the speed factor for beats per minute
	 * @param awaitNotes <tt>true</tt> if await notes function is selected
	 * @param markKeys <tt>true</tt> if mark keys function is selected
	 */
	public MidiSettings(String md5Hash, int keyboardTrack, int bpmFactor, Boolean awaitNotes, Boolean markKeys) {
		this.md5Hash = md5Hash;
		this.keyboardTrack = keyboardTrack;
		this.bpmFactor = bpmFactor;
		this.awaitNotes = awaitNotes;
		this.markKeys = markKeys;
		midiTrackSettings = new LinkedList<MidiTrackSettings>();
	}
	
	/**
	 * Returns the hash sum of the MIDI file
	 * 
	 * @return the hash sum of the MIDI file
	 */
	public String getMD5Hash() { return md5Hash; }
	
	/**
	 * Returns the track played by the keyboard
	 *
	 * @return the track played by the keyboard
	 */
	public int getKeyboardTrack() { return keyboardTrack; }
	
	/**
	 * Returns the speed factor for beats per minute
	 * 
	 * @return the speed factor for beats per minute
	 */
	public int getBeatsPerMinuteFactor() { return bpmFactor; }
	
	/**
	 * Return <tt>true</tt> if the await notes function is selected
	 * 
	 * @return <tt>true</tt> if the await notes function is selected
	 */
	public Boolean isAwaitNotesOn() { return awaitNotes; }

	/**
	 * Return <tt>true</tt> if the mark keys function is selected
	 * 
	 * @return <tt>true</tt> if the mark keys function is selected
	 */
	public Boolean isMarkKeysOn() { return markKeys; }
	
	
	/**
	 * Sets the hash sum to identify the MIDI file
	 * 
	 * @param md5Hash the hash sum to identify the MIDI file
	 */
	public void setMD5Hash(String md5Hash) { this.md5Hash = md5Hash; }
	 
	/**
	 * Sets the track played by the keyboard
	 * 
	 * @param keyboardTrack the track played by the keyboard
	 */
	public void setKeyboardTrack(int keyboardTrack) { this.keyboardTrack = keyboardTrack; }
	
	/**
	 * Sets the speed factor for beats per minute
	 * 
	 * @param bpmFactor the speed factor for beats per minute
	 */
	public void setBeatsPerMinuteFactor(int bpmFactor) { this.bpmFactor = bpmFactor; }
	
	/**
	 * Sets the status of the await notes function
	 * 
	 * @param awaitNotes <tt>true</tt> if the await notes function should be enabled
	 */
	public void setAwaitNotesOn(Boolean awaitNotes) { this.awaitNotes = awaitNotes; }

	/**
	 * Sets the status of the mark keys function
	 * 
	 * @param markKeys <tt>true</tt> if the mark keys function should be enabled
	 */
	public void setMarkKeysOn(Boolean markKeys) { this.markKeys = markKeys; }
	
	/**
	 * Adds custom settings for a track of a MIDI file
	 * 
	 * @param midiTrackSetting the settings for the track
	 */
	public void addMidiTrackSetting(MidiTrackSettings midiTrackSetting) { midiTrackSettings.add(midiTrackSetting); }
	
	/**
	 * Returns a list of all track settings
	 * 
	 * @return the list of all track settings
	 */
	public LinkedList<MidiTrackSettings> getMidiTrackSettings() { return midiTrackSettings; }
}