/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.settings;

/**
 * Detailed settings for a single track of a MIDI file
 *
 * @author  Sascha Westermann
 * @version 1.0, 07/30/08
 */
public class MidiTrackSettings {
	private String trackName;
	private int instrument, volume;
	private int color = -1;
	private Boolean audio, visual;
	
	/**
	 * Constructs a new track setting
	 * 
	 * @param trackName the name of the track
	 */
	public MidiTrackSettings(String trackName) {
		this.trackName = trackName;
	}
	
	/**
	 * Constructs a new track setting
	 * 
	 * @param trackName the name of the track
	 * @param instrument the instrument used for the track
	 * @param volume the volume of the track
	 * @param audio <tt>true</tt> if audio should be played
	 * @param visual <tt>true</tt> if track should be showed
	 */
	public MidiTrackSettings(String trackName, int instrument, int volume, Boolean audio, Boolean visual) {
		this.trackName = trackName;
		this.instrument = instrument;
		this.volume = volume;
		this.audio = audio;
		this.visual = visual;
	}
	
	/**
	 * Returns the color of the track
	 * 
	 * @return the color of the track
	 */
	public int getColor() { return color; }

	/**
	 * Returns the name of the track
	 * 
	 * @return the name of the track
	 */
	public String getTrackName() { return trackName; }
	
	/**
	 * Return the instrument used for the track
	 *
	 * @return the instrument used for the track
	 */
	public int getInstrument() { return instrument; }
	
	/**
	 * Returns the volume of the track
	 * 
	 * @return the volume of the track
	 */
	public int getVolume() { return volume; }
	
	/**
	 * Returns <tt>true</tt> if audio function is selected
	 * 
	 * @return <tt>true</tt> if audio function is selected
	 */
	public Boolean isAudioOn() { return audio; }

	/**
	 * Returns <tt>true</tt> if visual function is selected
	 * 
	 * @return <tt>true</tt> if visual function is selected
	 */
	public Boolean isVisualOn() { return visual; }
	
	/**
	 * Sets the color of the track
	 * 
 	 * @param color the color of the track
	 */
	public void setColor(int color) { this.color = color; }

	/**
	 * Sets the name of the track
	 * 
 	 * @param trackName the name of the track
	 */
	public void setTrackName(String trackName) { this.trackName = trackName; }
	
	/**
	 * Sets the instrument used for the track
	 * 
	 * @param instrument the instrument used for the track
	 */
	public void setInstrument(int instrument) { this.instrument = instrument; }
	
	/**
	 * Sets the volume of the track
	 * 
	 * @param volume the volume of the track
	 */
	public void setVolume(int volume) { this.volume = volume; }
	
	/**
	 * Sets the status of the audio function
	 * 
	 * @param audio <tt>true</tt> if audio should be played
	 */
	public void setAudioOn(Boolean audio) { this.audio = audio; }
	
	/**
	 * Sets the status of the visual function
	 * 
	 * @param visual <tt>true</tt> if track should be showed
	 */
	public void setVisualOn(Boolean visual) { this.visual = visual; }
}