/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.settings;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.BufferedReader;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.LinkedList;
import java.util.Scanner;
import javax.sound.midi.MidiDevice;
import javax.sound.midi.MidiSystem;
import openpiano.control.Controller;
import openpiano.midi.MidiInput;
import openpiano.midi.MidiOutput;

/**
 * A class managing the user settings
 *
 * @author  Sascha Westermann
 * @version 1.0, 07/30/08
 */
public class Settings {
	private int windowWidth = 800;
	private int windowHeight = 550;
	private int windowPosX = 0;
	private int windowPosY = 0;
	private int firstKey = 0;
	private int lastKey = 87;
	private float angle = 70.0f;
	private float zoomFactor = 1.0f;
	private int midiKeyboardIndex, midiOutputIndex;
	private MidiOutput midiOutput;
	private MidiInput midiInput;
	private File confFile;
	private File fileDir;
	private LinkedList<MidiSettings> midiSettings;
	private Controller controller;
	private String actualMidiInputDevice = "not set";
	private String midiInputDevice = "not set";
	private String actualMidiOutputDevice = "not set";
	private String midiOutputDevice = "not set";
	private Boolean downloadSoundbank = true;
	
	/**
	 * Creates a settings object and reads settings from a file
	 *
	 * @param controller the main controller of the application
	 */
	public Settings(Controller controller) {
		this.controller = controller;
		midiSettings = new LinkedList<MidiSettings>();
		String home = System.getProperty("user.home");
		String separator = System.getProperty("file.separator");
		File appDir = new File(home + separator + ".OpenPiano");
		fileDir = appDir;
		
		confFile = new File(appDir + separator + "app.conf");
		if (!appDir.exists())
			appDir.mkdir();
		try {
			BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(confFile)));
			String line;
			while ((line = br.readLine()) != null) {
				Scanner scanner = new Scanner(line);
				String option = scanner.next();
				String value = "";
				try {
					value = scanner.nextLine().split("\"")[1];
				} catch (Exception e) {}
			
				if (option.equals("downloadSoundbank"))
					downloadSoundbank = new Boolean(value);
				else if (option.equals("fileDir"))
					fileDir = new File(value);
				else if (option.equals("firstKey"))
					firstKey = new Integer(value);
				else if (option.equals("lastKey"))
					lastKey = new Integer(value);
				else if (option.equals("midiInput"))
					midiInputDevice = value;
				else if (option.equals("midiOutput"))
					midiOutputDevice = value;
				
				else if (option.equals("angle"))
					angle = new Float(value);
				else if (option.equals("zoomFactor"))
					zoomFactor = new Float(value);
				else if (option.equals("windowWidth"))
					windowWidth = new Integer(value);
				else if (option.equals("windowHeight"))
					windowHeight = new Integer(value);
				else if (option.equals("windowPosX"))
					windowPosX = new Integer(value);
				else if (option.equals("windowPosY"))
					windowPosY = new Integer(value);
				
				
				
				else if (option.equals("MD5"))
					midiSettings.add(new MidiSettings(value));
				else if (option.equals("BPM"))
					midiSettings.getLast().setBeatsPerMinuteFactor(new Integer(value));
				else if (option.equals("keyboardTrack"))
					midiSettings.getLast().setKeyboardTrack(new Integer(value));
				else if (option.equals("awaitNotes"))
					midiSettings.getLast().setAwaitNotesOn(new Boolean(value));
				else if (option.equals("markKeys"))
					midiSettings.getLast().setMarkKeysOn(new Boolean(value));
				else if (option.equals("trackName"))
					midiSettings.getLast().addMidiTrackSetting(new MidiTrackSettings(value));
				else if (option.equals("trackInstrument"))
					midiSettings.getLast().getMidiTrackSettings().getLast().setInstrument(new Integer(value));
				else if (option.equals("trackVolume"))
					midiSettings.getLast().getMidiTrackSettings().getLast().setVolume(new Integer(value));
				else if (option.equals("trackAudioOn"))
					midiSettings.getLast().getMidiTrackSettings().getLast().setAudioOn(new Boolean(value));
				else if (option.equals("trackVisualOn"))
					midiSettings.getLast().getMidiTrackSettings().getLast().setVisualOn(new Boolean(value));
				else if (option.equals("trackColor"))
					midiSettings.getLast().getMidiTrackSettings().getLast().setColor(new Integer(value));
			}
			br.close();
		} catch (Exception e) {}

		if (downloadSoundbank) {
			SoundbankCheck soundbankCheck = new SoundbankCheck(controller);
			if (soundbankCheck.hasFailed())
				downloadSoundbank = soundbankCheck.isRemindOn();
			else
				downloadSoundbank = false;
		}
	}
	
	/**
	 * Looks the configuration up for a previously saved hash
	 * 
	 * @param md5Hash the MD5 hash to look for
	 * @return the index of the known hash value, else <tt>-1</tt>
	 */
	public int lookupForFile(String md5Hash) {
		for (int i=0; i<midiSettings.size(); i++) {
			if (midiSettings.get(i).getMD5Hash().equals(md5Hash))
				return i;
		}
		return -1;
	}
	
	/**
	 * Returns the status of the mark keys function of the current loaded configuration
	 * 
	 * @return <tt>true</tt> if mark keys function is enabled
	 */
	public Boolean isMarkKeysOn() { 
		if (controller.getMidiLoader() != null)
			return midiSettings.get(controller.getMidiLoader().isFileKnown()).isMarkKeysOn(); 
		else
			return false;
	}
	
	/**
	 * Returns the status of the await notes function of the current loaded configuration
	 * 
	 * @return <tt>true</tt> if await notes function is enabled
	 */
	public Boolean isAwaitNotesOn() { 
		if (controller.getMidiLoader() != null)
			return midiSettings.get(controller.getMidiLoader().isFileKnown()).isAwaitNotesOn(); 
		else
			return false;
	}
	
	/**
	 * Returns the status of the audio function of a specified track of the current loaded configuration
	 * 
	 * @param track the relevant track
	 * @return <tt>true</tt> if audio function is enabled on the track
	 */
	public Boolean isAudioOn(int track) { 
		return midiSettings.get(controller.getMidiLoader().isFileKnown()).getMidiTrackSettings().get(track).isAudioOn(); 
	}

	/**
	 * Returns the status of the visual function of a specified track of the current loaded configuration
	 * 
	 * @param track the relevant track
	 * @return <tt>true</tt> if visual function is enabled on the track
	 */
	public Boolean isVisualOn(int track) { 
		return midiSettings.get(controller.getMidiLoader().isFileKnown()).getMidiTrackSettings().get(track).isVisualOn(); 
	}
	
	/**
	 * Returns the first key of the shown piano
	 * 
	 * @return the first key of the shown piano
	 */
	public int getFirstKey() { return firstKey; }

	/**
	 * Returns the last key of the shown piano
	 * 
	 * @return the last key of the shown piano
	 */
	public int getLastKey() { return lastKey; }

	/**
	 * Sets the first key of the shown piano
	 * 
	 * @param firstKey the first key of the shown piano
	 */
	public void setFirstKey(int firstKey) { this.firstKey = firstKey; }

	/**
	 * Sets the last key of the shown piano
	 * 
	 * @param lastKey the last key of the shown piano
	 */
	public void setLastKey(int lastKey) { this.lastKey = lastKey; }

	/**
	 * Returns the index of the MIDI input device
	 * 
	 * @return the index of the MIDI input device
	 */
	public int getMidiKeyboardIndex() { return midiKeyboardIndex; }

	/**
	 * Returns the index of the MIDI output device
	 * 
	 * @return the index of the MIDI output device
	 */
	public int getMidiOutputIndex() { return midiOutputIndex; }
	
	/**
	 * Returns the directory of the MIDI files
	 * 
	 * @return the directory of the MIDI files
	 */
	public File getFileDir() { return fileDir; }

	/**
	 * Sets the directory of the MIDI files
	 * 
	 * @param fileDir the directory of the MIDI files
	 */
	public void setFileDir(File fileDir) { this.fileDir = fileDir; }
	
	/**
	 * Returns the available MIDI ports
	 * 
	 * @param input <tt>true</tt> if MIDI input ports should be returned, <tt>false</tt> for MIDI output ports
	 * @return the available MIDI ports
	 */
	public MidiDevice.Info[] getMidiPorts(Boolean input) {
		LinkedList<MidiDevice.Info> list = new LinkedList<MidiDevice.Info>();
		try {
			for (int i=0; i<MidiSystem.getMidiDeviceInfo().length; i++) {
				MidiDevice device = MidiSystem.getMidiDevice((MidiSystem.getMidiDeviceInfo())[i]);

				if ( (input && (device.getMaxTransmitters() != 0))
					|| (!input && (device.getMaxReceivers() != 0)) )
						list.add((MidiSystem.getMidiDeviceInfo())[i]);
			}
		} catch (Exception e) {}

		String temp = null;
		for (int i=0; i<list.size(); i++) {
			if (list.get(i).getName().equals(temp)) {
				list.remove(i);
				i--;
			}
			temp = list.get(i).getName();
		}
		
		MidiDevice.Info[] devices = new MidiDevice.Info[list.size()];

		for (int i=0; i<list.size(); i++)
			devices[i] = list.get(i);

		return devices;
	}
	
	/**
	 * Adds a new configuration for a MIDI file
	 * 
	 * @param midiSettings the settings of a MIDI file to add
	 * @return the index of the configuration
	 */
	public int addMidiSettings(MidiSettings midiSettings) {
		this.midiSettings.add(midiSettings);
		return this.midiSettings.size()-1;
	}
	
	/**
	 * Saves the settings to the configuration file
	 */
	public void saveSettings() {
		try {
			BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(confFile)));
			LinkedList<String> options = new LinkedList<String>();
			options.add(new String("downloadSoundbank = \"" + downloadSoundbank + "\""));
			options.add(new String("firstKey = \"" + firstKey + "\""));
			options.add(new String("lastKey = \"" + lastKey + "\""));
			options.add(new String("fileDir = \"" + fileDir.toString() + "\""));
			options.add(new String("midiInput = \"" + midiInputDevice + "\""));
			options.add(new String("midiOutput = \"" + midiOutputDevice + "\""));
			options.add(new String("angle = \"" + angle + "\""));
			options.add(new String("zoomFactor = \"" + zoomFactor + "\""));
			options.add(new String("windowWidth = \"" + windowWidth + "\""));
			options.add(new String("windowHeight = \"" + windowHeight + "\""));
			options.add(new String("windowPosX = \"" + windowPosX + "\""));
			options.add(new String("windowPosY = \"" + windowPosY + "\""));
			for (MidiSettings m : midiSettings) {
				options.add(new String("MD5 = \"" + m.getMD5Hash() + "\""));
				options.add(new String("BPM = \"" + m.getBeatsPerMinuteFactor() + "\""));
				options.add(new String("keyboardTrack = \"" + m.getKeyboardTrack() + "\""));
				options.add(new String("awaitNotes = \"" + m.isAwaitNotesOn() + "\""));
				options.add(new String("markKeys = \"" + m.isMarkKeysOn() + "\""));
				for (int i=0; i<m.getMidiTrackSettings().size(); i++) {
					options.add(new String("trackName = \"" + m.getMidiTrackSettings().get(i).getTrackName() + "\""));
					options.add(new String("trackInstrument = \"" + m.getMidiTrackSettings().get(i).getInstrument() + "\""));
					options.add(new String("trackVolume = \"" + m.getMidiTrackSettings().get(i).getVolume() + "\""));
					options.add(new String("trackAudioOn = \"" + m.getMidiTrackSettings().get(i).isAudioOn() + "\""));
					options.add(new String("trackVisualOn = \"" + m.getMidiTrackSettings().get(i).isVisualOn() + "\""));
					if (m.getMidiTrackSettings().get(i).isVisualOn())
						options.add(new String("trackColor = \"" + m.getMidiTrackSettings().get(i).getColor() + "\""));
				}
			}
			
			for (String s : options) {
				bw.write(s);
				bw.write("\n");
				bw.flush();
			}
			bw.close();
		} catch (Exception e) {};
	}

	/**
	 * Returns the name of the actual MIDI output device
	 * @return the name of the actual MIDI output device
	 */
	public String getActualMidiOutputDevice() { return actualMidiOutputDevice; }
	
	/**
	 * Returns the name of the MIDI input device according to the configuration
	 * @return the name of the MIDI input device according to the configuration
	 */
	public String getMidiInputDevice() { return midiInputDevice; }

	/**
	 * Returns the name of the MIDI output device according to the configuration
	 * @return the name of the MIDI output device according to the configuration
	 */
	public String getMidiOutputDevice() { return midiOutputDevice; }
	
	//TODO: wtf? darunter läuft eine ganze menge schief, controller und settingsgui verwenden verschiedene geräte
	
	/**
	 * Sets the actual MIDI input device
	 * 
	 * @param midiInputDevice the actual MIDI input device
	 */
	public void setActualMidiInputDevice(String midiInputDevice) { actualMidiInputDevice = midiInputDevice; }

	/**
	 * Sets the actual MIDI output device
	 * 
	 * @param midiOutputDevice the actual MIDI output device
	 */
	public void setActualMidiOutputDevice(String midiOutputDevice) { actualMidiOutputDevice = midiOutputDevice; }
	
	/**
	 * Sets the MIDI output device
	 * 
	 * @param midiOutputDevice the MIDI output device
	 */
	public void setMidiOutputDevice(String midiOutputDevice) { this.midiOutputDevice = midiOutputDevice; }
	
	/**
	 * Returns the configuration for a MIDI file
	 * 
	 * @param index the index of the configuration to return
	 * @return the configuration for a MIDI file
	 */
	public MidiSettings getMidiSettings(int index) { return midiSettings.get(index); }
	
	/**
	 * Returns the configurations for the MIDI files
	 * 
	 * @return the configuration for the MIDI files
	 */
	public LinkedList<MidiSettings> getMidiSettings() { return midiSettings; }
	
	/**
	 * Creates a MIDI output device
	 * 
	 * @param midiOutputDevice the MIDI output device
	 */
	private void createMidiOutput(MidiDevice.Info midiOutputDevice) {
		midiOutput = new MidiOutput(midiOutputDevice);
    }
	
	/**
	 * Returns the MIDI transmitter
	 * 
	 * @return the MIDI transmitter
	 */
	public MidiOutput getMidiOutput() { return midiOutput;	}
	
	/**
	 * Handles the change of MIDI output devices
	 *  
	 * @param midiDevice the new MIDI output device
	 * @param saveSettings <tt>true</tt> if new device should be saved
	 */
	public void handleMidiOutputChange(MidiDevice.Info midiDevice, Boolean saveSettings) {
		String actualOutputDevice = getActualMidiOutputDevice();
		
		if (saveSettings)
			setMidiOutputDevice(midiDevice.getName());

		if (!actualOutputDevice.equals(midiDevice.getName()) || midiOutput == null) {
			if (controller.getMidiLoader() != null)
				controller.getMidiLoader().getMidiTimer().stopTimer();
	
			if (midiOutput != null && midiOutput.getReceiver() != null)
				midiOutput.close();
	
			createMidiOutput(midiDevice);
	
			if (controller.getMidiLoader() != null)
				controller.resetMidiMessages();
			
			if (controller.isPlayModeSet())
				controller.getMidiLoader().getMidiTimer().startTimer();
		}
	}

	/**
	 * Handles the change of MIDI output devices
	 *  
	 * @param midiDevice the new MIDI output device
	 */
	public void handleMidiOutputChange(MidiDevice.Info midiDevice) {
		handleMidiOutputChange(midiDevice, true);
	}
	
	/**
	 * Returns the MIDI receiver
	 * 
	 * @return the MIDI receiver
	 */
	public MidiInput getMidiInput() { return midiInput; }
	
	/**
	 * Handles the change of MIDI input devices
	 *  
	 * @param midiDevice the new MIDI input device
	 */
	public void handleMidiInputChange(MidiDevice.Info midiDevice) {
		handleMidiInputChange(midiDevice, true, false);
	}
	
	/**
	 * Handles the change of MIDI input devices
	 *  
	 * @param midiDevice the new MIDI input device
	 * @param saveSettings <tt>true</tt> if new device should be saved
	 */
	public void handleMidiInputChange(MidiDevice.Info midiDevice, Boolean saveSettings) {
		handleMidiInputChange(midiDevice, saveSettings, false);
	} 
	
	/**
	 * Handles the change of MIDI input devices
	 *  
	 * @param midiDevice the new MIDI input device
	 * @param saveSettings <tt>true</tt> if new device should be saved
	 * @param returnFromOptions <tt>true</tt> if called from keyboard size settings
	 */
	public void handleMidiInputChange(MidiDevice.Info midiDevice, Boolean saveSettings, Boolean returnFromOptions) {
		String actualInputDevice = actualMidiInputDevice;
		
		if (saveSettings)
			midiInputDevice = midiDevice.getName();

		if (!actualInputDevice.equals(midiDevice.getName()) || midiInput == null || returnFromOptions) {
			if (midiInput != null)
				midiInput.close();
			midiInput = new MidiInput(controller, midiDevice);
		}
	}
	/**
	 * Sets the height of the application window
	 * 
	 * @param windowHeight the height of the application window
	 */
	public void setWindowHeight(int windowHeight) { this.windowHeight = windowHeight; }

	/**
	 * Sets the width of the application window
	 * 
	 * @param windowWidth the widths of the application window
	 */
	public void setWindowWidth(int windowWidth) { this.windowWidth = windowWidth; }
	
	/**
	 * Returns the height of the application window
	 * 
	 * @return the height of the application window
	 */
	public int getWindowHeight() { return windowHeight; }
	
	/**
	 * Returns the width of the application window
	 * 
	 * @return the width of the application window
	 */
	public int getWindowWidth() { return windowWidth; }
	
	/**
	 * Sets the x-position of the application window
	 * 
	 * @param windowPosX the x-position of the application window
	 */
	public void setWindowPosX(int windowPosX) { this.windowPosX = windowPosX; }

	/**
	 * Sets the y-position of the application window
	 * 
	 * @param windowPosY the y-position of the application window
	 */
	public void setWindowPosY(int windowPosY) { this.windowPosY = windowPosY; }
	
	/**
	 * Returns the x-position of the application window
	 * 
	 * @return the x-position of the application window
	 */
	public int getWindowPosX() { return windowPosX; }

	/**
	 * Returns the y-position of the application window
	 * 
	 * @return the y-position of the application window
	 */
	public int getWindowPosY() { return windowPosY; }
	
	/**
	 * Sets the angle of the piano
	 * 
	 * @param angle the angle of the piano
	 */
	public void setAngle(float angle) { this.angle = angle; }

	/**
	 * Returns the angle of the piano
	 * 
	 * @return the angle of the piano
	 */
	public float getAngle() { return angle; }
	
	/**
	 * Sets the zoom factor of the piano
	 * 
	 * @param zoomFactor the zoom factor of the piano
	 */
	public void setZoomFactor(float zoomFactor) { this.zoomFactor = zoomFactor; }
	
	/**
	 * Returns the zoom factor of the piano
	 * 
	 * @return the zoom factor of the piano
	 */
	public float getZoomFactor() { return zoomFactor; }
}
