/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.settings;

import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import javax.sound.midi.MidiDevice;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.ShortMessage;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.UIManager;
import openpiano.control.Controller;
import openpiano.midi.MidiInput;

/**
 * The SettingsGUI class creates a graphical frontend for the settings 
 *
 * @author  Sascha Westermann
 * @version 1.0, 07/30/08
 */

public class SettingsGUI extends JPanel {
	private static final long serialVersionUID = -6191203592281530330L;
	private JComboBox inPortsBox, outPortsBox;
	private MidiDevice.Info[] inDevices, outDevices;
	private final Controller controller;
	private final Settings settings;
	private JDialog midiKeyboardSizeDialog;
	
	/**
	 * Constructs a new graphical frontend for the settings
	 * 
	 * @param controllr the main controller of the application
	 * @param settings the reference to the settings
	 */
	public SettingsGUI(Controller finalController, Settings finalSettings) {
		this.controller = finalController;
		this.settings = finalSettings;
		
		inDevices = settings.getMidiPorts(true);
		outDevices = settings.getMidiPorts(false);
		
		String[] inNames = new String[inDevices.length];
		for (int i=0; i<inDevices.length; i++) {
			String[] desc = inDevices[i].getDescription().split(", ");
			if (desc.length > 1)
				inNames[i] = inDevices[i].getName() + " (" + desc[desc.length-1] + ")";
			else
				inNames[i] = inDevices[i].getName();
		}
		
		String[] outNames = new String[outDevices.length];
		for (int i=0; i<outDevices.length; i++) {
			String[] desc = outDevices[i].getDescription().split(", ");
			if (desc.length > 1)
				outNames[i] = outDevices[i].getName() + " (" + desc[desc.length-1] + ")";
			else
				outNames[i] = outDevices[i].getName();
		}

		setLayout(new GridBagLayout());
		GridBagConstraints gbc = new GridBagConstraints();
		
		JLabel inPortsLabel = new JLabel("MIDI Keyboard");
		inPortsBox = new JComboBox(inNames);
		int foundInDevice = getIndexByName(inDevices, settings.getMidiInputDevice());
		if (foundInDevice != -1) {
			settings.setActualMidiInputDevice(inNames[foundInDevice]);
			inPortsBox.setSelectedIndex(foundInDevice);
			settings.handleMidiInputChange(inDevices[inPortsBox.getSelectedIndex()]);
		}
		else {
			settings.setActualMidiInputDevice(inNames[inNames.length-1]);
			inPortsBox.setSelectedIndex(inNames.length - 1);
			settings.handleMidiInputChange(inDevices[inPortsBox.getSelectedIndex()], false);
		}
		inPortsBox.addItemListener(new ItemListener() {
			public void itemStateChanged( ItemEvent e ) {
				  Object comp = e.getSource();
				  if (e.getStateChange() == ItemEvent.SELECTED) {
					  MidiDevice.Info inDevice = inDevices[((JComboBox)comp).getSelectedIndex()];
					  settings.handleMidiInputChange(inDevice);
				  }
			}
		});

		JLabel outPortsLabel = new JLabel("MIDI Output");
		outPortsBox = new JComboBox(outNames);
		int foundOutDevice = getIndexByName(outDevices, settings.getMidiOutputDevice());
		if (foundOutDevice != -1) {
			settings.setActualMidiOutputDevice(outNames[foundOutDevice]);
			outPortsBox.setSelectedIndex(foundOutDevice);
			settings.handleMidiOutputChange(outDevices[outPortsBox.getSelectedIndex()]);
		}
		else {
			settings.setActualMidiOutputDevice(outNames[outNames.length - 1]);
			outPortsBox.setSelectedIndex(outNames.length - 1);
			settings.handleMidiOutputChange(outDevices[outPortsBox.getSelectedIndex()], false);
		}
		outPortsBox.addItemListener(new ItemListener() {
			public void itemStateChanged( ItemEvent e ) {
				Object comp = e.getSource();
				if (e.getStateChange() == ItemEvent.SELECTED) {
					MidiDevice.Info outDevice = outDevices[((JComboBox)comp).getSelectedIndex()];
					settings.handleMidiOutputChange(outDevice);
				}

			}
		});
		
		JButton refreshButton = new JButton("Refresh Devices");
		refreshButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				controller.createSettingsDialog(true);
			}
		});

		JButton setMidiKeyboardButton = new JButton("Set MIDI Keyboard Size");
		setMidiKeyboardButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				setMidiKeyboardSize();
			}
		});
		
		JPanel buttonPanel = new JPanel();
		buttonPanel.add(refreshButton);
		buttonPanel.add(setMidiKeyboardButton);


		gbc.insets = new Insets(5, 5, 0, 5);
		gbc.fill = GridBagConstraints.BOTH;
		gbc.anchor = GridBagConstraints.WEST;
		add(inPortsLabel, gbc);
		
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		add(inPortsBox, gbc);
		
		gbc.insets = new Insets(0, 5, 0, 5);
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		add(outPortsLabel, gbc);

		gbc.gridwidth = GridBagConstraints.REMAINDER;
		add(outPortsBox, gbc);

		gbc.anchor = GridBagConstraints.CENTER;
		gbc.insets = new Insets(5, 5, 5, 5);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		add(buttonPanel, gbc);
	}
	
	private void setMidiKeyboardSize() {
		JPanel panel = new JPanel(new GridBagLayout());
		GridBagConstraints constraints = new GridBagConstraints();
		midiKeyboardSizeDialog = new JDialog(controller.getSettingsDialog(), "Set size of your MIDI Keyboard", true);
		JLabel firstKeyLabel = new JLabel("Press leftmost Key: ");
		firstKeyLabel.setOpaque(true);
		firstKeyLabel.setBackground((Color)UIManager.get("Slider.altTrackColor"));
		JLabel lastKeyLabel = new JLabel("Press rightmost Key: ");
		lastKeyLabel.setBackground((Color)UIManager.get("Slider.altTrackColor"));
		JLabel firstKeyValue = new JLabel("    ");
		JLabel lastKeyValue = new JLabel("    ");
		JButton okButton = new JButton("OK");
		okButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				midiKeyboardSizeDialog.setVisible(false);
				midiKeyboardSizeDialog = null;
				MidiDevice.Info inDevice = inDevices[inPortsBox.getSelectedIndex()];
				settings.handleMidiInputChange(inDevice, false, true);
			}
		});

		if (settings.getMidiInput() != null)
			settings.getMidiInput().close();
		new LocalMidiInput(controller, inDevices[inPortsBox.getSelectedIndex()], midiKeyboardSizeDialog, panel, firstKeyLabel, firstKeyValue, lastKeyLabel, lastKeyValue);

		
		constraints.gridx = 0;
		constraints.gridy = 0;
		constraints.fill = GridBagConstraints.BOTH;
		panel.add(firstKeyLabel, constraints);
		constraints.gridx = 1;
		constraints.insets = new Insets(0, 30, 0, 0);
		panel.add(firstKeyValue, constraints);
		constraints.insets = new Insets(0, 0, 0, 0);
		constraints.gridx = 0;
		constraints.gridy = 1;
		panel.add(lastKeyLabel, constraints);
		constraints.gridx = 1;
		constraints.insets = new Insets(0, 30, 0, 0);
		panel.add(lastKeyValue, constraints);
		constraints.insets = new Insets(0, 0, 0, 0);
		constraints.gridx = 0;
		constraints.gridy = 2;
		constraints.weightx = 0.0;
		constraints.fill = GridBagConstraints.NONE;
		constraints.gridwidth = 2;
		constraints.insets = new Insets(20, 0, 0, 0);
		panel.add(okButton, constraints);
		
		midiKeyboardSizeDialog.add(panel);
		midiKeyboardSizeDialog.pack();
		midiKeyboardSizeDialog.setVisible(true);
	}

	/**
	 * Returns the index of a named MIDI device or -1 if no device with this name is available
	 * 
	 * @param midiDevice the MIDI device Array to look in
	 * @param midiDeviceName the name of the MIDI device
	 * @return the index of the MIDI device or -1 if no device is available
	 */
	public int getIndexByName(MidiDevice.Info[] midiDevice, String midiDeviceName) {
		if (midiDeviceName == null)
			return midiDevice.length-1;
		for (int i=0; i<midiDevice.length; i++) {
			if (midiDeviceName.equals(midiDevice[i].getName()))
				return i;
		}
		return -1;
	}
}

class LocalMidiInput extends MidiInput {
 	private JLabel firstKeyLabel, lastKeyLabel, firstKeyValue, lastKeyValue;
	private int incomingEvent = 0;
	private JPanel panel;
	private JDialog dialog;
	private Controller controller;

	public LocalMidiInput(Controller c, MidiDevice.Info m, JDialog d, JPanel p, JLabel f, JLabel fv, JLabel l, JLabel lv) {
		super(c, m);
		controller = c;
		panel = p;
		dialog = d;
		firstKeyLabel = f;
		lastKeyLabel = l;
		firstKeyValue = fv;
		lastKeyValue = lv;
	}
	
	public void send(MidiMessage message, long timeStamp) {
		if (message instanceof ShortMessage) {
			ShortMessage m = (ShortMessage)message;
			if (m.getCommand() == ShortMessage.NOTE_ON && m.getData2() > 0) {
				if (incomingEvent == 0) {
					lastKeyLabel.setBackground((Color)UIManager.get("Slider.altTrackColor"));
					lastKeyLabel.setOpaque(true);
					firstKeyValue.setText(Integer.toString(m.getData1()-21));
					controller.getSettings().setFirstKey(m.getData1()-21);
					firstKeyLabel.setOpaque(false);
				}
				else {
					firstKeyLabel.setBackground((Color)UIManager.get("Slider.altTrackColor"));
					firstKeyLabel.setOpaque(true);
					lastKeyValue.setText(Integer.toString(m.getData1()-21));
					controller.getSettings().setLastKey(m.getData1()-21);
					lastKeyLabel.setOpaque(false);
				}	
				panel.revalidate();
				dialog.repaint();
				incomingEvent = (incomingEvent+1)%2;
			}
		}
	}
}