/*
 *  OpenPiano - a MIDI-based piano trainer with an easy note presentation
 *  Copyright (C) 2008  Sascha Westermann
 *
 *  This program is free software; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *  
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 */

package openpiano.settings;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.File;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import openpiano.control.Controller;

/**
 * A class to download and install the soundbank file for the java synthesizer
 *
 * @author  Sascha Westermann
 * @version 1.0, 07/30/08
 */
public class SoundbankCheck {
	private Boolean failed = false;
	private Boolean remind = true;
	
	/**
	 * Creates a soundbank object
	 *
	 * @param controller the main controller of the application
	 */
	public SoundbankCheck(Controller controller) {
		String separator = System.getProperty("file.separator");
		URL soundbank = null;
		File soundbankDir = new File(System.getProperty("java.home") + separator + "lib" + separator + "audio");
		File soundbankFile1 = new File(soundbankDir.getAbsoluteFile() + separator + "soundbank.gm");
		File soundbankFile2 = new File(soundbankDir.getAbsoluteFile() + separator + "soundbank-deluxe.gm");
		
	    GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
	    GraphicsDevice gd = ge.getDefaultScreenDevice();
        GraphicsConfiguration gc = gd.getDefaultConfiguration();
        Double centerX = gc.getBounds().getCenterX();
        Double centerY = gc.getBounds().getCenterY();
	
		if (!soundbankFile1.exists() && !soundbankFile2.exists()) {
			if (!soundbankDir.exists())
				if (!soundbankDir.mkdir())
					failed = true;
			try {
				soundbank = new URL("http://java.sun.com/products/java-media/sound/soundbank-deluxe.gm.zip");

				int downloadSize = soundbank.openConnection().getContentLength();
				
				InputStream is = soundbank.openStream();
				File soundbankZIP = new File(soundbankFile2.getAbsoluteFile() + ".zip");
				FileOutputStream fos = new FileOutputStream(soundbankZIP);

				BufferedInputStream bis = new BufferedInputStream(is);
				BufferedOutputStream bos = new BufferedOutputStream(fos);

				byte[] buff = new byte[2048];
				int bytesRead;
				int totalBytesRead = 0;
				
				JDialog downloadDialog = new JDialog(controller.getOpenPiano(), "Soundbank", false);
				JLabel progressLabel = new JLabel("Download in progress...   0%");
				JPanel progressPanel = new JPanel(new BorderLayout());
				progressPanel.add(progressLabel, BorderLayout.CENTER);
				downloadDialog.add(progressPanel);
		        downloadDialog.pack();
				downloadDialog.setLocation(Math.round(centerX.floatValue() - downloadDialog.getWidth()/2.0f),Math.round(centerY.floatValue() - downloadDialog.getHeight()/2.0f));
		        downloadDialog.setVisible(true);
				
				while( (bytesRead = bis.read(buff, 0, buff.length)) != -1 ) 
				{
					totalBytesRead+=bytesRead;
					progressLabel.setText("Download in progress... " + Math.round(((float)totalBytesRead/(float)downloadSize)*100) + "%");
					progressPanel.revalidate();
					downloadDialog.validate();
					bos.write(buff, 0, bytesRead);
					bos.flush();						
				}
				downloadDialog.dispose();
				bis.close();
				bos.close();
				
				ZipInputStream zis = new ZipInputStream(new FileInputStream(soundbankZIP));
				bos = new BufferedOutputStream(new FileOutputStream(soundbankFile2));
				ZipEntry entry;
				while ((entry = zis.getNextEntry()) != null) {
					if (entry.getName().equals("soundbank-deluxe.gm")) {
						while( (bytesRead = zis.read(buff, 0, buff.length)) != -1 ) {
							bos.write(buff, 0, bytesRead);
							bos.flush();						
						}
					}
				}
				zis.close();
				bos.close();

			} catch (MalformedURLException mfue) { failed = true; }
			catch (IOException ioe) { failed = true; };
		}
		
		if (failed) {
			JDialog soundbankDialog = new JDialog(controller.getOpenPiano(), "Installation of Soundbank failed", true);
			SoundbankFailedPanel soundbankQuestion = new SoundbankFailedPanel();
			soundbankDialog.add(soundbankQuestion);
	        soundbankDialog.pack();
			soundbankDialog.setLocation(Math.round(centerX.floatValue() - soundbankDialog.getWidth()/2.0f),Math.round(centerY.floatValue() - soundbankDialog.getHeight()/2.0f));
	        soundbankDialog.setVisible(true);
	        remind = soundbankQuestion.isRemindOn();
		}
	}
	
	/**
	 * Return <tt>true</tt> if the installation failed
	 * 
	 * @return <tt>true</tt> if the installation failed
	 */
	public Boolean hasFailed() { return failed; }
	
	/**
	 * Returns <tt>true</tt> if the reminder is on
	 * 
	 * @return <tt>true</tt> if the reminder is on
	 */
	public Boolean isRemindOn() { return remind; }
	
	private class SoundbankFailedPanel extends JPanel {
		private static final long serialVersionUID = 1069444189551259657L;
		private JCheckBox checkRemind;
		private JButton okButton;
		private Boolean remind = true;
		
		public SoundbankFailedPanel() {
			setLayout(new BorderLayout());
			
			String separator = System.getProperty("file.separator");
			String soundbankDir = new String(System.getProperty("java.home") + separator + "lib" + separator + "audio");

			JLabel informationLabel= new JLabel("<html>The Soundbank for the internal Java Synthesizer could not be installed. If<br>" +
					"you want to use the Synthesizer, you have to download it manually from<br>" +
					"http://java.sun.com/products/java-media/sound/soundbank-deluxe.gm.zip<br>" +
					"and unzip it to " + soundbankDir + "</html>" );
			JPanel information = new JPanel();
			information.add(informationLabel);
			add(information, BorderLayout.NORTH);
			
			checkRemind = new JCheckBox("Never remind me again");
			checkRemind.addItemListener(new ItemListener() {
				 public void itemStateChanged(ItemEvent e) {
					 if (checkRemind.isSelected())
						 remind = false;
					 else
						 remind = true;
				 }
			});
			
			okButton = new JButton("OK");
			okButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					Container c = okButton;
					while ( !(c.getParent() instanceof Window) ) {
						c = c.getParent();
					}
					Window w = (Window)(c.getParent());
					w.dispose();
				}
			});
			
			JPanel options = new JPanel(new BorderLayout());
			options.add(checkRemind, BorderLayout.WEST);
			options.add(okButton, BorderLayout.EAST);
			
			add(options, BorderLayout.SOUTH);
		}
		
		public Boolean isRemindOn() { return remind; }
	}
}